# -*- coding: utf-8 -*-
"""
Created on Mon Sep 17 16:16:22 2018

@author: jcnossen1
"""
import sys

sys.path.append("..")
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from smlmlib.gaussian import Gaussian
from smlmlib.base import SMLM
from smlmlib.spotdetect import SpotDetectorConfig
import smlmlib.util as su
import math
from scipy.stats import poisson
import time
import os
import tifffile


def generate_storm_movie(gaussian, emitterList, numframes=100, imgsize=512, intensity=500, bg=2, sigma=1.5, p_on=0.1):
    frames = np.zeros((numframes, imgsize, imgsize), dtype=np.uint16)
    emitters = np.array([[e[0], e[1], sigma, sigma, intensity] for e in emitterList])

    on_counts = np.zeros(numframes, dtype=np.int32)

    for f in range(numframes):
        frame = bg * np.ones((imgsize, imgsize), dtype=np.float32)
        frame_emitters = emitters * 1
        on = np.random.binomial(1, p_on, len(emitters))
        frame_emitters[:, 4] *= on

        frames[f] = gaussian.Draw(frame, frame_emitters)
        on_counts[f] = np.sum(on)

    return frames, on_counts


psfSigma = 3
roisize = int(2 + (psfSigma + 1) * 2)
w = 256
N = 2000
numframes = 100
R = np.random.normal(0, 0.2, size=N) + w * 0.3
angle = np.random.uniform(0, 2 * math.pi, N)
emitters = np.vstack((R * np.cos(angle) + w / 2, R * np.sin(angle) + w / 2)).T


with SMLM(debugMode=False) as smlm:
    gaussian = Gaussian(smlm)
    mov, on_counts = generate_storm_movie(gaussian, emitters, numframes, imgsize=w, sigma=psfSigma, p_on=20 / N)
    mov = np.random.poisson(mov)

    fn = "test_localization_queue.mp4"
    print("Saving movie to {0}".format(fn))
    os.makedirs("../data", exist_ok=True)
    tifffile.imsave("../data/palm_ring_generated.tiff", np.array(mov, dtype=np.uint16))
    # su.save_movie(mov, fn, 20)

    scale = 1
    result_image = np.zeros((w * scale, w * scale), dtype=np.float32)

    spotDetectorConfig = SpotDetectorConfig(psfSigma, roisize, minIntensity=2)

    with gaussian.CreateImageQueue(w, w, psfSigma, roisize, 500, spotDetectorConfig) as q:
        t0 = time.time()

        q.Start()
        for f in range(len(mov)):
            q.PushFrame(f, mov[f])

        while not q.IsIdle():
            print("Queue length: {0}, Frames processed: {1}".format(q.GetQueueLength(), q.GetResultFrameCount()))
            time.sleep(0.5)

        dt = time.time() - t0
        print(f"Processed {len(mov)} frames in {dt} seconds. {len(mov)/dt:.3f} fps")

        for f in range(len(mov)):
            fr = q.GetFrameResults(f)
            if f % 40 == 0:
                print("Frame {0} result count: {1}. true on-counts: {2}".format(f, len(fr), on_counts[f]))

            if len(fr)>0:
                gaussian.Draw(
                    result_image,
                    [
                        [
                            scale * (r.estimate[0] + r.roiPosition[0]),
                            scale * (r.estimate[1] + r.roiPosition[1]),
                            1,
                            1,
                            r.estimate[2],
                        ]
                        for r in fr
                    ],
                )

        for r in q.GetFrameResults(0):
            est = r.estimate
            roi = r.roiPosition
            print(f"x={est[0]}, y={est[1]}, I={est[2]}, bg={est[3]}. roix={roi[0]}, roiy={roi[1]}")


plt.figure()
plt.imshow(mov[0])
plt.title("Frame 0")

plt.figure()
plt.imshow(result_image)
plt.title("Result")
