# Localize.py does the following:
# - Runs 2D Gaussian pipeline on a tiff file
# - Runs drift estimation and correction
# - Save the resulting point cloud to picasso compatible HDF5 and CSV
import numpy as np
from smlmlib.base import SMLM
import matplotlib.pyplot as plt

import utils.localizations as loc
import os
import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process TIFF file")
    parser.add_argument("-i", "--input", help="Input TIFF file (.tif)")
    parser.add_argument("-o", "--output", default=None, help="Output dir")
    parser.add_argument("--sigma", default=2, help="PSF width in pixels")
    parser.add_argument("--offset", default=0, help="Camera calibration offset")
    parser.add_argument("--gain", default=1, help="Camera calibration gain")
    parser.add_argument("--roi", default=10, help="Gaussian fit ROI size")
    parser.add_argument("--threshold", default=5, help="Detection threshold")
    parser.add_argument("--driftbins", default=100, help="Timebin size for drift correction")    
    parser.add_argument("--rerun", default=False, action="store_true", help="Ignore cached results")
    parser.add_argument(
        "-n", "--num-frames", type=int, default=None, help="Number of frames to process. Leave blank to process all"
    )

    args = parser.parse_args()
    
    if args.input:
        
        cfg = { 
           'sigma':float(args.sigma),
           'roisize':int(args.roi),
           'maxSpotsPerFrame':2000,
           'detectionMinIntensity':float(args.threshold),
           'detectionMaxIntensity':1e6,
           'offset':float(args.offset),
           'gain':float(args.gain)
           }
        
        if args.num_frames:
            cfg['maxframes'] = args.num_frames
        
        if not args.output: 
            output =  os.path.splitext(args.input)[0] + '_results/'
        else:
            output = args.output
    
        os.makedirs(output, exist_ok=True)
        
        dataname = os.path.split(args.input)[1]
        
        with SMLM() as smlm:
            result = loc.load(args.input, cfg, smlm, rebuild=args.rerun)
            result.drift_estimate_cc(smlm, args.driftbins, smooth_sigma=4)
            result.apply_drift_correction(result.drift, args.driftbins)
                
            fig = result.plot_drift(f'Estimated drift {dataname}')
            fig.savefig(output + "drift.png")
            
            result.save_csv(output + "locs.csv")
            result.save_picasso_hdf5(output + "locs.hdf5")
            
    else:
        parser.print_help()
        
            
        