"""

Main SIMFLUX data processing pipeline.

- Takes TIFF path and known angle and frequencies
- Does spot detection and localization on summed-frames
- Estimates phases on the dataset split in 10 pieces (to know the precision of the phase estimates)
- Filters out all spots that have frames with intensity below certain threshold (= min filter)
- Run non-blinking SIMFLUX localization on it.
- For both Gauss2D and SIMFLUX:
    - Apply drift correction on the results

"""
import numpy as np
from smlmlib.base import SMLM
from smlmlib import util as su
import gaussian.fitters as gaussfit
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import scipy.io as sio
import utils.peaks as peaks
import utils.localizations as loc
from util.spotlist import SpotList
import matplotlib.patches
import math
import os,pickle

figsize=(9,7)


# Make sure the angles dont wrap around, so you can plot them and take mean
def unwrap_angle(ang):
    r = ang * 1
    ang0 = ang.flatten()[0]
    r[ang > ang0 + math.pi] -= 2 * math.pi
    r[ang < ang0 - math.pi] += 2 * math.pi
    return r


# Pattern angles wrap at 180 degrees
def unwrap_pattern_angle(ang):
    r = ang * 1
    ang0 = ang.flatten()[0]
    r[ang > ang0 + math.pi / 2] -= math.pi
    r[ang < ang0 - math.pi / 2] += math.pi
    return r


def print_phase_info(mod):
    for axis in [0, 1]:
        steps = np.diff(mod[axis::2, 3])
        steps[steps > np.pi] = -2 * np.pi + steps[steps > np.pi]
        print(f"axis {axis} steps: {-steps*180/np.pi}")

# Get the rows in mod for which the given axis is the dominant axis, get_axis_mod(mod,0) gives all X modulations
def get_axis_mod(mod, axis):
    dominant_axis = np.argmax(np.abs(mod[:,[0,1]]),1)
    return np.nonzero(dominant_axis==axis)[0]

# Flip kx,ky,phase
def flip_mod(mod_row):
    mod_row[[0,1,3]] = -1*mod_row[[0,1,3]]
    mod_row[3] += np.pi
    return mod_row

# Flip modulation directions so they all align with the first modulation pattern of a particular axis
def fix_mod_angles(mod):
    mod = 1*mod
    k = mod[:,[0,1]]
    for axis in range(2):
        am = get_axis_mod(mod,axis)
#        print(f"am0: {mod[am[0]]}")
        for i in np.arange(1,len(am)):
            if np.dot(k[am[0]],k[am[i]])<0: 
#                print(f"should flip axis {am[i]}: {mod[am[i]]}")
                mod[am[i]] = flip_mod(mod[am[i]])
    return mod

def draw_mod(allmod, showPlot=True, filename=None):
    fig,axes = plt.subplots(1,2)
    fig.set_size_inches(*figsize)
    for axis in range(2):
        axisname = ['X', 'Y']
        ax = axes[axis]
        indices = get_axis_mod(allmod,axis)
        freq = np.sqrt(allmod[indices[0],0]**2+allmod[indices[0],1]**2)
        period = 2*np.pi/freq
        x = np.linspace(0, period, 200)
        sum = x*0
        for i in indices:
            mod = allmod[i]
            q = (1+mod[2]*np.sin(x*freq-mod[3]) )*mod[4]
            ax.plot(x, q, label=f"Pattern {i}")
            sum += q
        ax.plot(x, sum, label=f'Summed {axisname[axis]} patterns')
        ax.legend()
        ax.set_title(f'{axisname[axis]} modulation')
        ax.set_xlabel('Pixels');ax.set_ylabel('Modulation intensity')
    fig.suptitle('Modulation patterns')
    if filename is not None: fig.savefig(filename)
    if not showPlot: plt.close(fig)
    return fig

def result_dir(path):
    dir, fn = os.path.split(path)
    return dir + "/results/" + os.path.splitext(fn)[0] + "/"


        
def load_mod(tiffpath):
    with open(os.path.splitext(tiffpath)[0]+"_mod.pickle", "rb") as pf:
        mod = pickle.load(pf)['mod']

        return mod
    
    
def density_filter(searchdist, mincount=5):
    def filterfn(lr : loc.LocResultList):
        counts = lr.count_neighbors(searchdist)
        def locfilter(i, locs):
            return counts[i] >= mincount
        lr.filter(locfilter)
        
    return filterfn


def save_spots_for_plots(path, spotlist: SpotList, mod, nspots=1000):
    data = (
            spotlist.IBg[:nspots],
            spotlist.rois[:nspots],
            spotlist.result.get_xyIBg()[:nspots],
            spotlist.result.get_roipos()[:nspots],
            mod)

    with open(os.path.splitext(path)[0]+"_selspots.pickle", "wb") as pf:
        pickle.dump(data, pf)
    


def print_mod(reportfn, mod, pixelsize):
    for k in range(len(mod)):
        reportfn(f"Pattern {k}: kx={mod[k,0]:.4f} ky={mod[k,1]:.4f} Phase {mod[k,3]*180/np.pi:8.2f} Depth={mod[k,2]:5.2f} "+
               f"Power={mod[k,4]:5.3f} ")

    xpat = get_axis_mod(mod, 0)
    ypat = get_axis_mod(mod, 1)

    depthX = np.mean(mod[xpat,2])
    depthY = np.mean(mod[ypat,2])

    phaseX = mod[xpat,3]
    phaseY = mod[ypat,3]
    
    shiftsX = (np.diff(phaseX[-1::-1]) % (2*np.pi)) * 180/np.pi
    shiftsY = (np.diff(phaseY[-1::-1]) % (2*np.pi)) * 180/np.pi
    
    shiftsX[shiftsX > 180] = 360 - shiftsX[shiftsX>180]
    shiftsY[shiftsY > 180] = 360 - shiftsY[shiftsY>180]
    
    with np.printoptions(precision=3, suppress=True):
        reportfn(f"X Shifts: {shiftsX} (deg) (patterns: {xpat}). Depth={depthX:.3f}")
        reportfn(f"Y Shifts: {shiftsY} (deg) (patterns: {ypat}). Depth={depthY:.3f}")

    freq = np.sqrt(mod[:,0]**2+mod[:,1]**2)
    pitch = 2*np.pi/freq * pixelsize
    reportfn(f"Refined pitch pattern X: {pitch[xpat[0]]:.5f} nm, pattern Y: {pitch[ypat[0]]:.5f} nm")

    return shiftsX, shiftsY

def estimate_sigma_from_tiff(path, reportfn, numframes=1):
    fr = loc.read_tiff_frames(path, np.arange(numframes))
#    fr = (fr.astype(np.float32) - cfg['offset']) / cfg['gain']
 #   fr[fr<=0] = 1e-6
    
    return gaussfit.estimate_sigma_from_movie(fr, roi_size=20)    

def process(path, cfg, num_patterns, pixelsize, num_phase_bins = 10,
            plot_ffts = False,
            estimate_angle_bins = 1, 
            drift_correct=False, 
            draw_silm_compare=False, 
            show_plots=True, 
            debugMode=False,
            fix_depths=False,
            freq_minmax=[1.5,2],
            minfilter=50,
            fix_phase_shifts=None): 
    
    if (len(np.array(num_patterns).shape) > 0):
        mod = num_patterns
    else:
        mod=np.zeros((num_patterns,5))
    
    dir, fn = os.path.split(path)
    resultsdir = result_dir(path)
    os.makedirs(resultsdir, exist_ok=True)
    resultprefix = resultsdir
        
    f = open(resultprefix + "report.txt", "w")
    def report(msg):
        print(msg)
        f.write(msg+"\n")
        
    report(f"Using spot detection threshold {cfg['detectionMinIntensity']}")
    report(f"Using gain {cfg['gain']}, offset {cfg['offset']}")
    report(f"Using PSF sigma {cfg['sigma']}")

    with SMLM(debugMode) as smlm:
        loc_sum = loc.load(path, {**cfg, "sumframes": len(mod)}, smlm, tag="sum")

        numnan = loc_sum.filter_nan()
        numbad = loc_sum.filter(lambda i,d: np.sum(d[:, [5, 6]], 1) < 4)  # CRLB X,Y < 4

#        loc_sum.save_picasso_hdf5(resultprefix+"g2d.hdf5")

        spotlist = SpotList(loc_sum, pixelsize=pixelsize, outdir=resultsdir)
        spotlist.estimate_IBg(path, smlm, len(mod), store_rois=True, rebuild=False)
        
        median_crlb_x = np.median(loc_sum.get_crlb()[:,0])
        median_I = np.median(loc_sum.get_xyI()[:,2])

        nanibg = np.nonzero(np.sum(np.isnan(spotlist.IBg),(1,2)))[0]
        report(f"#nan spots: {numnan}. unconverged: {numbad}. #nan ibg: {len(nanibg)}. g2d mean I={median_I:.1f}. mean crlb x {median_crlb_x:.4f}")

        fr = np.arange(len(loc_sum.frames))

        if estimate_angle_bins is not None:
            angles, pitch = spotlist.estimate_angle_and_pitch(
                len(mod), frame_bins=np.array_split(fr, estimate_angle_bins), smlm=smlm, plotTag=None,
                freq_minmax=freq_minmax
            )

            print("Pitch and angle estimation: ")
            for k in range(len(mod)):
                angles[angles[:, k] > 0.6 * np.pi] -= np.pi  # 180 deg to around 0
                angles[:, k] = unwrap_pattern_angle(angles[:, k])
                angles_k = angles[:, k]
                pitch_k = pitch[:, k]
                report(f"Pattern {k}: Angle: { np.mean(angles_k) :7.5f} [deg]. Pitch: {np.mean(pitch_k)*pixelsize:10.5f} ({2*np.pi/np.mean(pitch_k):3.3f} [rad/nm])")

                freq = 2 * np.pi / np.mean(pitch_k)
                kx = np.cos(np.mean(angles_k)) * freq
                ky = np.sin(np.mean(angles_k)) * freq
                mod[k, [0, 1]] = kx, ky
                            
        else:
            q = mod[:, 0] + 1j * mod[:, 1]
            pitch = 2 * np.pi / np.abs(q)
            angles = np.angle(q)
            print(angles)
            
        mod = fix_mod_angles(mod)

        # Average ky 
        mod[::2,1] = np.mean(mod[::2,1]) 
        mod[1::2,1] = np.mean(mod[1::2,1]) 

        # Average kx        
        mod[::2,0] = np.mean(mod[::2,0]) 
        mod[1::2,0] = np.mean(mod[1::2,0]) 
        
#        mod[0::2,0]= 0
 #       mod[1::2,1]= 0

        phase,depth,power = mod[:,3],mod[:,2],mod[:,4]
        phase_all=phase
        depth_all=depth
        mod_info = {"mod": mod, "pitch": pitch, "angles": angles, "phase": phase, "depth": depth, 'power': power}
            
        if num_phase_bins is not None:
            def phase_depth_estim(method):
                phase, depth, power = spotlist.estimate_phase_and_depth(mod, np.array_split(fr, num_phase_bins), method=method)
                phase_all, depth_all, power_all = spotlist.estimate_phase_and_depth(mod, [fr], method=method)
        
                axis = lambda k: ['x','y'][int(np.abs(mod[k,0])>np.abs(mod[k,1]))]
                
                fig = plt.figure(figsize=figsize)
                styles = [":", "-"]
                for k in range(len(mod)):
                    plt.plot(unwrap_angle(phase[:, k]) * 180 / np.pi, styles[k % 2], label=f"Phase {k} ({axis(k)})")
                plt.legend()
                plt.title(f"Phases for {fn}")
                plt.xlabel("Timebins"); plt.ylabel("Phase [deg]")
                plt.grid()
                plt.tight_layout()
                fig.savefig(resultprefix + "phases.png")
                if not show_plots: plt.close(fig)
        
                fig = plt.figure(figsize=figsize)
                for k in range(len(mod)):
                    plt.plot(depth[:, k], styles[k%2], label=f"Depth {k} ({axis(k)})")
                plt.legend()
                plt.title(f"Depths for {fn}")
                plt.xlabel("Timebins"); plt.ylabel("Modulation Depth")
                plt.grid()
                plt.tight_layout()
                fig.savefig(resultprefix + "depths.png")
                if not show_plots: plt.close(fig)
        
                fig = plt.figure(figsize=figsize)
                for k in range(len(mod)):
                    plt.plot(power[:, k], styles[k%2], label=f"Power {k} ({axis(k)})")
                plt.legend()
                plt.title(f"Power for {fn}")
                plt.xlabel("Timebins"); plt.ylabel("Modulation Power")
                plt.grid()
                plt.tight_layout()
                fig.savefig(resultprefix + "power.png")
                if not show_plots: plt.close(fig)
        
                # Update mod
                phase_std = np.zeros(len(mod))
                for k in range(len(mod)):
                    ph_k = unwrap_angle(phase[:, k])
                    mod[k, 3] = phase_all[0, k]
                    mod[k, 2] = depth_all[0, k]
                    mod[k, 4] = power_all[0, k]
                    phase_std[k] = np.std(ph_k)

                s=np.sqrt(num_phase_bins)
                for k in range(len(mod)):
                    report(f"Pattern {k}: Phase {mod[k,3]*180/np.pi:8.2f} (std={phase_std[k]/s*180/np.pi:6.2f}) "+
                           f"Depth={mod[k,2]:5.2f} (std={np.std(depth[:,k])/s:5.3f}) "+
                           f"Power={mod[k,4]:5.3f} (std={np.std(power[:,k])/s:5.5f}) ")

#            report("Estimating phase, depth, power with these pitch values:")
 #           for k in range(len(mod)):
  #              report(f"Pattern {k}: kx={mod[k,0]:.8f} ky={mod[k,1]:.8f} ")

            phase_depth_estim(method=1)
            mod=spotlist.refine_pitch(mod, smlm, minfilter, plot=True)[2]

            if fix_phase_shifts:
                report(f'Fixing phase shifts to {fix_phase_shifts}' )
                phase_shift_rad = fix_phase_shifts / 180 * np.pi
                mod[::2,3] = mod[0,3] + np.arange(len(mod)//2) * phase_shift_rad
#                mod[::2,3] += 0.1
                mod[1::2,3] = mod[1,3] + np.arange(len(mod)//2) * phase_shift_rad
            
                for n in range(3):
                    mod=spotlist.refine_pitch(mod, smlm, minfilter, plot=(n==0))[2]

            info_mat = {**mod_info, 'phase': phase_all, 'depth': depth_all}
            sio.savemat(resultprefix + "info.mat", info_mat)

            mod_info = {"mod": mod, "pitch": pitch, "angles": angles, "phase": phase, "depth": depth, 'power': power}
            with open(os.path.splitext(path)[0]+"_mod.pickle", "wb") as df:
                pickle.dump(mod_info, df)

        # Average relative intensity
        mod[::2,4] = np.mean(mod[::2,4])
        mod[1::2,4] = np.mean(mod[1::2,4])
        mod[:,4] /= np.sum(mod[:,4])
        
        # Average modulation depth
        mod[::2,2] = np.mean(mod[::2,2])
        mod[1::2,2] = np.mean(mod[1::2,2])
        
        if fix_depths:
            report(f'Fixing modulation depth to {fix_depths}' )
            mod[::2,2]=fix_depths
            mod[1::2,2]=fix_depths

        report("Final modulation pattern parameters:")
        print_mod(report, mod, pixelsize)
        
        draw_mod(mod, False, resultprefix + "patterns.png")
        #spotlist.silm_bias_plot2D(mod, smlm, silm_min_intensity, tag='')

        if plot_ffts:
            spotlist.generate_projections(mod, 4, smlm)
            spotlist.plot_proj_fft()

        med_sum_I = np.median(spotlist.IBg[:,:,0].sum(1))
        lowest_power = np.min(mod[:,4])
        depth = mod[np.argmin(mod[:,4]),2]
        median_intensity_at_zero = med_sum_I * lowest_power * (1-depth)
        report(f"Median summed intensity: {med_sum_I:.1f}. Median intensity at pattern zero: {median_intensity_at_zero:.1f}")
        report(f"Using min-filter at { minfilter }")

        for k in range(len(mod)):
            spotlist.draw_spots_in_pattern(f"{spotlist.outdir}patternspots{k}.png", mod, 
                                       k, tiffname=os.path.split(fn)[1], numpts= 2000, min_intensity=minfilter)



        # g2d_results are the same set of spots used for silm, for fair comparison
        silm_results,g2d_results = spotlist.silm(mod,smlm, min_intensity=minfilter,
                                                 draw_silm_compare=draw_silm_compare)

        drift = None
        framebinning=200
        
        if drift_correct == True or drift_correct == 'cc':
            drift = loc_sum.drift_estimate_cc(smlm, framebinning, scale=6, useIntensities=False, smooth_sigma=4, maxdrift=3)
            loc_sum.plot_drift(f"Drift estimate for {fn}").savefig(resultprefix + "drift.png")

            sio.savemat(resultprefix+"drift.mat", {'drift': loc_sum.drift})
            silm_results.apply_drift_correction(drift, framebinning)
            g2d_results.apply_drift_correction(drift, framebinning)
        elif drift_correct:
            drift_fn = os.path.split(path)[0]+ "/" + drift_correct
            g2d_results.drift_correct_from_file(drift_fn, len(mod))
            silm_results.drift_correct_from_file(drift_fn, len(mod))
            silm_results.plot_drift('drift' )

        elif drift_correct == "nn":
            raise NotImplementedError()

        silm_results.save_picasso_hdf5(resultprefix+"simflux.hdf5")
        g2d_results.save_picasso_hdf5(resultprefix+"g2d.hdf5")

        silm_data = silm_results.save_csv_with_drift(resultprefix + "simflux.csv")
        g2d_data = g2d_results.save_csv_with_drift(resultprefix + "g2d.csv")
        sio.savemat(resultprefix + "localizations.mat", {"simflux": silm_data, "g2d": g2d_data})
                                    
        image = loc_sum.draw(2,smlm)
        f.close()
        
#        save_spots_for_plots(path, spotlist,mod)
        
        return spotlist, image, mod, silm_results, g2d_results

