"""
Minimal example of using the SILM C++ methods

INSTALL:
pip install PyQt5==5.9.2
pip install scipy matplotlib numpy

Recommended IDE:
pip install spyder
spyder3
"""
import numpy as np
import math
import sys
import matplotlib.pyplot as plt

sys.path.append("../..")

import smlmlib.gaussian
import smlmlib.silm
import smlmlib.util as smlm_util

    
def test():
    np.random.seed(0)

    # Load DLL and API objects
    with smlmlib.base.SMLM(debugMode=True) as smlm:
        silm = smlmlib.silm.SILM(smlm)
        g2d_api = smlmlib.gaussian.Gaussian(smlm)
        
        # Generate some spots
        numspots = 2
        numEP = 6
        sigma = 2
        roisize = 12
        silm_param = silm.Params(roisize, numEP, sigma, levMarIt=40)
    
        # Compute phases such that sum of excitation patterns is always 1
        mod = np.array([
                   [0, 1.8,   0.95, 0, 1/6],
                   [1.9, 0, 0.95, 0, 1/6],
                   [0, 1.8,   0.95, 2*np.pi/3, 1/6],
                   [1.9, 0, 0.95, 2*np.pi/3, 1/6],
                   [0, 1.8,   0.95, 4*np.pi/3, 1/6],
                   [1.9, 0, 0.95, 4*np.pi/3, 1/6]
                  ])
        patternPitch = 2*np.pi/mod[0,1] # pixels
        mod = np.tile(mod, (numspots,1,1))
    
        bg = 100
        # Generate positions. We set bg=numEP so every individual pattern image has 6/6 = 1 photon bg noise
        theta = (
            np.tile([roisize / 2, roisize / 2, 6000, bg*numEP], (numspots, 1))
            + np.random.uniform([-1, -1, 0, 0], [1, 1, 0, 0], size=(numspots, 4)) * patternPitch / 2
        )  #  move xy in a range of 2 pixels
        print(theta)
        
        startPatterns = np.zeros(numspots)
        roipos = np.zeros((numspots,2),dtype=np.int32)
    
        # Compute expected values
        numframes = 6
        mu, fi = silm.SILM_ASW_ComputeFisherMatrix(mod, theta, numframes, startPatterns, silm_param)
    
        # Sample
        smp = np.random.poisson(mu)

        # Gauss fits
        with g2d_api.CreatePSF_XYIBg(roisize,sigma, cuda=True) as psf:
            g2d_theta = psf.ComputeMLE(smp.sum(1))[0]
            
            g2d_theta_repeat=np.repeat(g2d_theta[:,[0,1]], numframes,0)
            # Fit intensities
            IBg0, IBg_crlb = g2d_api.EstimateIBg(
                    smp.reshape((numframes*numspots,roisize,roisize)), 
                    sigma, g2d_theta_repeat)
            IBg = IBg0.reshape((numspots,numframes,2))

        # Compute on/off probablities per frame per spot
        pOnOff, IBg2, crlbVariance, expectedIntensity = silm.SILM_ASW_ComputeOnOffProb(smp, mod, g2d_theta, silm_param, startPatterns, roipos, useCuda=False)
        pOnOff = np.exp(pOnOff)
        totals = np.sum(pOnOff,2)
        pOnOff /= totals[:,:,np.newaxis]

        for s in range(numspots):
#            print(f"spot {s}: p(on): {pOnOff[s,:,0]}\n I_f={IBg[s,:,0]} \n")
            fig,axes=plt.subplots(2,1,sharex=True)
            axes[0].plot(IBg2[s,:,0], label="I(f)")
            axes[0].plot(IBg[s,:,0], label="I(f) (check)")
            axes[0].plot(expectedIntensity[s,:], label="Expected I(f)")
            axes[0].plot(crlbVariance[s,:,0], '--', label="Var I(f) on")
            axes[0].plot(crlbVariance[s,:,1], '--', label="Var I(f) off")
#            plt.plot(expectedIntensity[s,:] + crlbVariance[s,:,0], label="Expected I(f) + Var(I(f))")
            axes[0].legend()
            
            axes[1].plot(pOnOff[s,:,0], label="p(on)")
            axes[1].plot(pOnOff[s,:,1], label="p(off)")
            axes[1].legend()
    
        # Plot samples
        for s in range(numspots):
            smlm_util.imshow_hstack(smp[s],f'spot {s}' )
        
        est, results = silm.SILM_ASW_ComputeMLE(smp, mod, silm_param, startPatterns, roipos=roipos)
        print (f"Non-blink estimates: \n{est}")

if __name__ == "__main__":
    test()
    