# SILM CreateImageQueue example
import numpy as np
import matplotlib.pyplot as plt
import math
import time
import tifffile

import sys

sys.path.append("..")

from smlmlib.gaussian import Gaussian
import smlmlib.silm
import smlmlib.spotdetect as spotdetect
import smlmlib.util as su

# Select which DLL to use
debugMode = False
smlm = smlmlib.base.SMLM(debugMode)


# Simulate blinking of a set of emitters at known positions. ExcitationPattern is a function (theta, frame)
def generate_blink_trace(T_on, T_off, theta, nframes, excitationPattern, fps):
    numEmitters = len(theta)
    states = np.random.binomial(1, T_on / (T_on + T_off), numEmitters)
    ontimes = np.zeros((numEmitters, nframes))

    sys.stdout.write("simulating blinking")

    for i in range(numEmitters):
        state = states[i]
        remaining = np.random.exponential(T_on if state else T_off)
        if i % 100 == 0:
            sys.stdout.write(".")

        for f in range(nframes):
            dt = 1 / fps
            q = excitationPattern(theta[i], f)  # the intensity q between 0 and 1

            onTime = 0
            while remaining <= dt * q:
                dt -= remaining / q
                if state:
                    onTime += remaining
                    state = 0
                    remaining = np.random.exponential(T_off)
                else:
                    remaining = np.random.exponential(T_on)
                    state = 1

            # the emitter will remain in same state for at least dq time, so we can go to next frame
            remaining -= dt * q
            if state:
                onTime += dt
            onTime *= fps  # scaled so onTime is a fraction off ontime during the frame
            ontimes[i, f] = onTime

    print()
    return ontimes


def generate_movie(
    emitterList, modulation, numframes, T_on, T_off, imgsize=512, intensity=2000, bg=1, sigma=1.5, p_on=0.1
):
    frames = np.zeros((numframes, imgsize, imgsize), dtype=np.uint16)
    emitters = np.array([[e[0], e[1], sigma, sigma, intensity] for e in emitterList])

    on_counts = np.zeros(numframes, dtype=np.int32)

    def excitationPattern(pos, frame):
        m = modulation[frame % len(modulation)]
        return (1 + m[2] * np.sin(m[0] * pos[..., 0] + m[1] * pos[..., 1] - m[3])) / len(modulation)

    ontimes = generate_blink_trace(T_on, T_off, emitters, numframes, excitationPattern, 1)

    print(f"drawing {numframes} frames")
    g = Gaussian(smlm)
    for f in range(numframes):
        frame = bg * np.ones((imgsize, imgsize), dtype=np.float32)
        frame_emitters = emitters * 1
        frame_emitters[:, 4] *= ontimes[:, f] * excitationPattern(emitters, f)

        g.Draw(frame, frame_emitters)
        frames[f] = frame
        on_counts[f] = np.sum(ontimes[:, f] > 0)

    return frames, on_counts


def process_movie_with_queue(q, mov, on_counts, result_image, label):
    scale = 1
    t0 = time.time()
    q.Start()
    for f in range(len(mov)):
        q.PushFrame(f, mov[f])

    while not q.IsIdle():
        q.ReportState()
        time.sleep(1)

    dt = time.time() - t0
    print(f"Processed {len(mov)} frames in {dt} seconds. {len(mov)/dt:.3f} fps")

    gaussian = Gaussian(smlm)
    for f in range(len(mov)):
        fr = q.GetFrameResults(f)

        if f < 20:
            nz = ",".join([str(i) for i, r in enumerate(fr) if r.iterations > 0])

            print(
                "Frame {0} result count: {1}. true on-counts: {2}. where iterations>0 {3}".format(
                    f, len(fr), on_counts[f], nz
                )
            )

        if len(fr) > 0:
            gaussian.Draw(
                result_image,
                [
                    [
                        scale * (r.estimate[0] + r.roiPosition[0]),
                        scale * (r.estimate[1] + r.roiPosition[1]),
                        0.5,
                        0.5,
                        r.estimate[2],
                    ]
                    for r in fr
                ],
            )

    for r in q.GetFrameResults(0):
        est = r.estimate
        roi = r.roiPosition
        print(f"x={est[0]}, y={est[1]}, I={est[2]}, bg={est[3]}. roix={roi[0]}, roiy={roi[1]}")


def unpack64(v):
    r = np.zeros(64, dtype=np.int)
    for i in range(64):
        r[i] = (v & (1 << i)) != 0
    return r


def show_result_rois(frames, results, roisize):

    nspots = len(results)
    rois = np.zeros((nspots, len(frames), roisize, roisize))

    for i, r in enumerate(results):
        for f in range(len(frames)):
            x = r.roiPosition[0]
            y = r.roiPosition[1]
            rois[i, f] = frames[f, y : y + roisize, x : x + roisize]

        #    fig, axes = plt.subplots(1, len(frames), True, True)
        #   fig.suptitle( f"Result {i}. Iterations={r.iterations}. SilmFrames={r.silmFrames}")

        #        axes[i][len(frames)].plot()

        su.imshow_many(rois[i], f"Result {i}. Iterations={r.iterations}. SilmFrames={r.silmFrames}")
        silmFrames = unpack64(r.silmFrames)[: len(frames)]

        plt.figure()
        plt.plot(silmFrames)


psfSigma = 3
roisize = int(2 + (psfSigma + 1) * 2)
w = 256
N = 1000
numframes = 500

# Simulate a ring of emitters
R = np.random.normal(0, 0.001, size=N) + w * 0.3
angle = np.random.uniform(0, 2 * math.pi, N)
emitters = np.vstack((R * np.cos(angle) + w / 2, R * np.sin(angle) + w / 2)).T
omega = 3  # [rad/pixel]


spotDetector = spotdetect.SpotDetectorConfig(psfSigma, roisize, minIntensity=10)
numEP = 6

T_on = 6
T_off = T_on * 10
numFitFrames = 12

silm = smlmlib.silm.SILM(smlm)
silm_p = silm.Params(roisize, numep=numEP, sigma=psfSigma)
hmm_p = smlmlib.silm.BlinkHMM_Params.make(0.1, 0.5, 0.5)
phases = np.tile(np.arange(numEP / 2) * 4 * math.pi / numEP, 2)
mod = silm.phase_to_mod(phases, omega)

mov, on_counts = generate_movie(emitters, mod, numframes, T_on, T_off, imgsize=w, sigma=psfSigma, intensity=5000, bg=1)
mov = np.random.poisson(mov)
tiffpath = "../../../data/test_silm_queue.tiff"
tifffile.imsave(tiffpath, np.array(mov, dtype=np.uint16))

plt.figure()
plt.imshow(np.sum(mov[:6], 0))
plt.title("First 6 frames summed")

if True:
    result_image_g2d = np.zeros((w, w), dtype=np.float32)
    with Gaussian(smlm).CreateImageQueue(w, w, psfSigma, roisize, 500, spotDetector) as q:
        process_movie_with_queue(q, mov, on_counts, result_image_g2d, "Gaussian")
    plt.figure()
    plt.imshow(result_image_g2d)
    plt.title("2D Gaussian Fit Result")

result_image_silm = np.zeros((w, w), dtype=np.float32)
with silm.CreateImageQueue(w, w, numFitFrames, silm_p, hmm_p, 1000, spotDetector, mod) as q:
    q.SetFrameStep(1)  # Only process every 6th frame
    process_movie_with_queue(q, mov, on_counts, result_image_silm, "SILM")

    f = 20
    # show_result_rois(mov[f-numFitFrames+1:f+1], q.GetFrameResults(f), roisize)


plt.figure()
plt.imshow(result_image_silm)
plt.title("SILM Fit Result")
