# -*- coding: utf-8 -*-
import os,sys
sys.path.append("../..")


import numpy as np
import utils.localizations as loc
from smlmlib.gaussian import Gaussian
from smlmlib.silm import SILM
from smlmlib.base import SMLM
import tqdm
import tifffile
import matplotlib.pyplot as plt
import utils.caching as cache
import sys
import utils.findpeak1D as findpeak1D
import util.config as silm_config
from smlmlib import util as su
import pickle
import os
import time
import smlmlib.spotdetect as spotdetect


# Reads all results from the queue object
def fetch_from_queue(q):
    nframes = q.GetResultFrameCount()

    result_list = []
    frames = []
    spot_count = 0
    for f in range(nframes):
        fr = q.GetFrameResults(f)
        if f % 5 == 0:
            sys.stdout.write(f"\rreading results {f/nframes*100:3.0f}% {f}/{nframes} (#spots={spot_count})")
            sys.stdout.flush()

        if len(fr) > 0:
            result = np.array( [
                    [
                        f,
                        r.initialValue[0] + r.roiPosition[0],
                        r.initialValue[1] + r.roiPosition[1],
                        r.initialValue[2],
                        r.initialValue[3],
                        r.estimate[0] + r.roiPosition[0],
                        r.estimate[1] + r.roiPosition[1],
                        r.estimate[2],
                        r.estimate[3],
                        r.score,
                        *r.roiPosition,
                    ] for r in fr
                ],
                dtype=np.float32,
            )
        else:
            result = np.zeros((0, 12), dtype=np.float32)

        result_list.append(result)
        frames.append(np.arange(len(result)) + spot_count)
        spot_count += len(result)
    
    results = np.concatenate(result_list)
    return results,frames    


def process_images(cfg, image_iterator, totalframes, smlm):
    img0 = next(image_iterator)

    width = img0.shape[1]
    height = img0.shape[0]

    spotDetectConfig = spotdetect.SpotDetectorConfig(
        cfg["sigma"], cfg["roisize"], cfg["detectionMinIntensity"], cfg["detectionMaxIntensity"]
    )
    
    mod = cfg['mod']
        
    silm_p = SILM(smlm).Params(
        cfg["roisize"], len(mod), sigma=cfg["sigma"], startLambdaStep=1
    )

    with SILM(smlm).SILM_CreateImageQueue(
            width,height,len(mod),silm_p,cfg['maxSpotsPerFrame'],cfg['silmMinIntensity'],
            spotDetectConfig, mod) as q:
        q.Start()

        raw_image_sum = np.zeros((height, width), dtype=np.float32)
        t0 = time.time()

        q.PushFrame(0, img0)
        numframes = 1

        for index, img in enumerate(image_iterator):
            f = index + 1

            raw_image_sum += img
            while q.GetQueueLength() > 50:
                time.sleep(0.05)

            q.PushFrame(f, img)
            if numframes > 1 and numframes % 20 == 0:
                t = time.time()
                fps = numframes / (t - t0)
                remaining = (totalframes - numframes) / fps
                sys.stdout.write(
                    f"\rprocessed {numframes}/{totalframes} frames. Queued: {q.GetQueueLength()}. Time remaining: {remaining:4.0f}s"
                )
                sys.stdout.flush()
            numframes += 1

        while not q.IsIdle():
            print(f'Frames processed: {q.GetResultFrameCount()}/{numframes}')
            time.sleep(0.5)

        dt = time.time() - t0
        print(f"\r{numframes} frames in {dt:3.0f} seconds. {numframes/dt:.2f} fps\t\t\t\t\t\t\t\t\t")

        silm_results,g2d_results,frames=fetch_from_queue(q)
        
        return silm_results,g2d_results,frames, [width,height]


def read_tiff(filename, cfg, smlm):

    with tifffile.TiffFile(filename) as tif:
        p0 = tif.pages[0]
        height = p0.shape[0]
        width = p0.shape[1]
        numframes = len(tif.pages)
        print(f"{filename}: width: {width}, height: {height}")

        def generator():
            for f, page in enumerate(tif.pages):
                if "maxframes" in cfg and f == cfg["maxframes"]:
                    raise StopIteration()

                img = (page.asarray().astype(np.float) - cfg["offset"]) / cfg["gain"]
                img[img<0] = 0
                yield img

        return process_images(cfg, generator(), numframes, smlm)


def process_tiff(tiffpath, cfg, mod=None):
    if mod is None:
        with open(os.path.splitext(tiffpath)[0]+"_mod.pickle", "rb") as pf:
            mod = pickle.load(pf)['mod']
    
    cfg = {**cfg, 'mod': mod}
    
    with SMLM(debugMode=True) as smlm:
        return read_tiff(tiffpath, cfg, smlm)
    
if __name__ == '__main__':
    fn='../../../data/80nm/80nm-3rdshift-2.tif'
    
    
    # Configuration for spot detection and 2D Gaussian fitting
    cfg = { 'sigma':1.8304,
       'roisize':12,
       'maxSpotsPerFrame':2000,
       'detectionMinIntensity':50,
       'detectionMaxIntensity':2000,
       'offset':100,
       'gain':1/0.47,
       'silmMinIntensity': 30
       }
    
    process_tiff(fn, cfg)