# -*- coding: utf-8 -*-

import numpy as np
import utils.localizations as loc
from smlmlib.gaussian import Gaussian
from smlmlib.silm import SILM
from smlmlib.base import SMLM
import tqdm
import tifffile
import matplotlib.pyplot as plt
import utils.caching as cache
import sys
import utils.findpeak1D as findpeak1D
import util.config as silm_config
from smlmlib import util as su
import utils.dftpeak as dftpeak
import utils.fitpoints2D as fitpoints2D

def extract_rois(img, roipos, roisize):
    result = np.zeros((len(roipos), roisize, roisize))
    for i, pos in enumerate(roipos):
        roi = img[pos[1] : pos[1] + roisize, pos[0] : pos[0] + roisize]
        result[i] = roi
    mean = np.mean(result)
    std = np.std(result)
    result[np.abs(result - mean) > 15 * std] = mean
    return result


def roi_iterator(r: loc.LocResultList, tiffpath, num_patterns):
    with tifffile.TiffFile(tiffpath) as tif:
        roisize = r.cfg["roisize"]
        roipos = r.get_roipos()

        for f, indices in enumerate(r.frames):
            rois = np.zeros((len(indices), num_patterns, roisize, roisize))
            for e in range(num_patterns):
                pageIndex = f * num_patterns + e
                raw = tif.pages[pageIndex].asarray().astype(np.float)
                img = (raw - r.cfg["offset"]) / r.cfg["gain"]
                img[img < 0] = 0
                rois[:, e] = extract_rois(img, roipos[indices], roisize)
            yield rois


def polyfit_ransac(x,y,deg,filterfrac=0.5,iterations=3):
    indices = np.arange(len(x))    

    for k in range(iterations):    
        c = np.polyfit(x[indices],y[indices],deg)
#        print(f"{k}: {c}")
        if k == iterations-1:
            break
        
        errs = (np.polyval(c, x) - y)**2
        sorted_errs = np.argsort(errs)
        indices = sorted_errs[:int(len(sorted_errs)*filterfrac)]

    return c


def _estimate_ibg(r: loc.LocResultList, tiffpath, smlm, num_patterns, store_rois):
    roisize = r.cfg["roisize"]
    IBg = np.zeros((len(r.data), num_patterns, 2), dtype=np.float32)
    IBg_crlb = np.zeros((len(r.data), num_patterns, 2), dtype=np.float32)

    if store_rois:
        all_rois = np.zeros((len(r.data), num_patterns, roisize, roisize), dtype=np.uint16)
    else:
        all_rois = None

    xy = r.get_xyI()[:, 0:2]
    roipos = r.get_roipos()
    g = Gaussian(smlm)
    #    with tqdm.tqdm(total=len(r)) as pbar:
    for f, rois in enumerate(roi_iterator(r, tiffpath, num_patterns)):
        indices = r.frames[f]
        pts = xy[indices] - roipos[indices]
        pts_tiled = np.repeat(pts, num_patterns, 0)
        IBg_, IBg_crlb_ = g.EstimateIBg(
            rois.reshape((num_patterns * len(pts), roisize, roisize)), r.cfg["sigma"], pts_tiled, useCuda=True
        )

        IBg[indices] = IBg_.reshape((len(indices), num_patterns, 2))
        IBg_crlb[indices] = IBg_crlb_.reshape((len(indices), num_patterns, 2))
        if store_rois:
            all_rois[indices] = rois

        sys.stdout.write(f"\r{f}/{len(r)}.  #spots={len(indices)}")
    #            pbar.n = f
    #            pbar.refresh()
    return (IBg, IBg_crlb, all_rois)


class SpotList:
    IBg = None
    IBg_crlb = None
    num_patterns = 0
    result = None  # Type: loc.LocResultList
    outdir = ""  # output directory for plots etc
    rois = None

    def __init__(self, r: loc.LocResultList, pixelsize, outdir, IBg=None, IBg_crlb=None, rois=None):
        self.result = r
        self.pixelsize = pixelsize
        self.outdir = outdir
        self.rois = rois
        self.IBg = IBg
        self.IBg_crlb = IBg_crlb

    def estimate_IBg(self, tiffpath, smlm, num_patterns, store_rois=False, rebuild=False):
        tag = "ibg_rois" if store_rois else "ibg"
        self.IBg, self.IBg_crlb, self.rois = cache.read_npy(tiffpath, tag,
            (self.result.cfg, store_rois),
            lambda path, cfg: _estimate_ibg(self.result, tiffpath, smlm, num_patterns, store_rois),
            rebuild=rebuild
        )
        self.num_patterns = num_patterns

    def get_xyI(self, ep):
        xyI = self.result.get_xyI() * 1
        xyI[:, 2] = self.IBg[:, ep, 0]
        return xyI

    def generate_projections(self, mod, scale, smlm):
        w = self.result.imgshape[0]
        self.proj = np.zeros((len(mod), w * scale))
        self.proj_shift = np.zeros(len(mod))
        for k in range(len(mod)):
            m = mod[k]
            ang = np.arctan2(m[1], m[0])
            xyI = self.get_xyI(k)
            xyI = xyI[~np.isnan(xyI[:, 2]), :]
            projected, shift = SILM(smlm).ProjectPoints(xyI, w * scale, scale, [ang])
            self.proj[k] = projected[0]
            self.proj_shift[k] = shift

    def plot_proj_fft(self):
        from matplotlib.backends.backend_pdf import PdfPages

        pp = PdfPages(f"{self.outdir}fft-spectra.pdf")
        for k in range(self.num_patterns):
            fig = plt.figure()
            scale = len(self.proj[k]) / self.result.imgshape[0]
            freq = np.fft.fftshift(np.fft.fftfreq(len(self.proj[k])))
            plt.plot(freq * scale, np.fft.fftshift(np.abs(np.fft.fft(self.proj[k] ** 2))))
            plt.title(f"FFT for pattern {k}")
            # plt.xlabel("pixel^-1");
            plt.tight_layout()
            pp.savefig(fig)
            plt.close(fig)
        pp.close()

    def _estimate_angle_and_pitch_dft(self, pattern_indices, frames, smlm:SMLM, freq_minmax):
        w,h=self.result.get_image_size()

        z=6
        ft_sum = np.zeros((h*z,w*z))
        for ep in pattern_indices:
            img = self.render_sr_image_per_pattern(ep, frames, sr_factor=z,smlm=smlm)
            ft_img = smlm.FFT2(img)
            ft_sum += np.abs(ft_img**2)
        ft_sum = np.fft.fftshift(ft_sum)
        
        freq = np.fft.fftshift( np.fft.fftfreq(h*z) )*z*2*np.pi
        XFreq, YFreq = np.meshgrid(freq,freq)
        Freq = np.sqrt(XFreq**2+YFreq**2)
        
        mask = (Freq>freq_minmax[0]) & (Freq<freq_minmax[1])
#        plt.imsave(self.outdir + f"pattern-{pattern_indices}-FFT-mask.png", mask)
        
        ft_sum[~mask] = 0
        plt.imsave(self.outdir + f"pattern-{pattern_indices}-FFT-ep{pattern_indices}.png", ft_sum)
        
        max_index = np.argmax(ft_sum)
        max_indices = np.unravel_index(max_index, ft_sum.shape)
                
        yx = freq[list(max_indices)]
        xy = self._find_dft2_peak(yx[[1,0]], pattern_indices, frames, smlm)
                
        ang = np.arctan2(xy[1],xy[0])
        freq = np.sqrt(xy[0]**2+xy[1]**2)
        return ang,2*np.pi/freq
        
    def _find_dft2_peak(self, xy, pattern_indices, frames, smlm:SMLM):
        indices = np.concatenate([self.result.frames[i] for i in frames])
        S = 0.02
        kxrange = np.linspace(xy[0]-S, xy[0]+S, 50)
        kyrange = np.linspace(xy[1]-S, xy[1]+S, 50)

        img = np.zeros((len(kyrange),len(kxrange)))
        for ep in pattern_indices:
            xyI = self.get_xyI(ep)[indices]
            sig = dftpeak.dft_points(xyI, kxrange, kyrange, smlm, useCuda=True)
            img += np.abs(sig**2)
            
#        plt.figure()
 #       plt.imshow(img)
        plt.imsave(self.outdir + f"pattern-{pattern_indices}-FFT-peak.png", img)
        
        peak = np.argmax(img)
        peak = np.unravel_index(peak, img.shape)

        kx_peak = findpeak1D.quadraticpeak(img[peak[0], :], kxrange, npts=11, plotTitle=None)#='X peak')
        ky_peak = findpeak1D.quadraticpeak(img[:, peak[1]], kyrange, npts=11, plotTitle=None)#='Y peak')

        return kx_peak, ky_peak

    def estimate_angle_and_pitch(self, num_patterns, frame_bins, smlm, plotTag=None,freq_minmax=[1.5,2]):
        angle = np.zeros((len(frame_bins), num_patterns))
        pitch = np.zeros((len(frame_bins), num_patterns))
        for b in range(len(frame_bins)):
            sys.stdout.write(f"Estimating angle/pitch {b}/{len(frame_bins)}")
            for m in tqdm.trange(num_patterns):
                angle[b, m], pitch[b, m] = self._estimate_angle_and_pitch_dft(
                    [m], frame_bins[b], smlm,
                    freq_minmax=freq_minmax
                )
            sys.stdout.write("\n")
        return angle, pitch
    
    def _estimate_angle_and_pitch(self, ep, frames, smlm, plotTag,freq_minmax):
        indices = np.concatenate([self.result.frames[i] for i in frames])
        xyI = self.get_xyI(ep)[indices]
        w = self.result.get_area()[2]

        ang, freq, output = self.pattern_angle_and_freq(
            xyI, w, angle_range=np.linspace(0, np.pi, 300), freq_range=np.linspace(freq_minmax[0], freq_minmax[1], 300), smlm=smlm
        )

        freq_range = np.linspace(freq * 0.95, freq * 1.05, 200)
        ang_range = np.linspace(ang - 0.02, ang + 0.02, 200)
        deg = np.rad2deg(ang_range)
        ang2, freq2, output = self.pattern_angle_and_freq(
            xyI, w, ang_range, freq_range, smlm, useFit=True, plotTag=plotTag
        )

        if False:
            plt.figure()
            plt.imshow(output, extent=[freq_range[0], freq_range[-1], deg[-1], deg[0]], aspect="auto")
            plt.title(f"corr ep{ep}. Angle={ang*180/np.pi:.4f} deg, pitch={2*np.pi/freq} pixels")
        return ang2, 2 * np.pi / freq2

    def pattern_angle_and_freq(self, xyI, w, angle_range, freq_range, smlm, useFit=False, plotTag=None):
        silm = SILM(smlm)
        projWidth = w * 11
        scaling = 10

        projpts, _ = silm.ProjectPoints(xyI, projWidth, scaling, angle_range)
        X, F = np.meshgrid(range(projWidth), freq_range)
        sin = np.sin(X * F / scaling)
        cos = np.cos(X * F / scaling)

        # Autocorrelation
        output = np.abs(silm.SignalCorrelation(projpts, sin) ** 2 + silm.SignalCorrelation(projpts, cos) ** 2)
        #output = np.sqrt(silm.SignalCorrelation(projpts, sin) ** 2 + silm.SignalCorrelation(projpts, cos) ** 2)
        max_index = np.argmax(output)
        max_indices = np.unravel_index(max_index, output.shape)

        freq = freq_range[max_indices[1]]
        ang = angle_range[max_indices[0]]

        if useFit:
            ang = findpeak1D.quadraticpeak(output[:, max_indices[1]], angle_range, npts=20)  # , plotTitle="angle")
            freq = findpeak1D.quadraticpeak(output[max_indices[0], :], freq_range, npts=10)  # ,plotTitle="freq")

            fig, axes = plt.subplots(1, 2)
            axes[0].plot(np.rad2deg(angle_range), output[:, max_indices[1]])
            axes[0].set_xlabel(f"Angle [degrees]. Found angle: {np.rad2deg(ang):.4f} deg")
            axes[1].plot(freq_range, output[max_indices[0], :])
            axes[1].set_xlabel(f"Frequency [1/pixel]. Found pitch={2*np.pi/freq*self.pixelsize:.5f} nm")
            plt.suptitle(f"{plotTag} angle and freq. fits")
            fig.savefig(self.outdir + f"{plotTag}-angle-freq-fit.png")
            plt.close(fig)

        return ang, freq, output

    def estimate_phase_and_depth(self, mod_list, frame_bins, method=0):
        phase = np.zeros((len(frame_bins), len(mod_list)))
        depth = np.zeros((len(frame_bins), len(mod_list)))
        power = np.zeros((len(frame_bins), len(mod_list)))
        for b, fr_bin in enumerate(frame_bins):
            for m, mod in enumerate(mod_list):
                if method == 0:
                    phase[b, m], depth[b, m], power[b, m] = self._estimate_phase_and_depth(mod, m, fr_bin)
                else:
                    phase[b, m], depth[b, m], power[b, m] = self._estimate_phase_and_depth2(mod, m, fr_bin)

            #print(power[b,:])                    
            power[b] /= np.sum(power[b,:])
        return phase, depth, power

    # Re-estimate the phase and depth of a single modulation pattern
    def _estimate_phase_and_depth(self, mod, ep, frames):
        indices = np.concatenate([self.result.frames[i] for i in frames])
        xyI = self.get_xyI(ep)[indices]

        spotPhaseField = mod[0] * xyI[:, 0] + mod[1] * xyI[:, 1]

        I0 = np.sum(xyI[:, 2])
        Iw = np.sum(xyI[:, 2] * np.exp(-1j * spotPhaseField))

        phase = -np.angle(Iw / I0 * 1j)
        depth = 2 * np.abs(Iw / I0)
        power = np.mean(xyI[:,2] / self.result.get_xyI()[indices,2])
        return phase, depth, power

    def _estimate_phase_and_depth2(self, mod, ep, frames):
        indices = np.concatenate([self.result.frames[i] for i in frames])
        sumI = np.sum(self.IBg[indices, :, 0],1)

        xyI = self.result.get_xyI()[indices] * 1
        
        xyI_modulated = xyI*1
        xyI_modulated[:,2] = self.IBg[indices,ep,0]

        I = xyI[:,2]*1
        spotPhaseField = mod[0] * xyI[:, 0] + mod[1] * xyI[:, 1]
        
        basefreq = np.linspace(-1, 1, 3)
        
        weights = np.ones(len(I))

        for it in range(6):

            # DFT on modulated and unmodulated data
            f = [ np.sum(weights * xyI_modulated[:,2] * np.exp(-1j * spotPhaseField * k)) for k in basefreq ]
            B = [ np.sum(weights * I * np.exp(-1j * spotPhaseField * k)) for k in basefreq - 1]
            A = [ np.sum(weights * I * np.exp(-1j * spotPhaseField * k)) for k in basefreq ]
            C = [ np.sum(weights * I * np.exp(-1j * spotPhaseField * k)) for k in basefreq + 1]
            
            # Solve Ax = b
            nrows = len(basefreq)
            M = np.zeros((nrows,3), dtype=np.complex)
            b = np.zeros(nrows, dtype=np.complex)
            for i in range(nrows):
                M[i] = [ B[i], C[i], A[i] ] 
#                M[i] = [ B[i], C[i], A[i] ] 
                b[i] = f[i]
                            
            # Actually x[1] is just the complex conjugate of x[0], 
            # so it seems a solution with just 2 degrees of freedom is also possible
#            x, residual, rank, s = np.linalg.lstsq(M,b,rcond=None)
            x = np.linalg.solve(M,b)
            b,c,a = x
           
            depth = np.real(2*np.abs(b)/a)
            phase = - np.angle(b*1j)
            power = np.real(2*a)/2  # This is the power in 1D sine case

            q = power * (1+depth*np.sin(mod[0]*xyI[:,0]+mod[1]*xyI[:,1]-phase))            
            xyQ = xyI_modulated*1
            sumI = np.sum(self.IBg[indices, :, 0],1)
            xyQ[:,2] /= sumI

            errs = np.abs(xyQ[:,2]-q)
            
           # print(f"{it}: err sum: {np.sum(errs)}. phase={phase}")
 #           plt.figure()
#            plt.hist(errs)
            median_err = np.percentile(errs,50)# median(errs)
            weights = errs < median_err
            

#        print(f"ep{ep}. depth={depth}, phase={phase}")
        return phase, depth, power


    def silm(self,mod,smlm:SMLM, min_intensity=50, draw_silm_compare=False):
        sel = np.nonzero((self.IBg[:,0,0] > min_intensity ) & 
                         (self.IBg[:,-1,0] > min_intensity )) [0]
       
#        print(f"{len(sel)} spots for silm... ({len(self.IBg)} unfiltered).  ")
 #       lowest_power = np.min(mod[:,4])
#        depth = mod[np.argmin(mod[:,4],2]
#        if min_intensity / (med_I*lowest_power) > 1-depth:
            

        rois = self.rois[sel]

        silm_p = SILM(smlm).Params(
            self.result.cfg["roisize"], len(mod), sigma=self.result.cfg["sigma"], startLambdaStep=1
        )

        roipos = self.result.get_roipos()[sel]
        startPatterns = np.zeros(len(sel))
        mod_n = np.tile(mod, (len(rois), 1, 1))
        est, results = SILM(smlm).SILM_ASW_ComputeMLE(rois, mod_n, silm_p, startPatterns, cuda=True, roipos=roipos)

        crlb = np.array([r.crlb for r in results])
        iterations = np.array([r.iterations for r in results])
        ll = np.array([r.loglikelihood for r in results])

        data = self.result.data[sel, :] * 1
        data[:, [loc.DataColumns.X, loc.DataColumns.Y]] = est[:, [0, 1]] + roipos
        data[:, [loc.DataColumns.I, loc.DataColumns.BG]] = est[:, [2, 3]]
        data[:, [loc.DataColumns.CRLB_X, loc.DataColumns.CRLB_Y, loc.DataColumns.CRLB_I, loc.DataColumns.CRLB_BG]] = crlb
        data[:, loc.DataColumns.ITERATIONS] = iterations
        data[:, loc.DataColumns.LL] = ll 

        extraColumns = { 
                'intensities': self.IBg[sel,:,0].astype(np.float32),
                'backgrounds': self.IBg[sel,:,1].astype(np.float32)
                }
                
        if draw_silm_compare:
            scfg = silm_config.from_mod(self.result.cfg,mod)
            import matplotlib.ticker as ticker
            # matplotlib gets really slow with large point counts
            draw_sel = sel[:50000 if len(sel)>50000 else len(sel)]*1
            gxy = self.result.get_xyI()[draw_sel]
            fig,axes=plt.subplots(2,len(mod)//2,sharey=True,sharex=True)
            for k in range(self.num_patterns):
                ax = axes[k%2,k//2]
                # The gaussian localizations projected onto the modulation axis
                k_len = np.sqrt(mod[k,0]**2+mod[k,1]**2)
                g2d_c = mod[k,0] * gxy[:,0] + mod[k,1] * gxy[:,1]
                # The SILM localizations projected onto the modulation axis
                silm_c = mod[k,0] * data[:len(draw_sel),loc.DataColumns.X] + mod[k,1] * data[:len(draw_sel),loc.DataColumns.Y]
                phases = ( (g2d_c - mod[k,3]) % (2*np.pi) ) / np.pi
                ax.scatter( (g2d_c - silm_c) / k_len, phases,s=0.5,marker='.')
                ax.set_xlim(-2,2)
                ax.set_title(f'[{k}]: Angle={scfg.angle[k]:.3f}')
                if k%2!=0: ax.set_xlabel('[pixels]'); 
                ax.set_ylabel(f'Phase in pattern {k}')
                ax.yaxis.set_major_formatter(ticker.FormatStrFormatter('%g $\pi$'))
                ax.yaxis.set_major_locator(ticker.MultipleLocator(base=1.0))
            fig.suptitle("Difference between Gaussian and SILM localization mapped to position in pattern")
            fig.savefig(self.outdir + "phase-bias.png")
            plt.close(fig)
        
        return (loc.LocResultList(data,self.result.cfg, self.result.get_area(), extraColumns=extraColumns),
                loc.LocResultList(self.result.data[sel,:]*1,self.result.cfg, self.result.get_area(), extraColumns=extraColumns))
    
    
    
    def draw_spot_errors(self, mod, smlm:SMLM):
        sumI = np.sum(self.IBg[:, :, 0],1)
                
        coeff = np.zeros(6)
                
        coeff_k = np.zeros((len(mod),6))
        center = self.result.get_image_size()*.5
        
        silm = SILM(smlm)
        for k in range(len(mod)):        
            xyQ = self.get_xyI(k)
            xyQ[:,2] /= sumI

            if False:
#                1+m*cos( 2*pi* (1-eps*(x^2+y^2)/R^2) *x/p + psi )
                last_err=0
                for i in range(2000):
                    err,err_deriv,spot_err=silm.PhaseFitErrorDeriv(xyQ, mod[k], coeff)
                    
                    if i>0 and np.abs(last_err-err)<1e-2:
                        break
                    last_err = err
                    if i % 50 == 0:
                        print(f"{i}: err={err[0]}, coeff={coeff}")
                    coeff = coeff - err_deriv*1e-14

            err,err_deriv,spot_err=silm.PhaseFitErrorDeriv(xyQ, mod[k], coeff)

            pts = xyQ*1
            pts[:,2] = spot_err
            
            sel = np.argsort(spot_err)
            pts = pts[sel]
            
            r = np.sqrt( pts[:,0]**2+pts[:,1]**2 )
            
            plt.figure()
            numspots = 2000
            plt.scatter(r[:numspots], spot_err[:numspots], marker='.')
            plt.ylim(0,0.2)
            plt.title(f"Spot errors pattern {k}")
            
            plt.figure()
            x = np.arange(0,self.result.get_image_size()[0])
            plt.plot(x,coeff[0]*x**2+coeff[3]*x+coeff[5])
            plt.title(f"Pattern {k} X coeff")
            coeff_k[k] = coeff
            
            plt.figure()
            plt.hist(spot_err,30)
        return coeff_k


    def refine_pitch(self, mod, smlm:SMLM, silm_min_intensity, plot=False):
        mod=mod*1
        
        # Get the rows in mod for which the given axis is the dominant axis, get_axis_mod(mod,0) gives all X modulations
        def get_axis_mod(mod, axis):
            dominant_axis = np.argmax(np.abs(mod[:,[0,1]]),1)
            return np.nonzero(dominant_axis==axis)[0]

        xmod = get_axis_mod(mod,0)
        ymod = get_axis_mod(mod,1)
        
        kx = np.mean(mod[xmod,0])
        ky = np.mean(mod[ymod,1])
        
        def estim_shift(mod,plot=False):
            silm_d,g2d_d = self.silm(mod,smlm,min_intensity=silm_min_intensity)
            shift = silm_d.get_xyI()[:,[0,1]] - g2d_d.get_xyI()[:,[0,1]]
            pos=g2d_d.get_xyI()
            cx = np.polyfit(pos[:,0],shift[:,0],1)
            cy = np.polyfit(pos[:,1],shift[:,1],1)
#            cx = polyfit_ransac(pos[:,0],shift[:,0],1,0.4)
#           cy = polyfit_ransac(pos[:,1],shift[:,1],1,0.4)
 
            if plot:
                # plot a random subset of the points
                np.random.seed(0)
                indices = np.arange(len(pos))# np.nonzero(np.min( self.IBg[:,:,0],1) > min_intensity)[0]
                np.random.shuffle(indices)
                indices = indices[:2000]
                                
                w,h=self.result.get_image_size()
                fig,axes=plt.subplots(2,1)
                px = np.polyfit(pos[indices,0],shift[indices,0],1)
                py = np.polyfit(pos[indices,1],shift[indices,1],1)

                sx=self.pixelsize/1000
                sy=self.pixelsize
                
                axes[0].scatter(sx*pos[indices,0],sy*shift[indices,0])
                axes[0].plot(np.linspace(0,w)*sx,np.polyval(px,np.linspace(0,w))*sy, 'g-')
                axes[0].set_title(f'X slope ({cx[0]*kx:.5f} rad/pixel), offset ({cx[1]:.3f})')

                axes[0].set_title('Difference between SIMFLUX and SMLM in X')
                axes[0].set_xlabel('X position in FOV [um]')
                axes[0].set_ylabel('Difference [nm]')

#                axes[0].set_ylim(-50,50)
                axes[1].scatter(sx*pos[indices,1],sy*shift[indices,1])
                axes[1].plot(np.linspace(0,h)*sx,np.polyval(py,np.linspace(0,h))*sy,'g-')
                axes[1].set_title(f'Y shift ({cy[0]*ky:.5f} rad/pixel), offset ({cy[1]:.3f}))')
                axes[1].set_title('Difference between SIMFLUX and SMLM in Y')
                axes[1].set_xlabel('Y position in FOV [um]')
                axes[1].set_ylabel('Difference [nm]')
#               axes[1].set_ylim(-50,50)
                fig.tight_layout()
                fig.savefig(self.outdir+"fov-shift.png")
              #  fig.savefig(self.outdir+"fov-shift.svg")
#                plt.close(fig)
            
            return silm_d,g2d_d,cx,cy
        
        for k in range(5):            
            silm_d,g2d_d,cx,cy=estim_shift(mod,plot and k==0)
            # cx[0] is shifted pixel / pixel
            # cx[0]*kx shifted radians / pixel, iaw frequency change

            # Adjust pitch
            print(f"Adjust kx: {cx[0]*kx:.4f}, ky: {cy[0]*ky:.4f}. phase shift x: {cx[1]*kx:.4f}, phase shift y: {cy[1]*ky:.4f}")
            mod[xmod,0] += cx[0]*kx
            mod[ymod,1] += cy[0]*ky

            mod[xmod,3] -= cx[1]*kx
            mod[ymod,3] -= cy[1]*ky
            
        return silm_d,g2d_d,mod

    def silm_bias_plot2D(self, mod, smlm:SMLM, silm_min_intensity, tag):
        
        silm_d,g2d_d = self.silm(mod,smlm,min_intensity=silm_min_intensity)
        shift = silm_d.get_xyI()[:,[0,1]] - g2d_d.get_xyI()[:,[0,1]]
        pos = g2d_d.get_xyI()

        rms=np.sqrt(np.mean(shift**2,0))
        print(f"RMS x shift={rms[0]*self.pixelsize}, y shift={rms[1]*self.pixelsize}")

        shape = self.result.get_image_size()
        w = 100

        print(f"running 2D bias fit over {len(pos)} points")

        gridshape = [w,w]        

        plt.set_cmap('viridis')
#        statfn = 'median'
        
        def median_mincount(x, minsize):
            if len(x) < minsize:
                return np.nan
            return np.median(x)
        
        def rms_mincount(x,minsize):
            if len(x)<minsize:
                return np.nan
            return np.sqrt(np.mean(x**2))
        
        statfn = lambda x: median_mincount(x,20)
        metric='Median bias'
 #       statfn = lambda x: rms_mincount(x,20)
#        metric='RMS'
        scatter=True
         
        clustergrid=[w,w]
        xysx = fitpoints2D.hist2d_statistic_points(pos[:,0],pos[:,1],shift[:,0], shape, clustergrid, statistic=statfn)
        xysy = fitpoints2D.hist2d_statistic_points(pos[:,0],pos[:,1],shift[:,1], shape, clustergrid, statistic=statfn)
        
        print(f"2D bias plot: plotting {len(xysx)} clusters")
        
        xysx[:,2] = np.clip(xysx[:,2],-0.25,0.25)
        xysy[:,2] = np.clip(xysy[:,2],-0.25,0.25)
                
        img_x,coeff_x = fitpoints2D.fitlsq(xysx[:,0], xysy[:,1], xysx[:,2], shape, gridshape)
        img_y,coeff_y = fitpoints2D.fitlsq(xysy[:,0], xysy[:,1], xysy[:,2], shape, gridshape)

        plt.figure()
#        plt.imshow(img_x, origin='lower',extent=[0,shape[1],0,shape[0]])
        s=self.pixelsize/1000
        sc=self.pixelsize
        if scatter: plt.scatter(s*xysx [:,0],s*xysx [:,1],c=sc*xysx [:,2])
        cb=plt.colorbar()
        cb.set_label('Bias in X [nm]')
        plt.xlabel(u'X position in FOV [um]')
        plt.ylabel(u'Y position in FOV [um]')
        plt.title(f'{metric} in X over FOV')
        plt.savefig(self.outdir + tag+"x-bias-fov.png" )
      #  plt.savefig(self.outdir + tag+"x-bias-fov.svg" )
        
        plt.figure()
 #       plt.imshow(img_y, origin='lower',  extent=[0,shape[1],0,shape[0]])
        if scatter: plt.scatter(s*xysy[:,0],s*xysy[:,1],c=sc*xysy[:,2])
        cb=plt.colorbar()
        cb.set_label('Bias in Y [nm]')
        plt.xlabel(u'X position in FOV [um]')
        plt.ylabel(u'Y position in FOV [um]')
        plt.title(f'{metric} in Y over FOV')
        plt.savefig(self.outdir + tag+"y-bias-fov.png" )
      #  plt.savefig(self.outdir + tag+"y-bias-fov.svg" )
        

    
    def render_sr_image_per_pattern(self, ep, frames, sr_factor, smlm:SMLM):
        indices = np.concatenate([self.result.frames[i] for i in frames])
        xyI = self.get_xyI(ep)[indices]
        w,h=self.result.get_image_size()
        img = np.zeros((h*sr_factor,w*sr_factor))
        img = loc.draw_spots(img,xyI[:,0],xyI[:,1],xyI[:,2],1/sr_factor,1/sr_factor,sr_factor,smlm)
        return img
            
    def draw_spots_in_pattern(self, filename, allmod, ep, tiffname, numpts=2000, min_intensity=30):

        med_I = np.median(np.sum(self.IBg[:,:,0],1))
        min_intensity_line_y = min_intensity / med_I
        
        np.random.seed(0)
        indices = np.arange(len(self.IBg))# np.nonzero(np.min( self.IBg[:,:,0],1) > min_intensity)[0]
        np.random.shuffle(indices)
        
        sel = indices[:np.minimum(numpts,len(indices))]
        
        mod = allmod[ep]
        xyI = self.get_xyI(ep)[sel]
        sumI = np.sum(self.IBg[sel, :, 0],1)
        
        # normalize
        normI = xyI[:,2] / sumI
        
        k = np.sqrt(mod[0]**2+mod[1]**2)
        proj = xyI[:,0]*mod[0]+xyI[:,1]*mod[1] 
        x = proj % (np.pi*2) 
        sigx = np.linspace(0,2*np.pi,400)

        center = self.result.get_image_size()/2
        r = ( np.abs(proj - (center[0]*mod[0]+center[1]*mod[1])) ) / np.linalg.norm(mod[0:2])
        
        fig=plt.figure()
        plt.scatter(x, normI,c=r,cmap='viridis', marker='.')
        plt.colorbar()
                    
        plt.plot(sigx, mod[4]*(1+mod[2]*np.sin(sigx-mod[3])), 'r', linewidth=4, label='Estimated Q(x)')
        
       # plt.plot(sigx, np.ones(len(sigx))*min_intensity_line_y, label='Min intensity filter cutoff')
        
        plt.xlabel('Phase [radians]')
        plt.ylabel(r'Normalized intensity ($I_k$)')
        plt.title(tiffname + f': Pattern {ep}. K={k:.4f} ' + f" Phase={allmod[ep,3]*180/np.pi:.3f})")
        plt.tight_layout()
        plt.savefig(filename)
        plt.close(fig)

    def draw_spots_in_patterns(self, title, allmod, pattern_indices, numpts=2000):
        np.random.seed(0)
        indices = np.arange(len(self.IBg))# np.nonzero(np.min( self.IBg[:,:,0],1) > min_intensity)[0]
        np.random.shuffle(indices)   # this gets rid of blobs
                
        fig=plt.figure(figsize=(5,4))
        
        for i,ep in enumerate(pattern_indices):
            mod = allmod[ep]
            xyI = self.get_xyI(ep)[indices]
            sumI = np.sum(self.IBg[indices, :, 0],1)
            
            # normalize
            normI = xyI[:,2] / sumI
            proj = xyI[:,0]*mod[0]+xyI[:,1]*mod[1]
            x = proj % (np.pi*2)
            sigx = np.linspace(0,2*np.pi,400)

            q_expected=mod[4]*(1+mod[2]*np.sin(proj - mod[3]))
            err = ( q_expected-normI ) ** 2
            sel = np.arange(len(err))[:numpts]#np.nonzero(err<np.percentile(err, 80))[0]#[:numpts]
            
            s=1/(2*np.pi)
            plt.scatter(x[sel]*s, normI[sel], marker='.', s=10)#, c=['b','r','c','y','g'][i])
            plt.plot(sigx*s, mod[4]*(1+mod[2]*np.sin(sigx-mod[3])), '-', linewidth=6, label=f'Pattern {ep}')
        
        plt.ylim([0,0.4])

        plt.xlabel('Position in pattern [0-1]')
        plt.ylabel(f'Excitation power')
        plt.title(title)
#        plt.legend()
#        plt.tight_layout()
        fn=f"combined-phases-{pattern_indices}"
        plt.savefig(self.outdir + fn+".svg")
        plt.savefig(self.outdir + fn+".png")

        plt.close(fig)
