# -*- coding: utf-8 -*-

import ctypes
import os
import math
import numpy as np
import numpy.ctypeslib as ctl
import matplotlib.pyplot as plt
import sys

from .util import imshow_many

# https://stackoverflow.com/questions/32120178/how-can-i-pass-null-to-an-external-library-using-ctypes-with-an-argument-decla/32138619
FloatArrayTypeBase = np.ctypeslib.ndpointer(dtype=np.float32, flags="C_CONTIGUOUS")


def debugPrint(msg):
    sys.stdout.write(msg.decode("utf-8"))


def _from_param(cls, obj):
    if obj is None:
        return obj
    return FloatArrayTypeBase.from_param(obj)


NullableFloatArrayType = type("FloatArrayType", (FloatArrayTypeBase,), {"from_param": classmethod(_from_param)})


class SMLM:
    def __init__(self, debugMode=False):
        thispath = os.path.dirname(os.path.abspath(__file__))

        if ctypes.sizeof(ctypes.c_voidp) == 4:
            raise RuntimeError(f"The SMLM library can only be used with 64-bit python.")

        if debugMode:
            dllpath = "/../../x64/Debug/smlm_cuda.dll"
        else:
            dllpath = "/../../x64/Release/smlm_cuda.dll"

        abs_dllpath = os.path.abspath(thispath + dllpath)

        if debugMode:
            print("Using " + abs_dllpath)
        self.debugMode = debugMode

        smlmlib = ctypes.CDLL(abs_dllpath)
        self.lib = smlmlib

        self.DebugPrintCallback = ctypes.CFUNCTYPE(None, ctypes.c_char_p)
        self._SetDebugPrintCallback = smlmlib.SetDebugPrintCallback
        self._SetDebugPrintCallback.argtypes = [self.DebugPrintCallback]

        self.DebugImageCallback = ctypes.CFUNCTYPE(
            None, ctypes.c_int32, ctypes.c_int32, ctypes.c_int32, ctypes.POINTER(ctypes.c_float), ctypes.c_char_p
        )
        self._SetDebugImageCallback = smlmlib.SetDebugImageCallback
        self._SetDebugImageCallback.argtypes = [self.DebugImageCallback]

        # CDLL_EXPORT void FFT(const cuFloatComplex* src, cuFloatComplex* dst, int batchsize, int siglen, int forward)
        self._FFT = smlmlib.FFT
        self._FFT.argtypes = [
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_int32,
        ]
        self._FFT2 = smlmlib.FFT2
        self._FFT2.argtypes = [
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_int32,
            ctypes.c_int32,
        ]

        self.SetDebugPrintCallback(debugPrint)

    def ListDebugImages(self):
        for i, v in enumerate(self.debugImages):
            print(f"{v[1]}: {v[0].shape}")

    def SetDebugPrintCallback(self, fn):
        self.dbgPrintCallback = self.DebugPrintCallback(fn)  # make sure the callback doesnt get garbage collected
        self._SetDebugPrintCallback(self.dbgPrintCallback)

    def SetDebugImageCallback(self, fn):
        self.dbgImgCallback = self.DebugImageCallback(fn)  # make sure the callback doesnt get garbage collected
        self._SetDebugImageCallback(self.dbgImgCallback)

    def FFT(self, src, forward=True):
        src_shape = src.shape
        if len(src.shape) == 1:
            src = [src]

        src = np.ascontiguousarray(src, dtype=np.complex64)
        dst = np.zeros(src.shape, dtype=np.complex64)

        batchsize = src.shape[0]
        siglen = src.shape[1]

        self._FFT(src, dst, batchsize, siglen, forward)
        if len(src_shape) == 1:
            return dst[0]
        return dst

    def IFFT(self, src):
        return self.FFT(src, forward=False)

    def FFT2(self, src, forward=True):
        src_shape = src.shape
        if len(src.shape) == 2:
            src = [src]

        src = np.ascontiguousarray(src, dtype=np.complex64)
        dst = np.zeros(src.shape, dtype=np.complex64)

        batchsize = src.shape[0]
        sigw = src.shape[1]
        sigh = src.shape[2]

        self._FFT2(src, dst, batchsize, sigw, sigh, forward)

        if len(src_shape) == 2:
            return dst[0]
        return dst

    def IFFT2(self, src):
        return self.FFT2(src, forward=False)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        # Free DLL so we can overwrite the file when we recompile
        ctypes.windll.kernel32.FreeLibrary.argtypes = [ctypes.wintypes.HMODULE]
        ctypes.windll.kernel32.FreeLibrary(self.lib._handle)
