import ctypes
import numpy as np
import matplotlib.pyplot as plt
import numpy.ctypeslib as ctl
import numbers

from smlmlib import image_queue
from smlmlib import psf
from smlmlib import spotdetect
from .base import SMLM, NullableFloatArrayType

Theta = ctypes.c_float * 4
FisherMatrix = ctypes.c_float * 16

self = None


class EstimationResult(ctypes.Structure):
    _fields_ = [
        ("estimate", Theta),
        ("initialValue", Theta),
        ("crlb", Theta),
        ("iterations", ctypes.c_int32),
        ("roiPosition", ctypes.c_int32 * 2),
        ("score", ctypes.c_float),
        ("loglikelihood", ctypes.c_float),
    ]


class Gauss3D_Calibration(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float * 4), ("y", ctypes.c_float * 4)]

    def __init__(self, x, y):
        self.x = (ctypes.c_float * 4)(*x)
        self.y = (ctypes.c_float * 4)(*y)

    @classmethod
    def from_file(cls, filename):
        calibration = np.load(filename).item()
        return cls(calibration.get("x"), calibration.get("y"))


class GaussianImageQueue(image_queue.ImageProcessingQueue):
    def GetFrameResults(self, frameIdx):
        nresults = self.GetResultCount(frameIdx)
        if nresults < 0:
            return []

        results = (EstimationResult * nresults)()
        self._GetResults(frameIdx, results)
        return results


class Gaussian:
    def __init__(self, smlm: SMLM):
        smlmlib = smlm.lib
        self.lib = smlm

        self._Gauss2D_ComputeMLE = smlmlib.Gauss2D_ComputeMLE
        self._Gauss2D_ComputeMLE.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctypes.POINTER(EstimationResult),
            ctypes.c_int32,
            NullableFloatArrayType,
            ctypes.c_float,  # sigma
            ctypes.c_int32,  # imgw
            ctypes.c_int32,  # maxiterations
            ctypes.c_float,  # lambdastep
            ctypes.c_int32,  # cuda
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # tracebuf
            ctypes.c_int32,  # tracebuflen per spot
        ]

        # CDLL_EXPORT void Gauss2D_EstimateIntensityBg(const float* imageData, Vector2f *IBg, int numspots, const Vector2f* xy,
        # float sigma, int imgw, int maxIterations, bool cuda)
        self._EstimateIBg = smlmlib.Gauss2D_EstimateIntensityBg
        self._EstimateIBg.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # images
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # IBg (result)
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # IBg_crlb (result)
            ctypes.c_int32,  # numspots
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # xy (input)
            ctypes.c_float,  # sigma
            ctypes.c_int32,  # imgw
            ctypes.c_int32,  # maxiterations
            ctypes.c_int32  # cuda
        ]
        
        self._Gauss2D_IntensityBg_CRLB = smlmlib.Gauss2D_IntensityBg_CRLB
        self._Gauss2D_IntensityBg_CRLB.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # crlb
            ctypes.c_int32,  # numspots
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # xyIBg
            ctypes.c_float,  # sigma
            ctypes.c_int32,  # imgw
            ctypes.c_bool  # cuda
        ]
         

        self._Gauss2D_CreateImageQueue = smlmlib.Gauss2D_CreateImageQueue
        self._Gauss2D_CreateImageQueue.argtypes = [
            ctypes.POINTER(image_queue.ImageQueueConfig),
            ctypes.c_float,  # sigma
            ctypes.c_int32,  # imgw
            ctypes.c_int32,  # maxSpots
            ctypes.c_void_p,  # spotDetectionFactory
            ctypes.c_void_p,
        ]  # calibrationTransformPtr
        self._Gauss2D_CreateImageQueue.restype = ctypes.c_void_p

        # (float * image, int imgw, int imgh, float * spotList, int nspots)
        self._Gauss2D_Draw = smlmlib.Gauss2D_Draw
        self._Gauss2D_Draw.argtypes = [
            ctl.ndpointer(np.float64, flags="aligned, c_contiguous"),  # mu
            ctypes.c_int32,
            ctypes.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # mu
            ctypes.c_int32,
            ctypes.c_float
        ]

        self._Gauss_CreatePSF_XYIBg = smlmlib.Gauss_CreatePSF_XYIBg
        self._Gauss_CreatePSF_XYIBg.argtypes = [ctypes.c_int32, ctypes.c_float, ctypes.c_int32]
        self._Gauss_CreatePSF_XYIBg.restype = ctypes.c_void_p

        self._Gauss_CreatePSF_XYI = smlmlib.Gauss_CreatePSF_XYI
        self._Gauss_CreatePSF_XYI.argtypes = [ctypes.c_int32, ctypes.c_float, ctypes.c_int32]
        self._Gauss_CreatePSF_XYI.restype = ctypes.c_void_p

        self._Gauss_CreatePSF_XYZIBg = smlmlib.Gauss_CreatePSF_XYZIBg
        self._Gauss_CreatePSF_XYZIBg.argtypes = [ctypes.c_int32, ctypes.POINTER(Gauss3D_Calibration), ctypes.c_int32]
        self._Gauss_CreatePSF_XYZIBg.restype = ctypes.c_void_p

    def ComputeFisherMatrix(self, theta, sigma, imgw):
        with self.CreatePSF_XYIBg(imgw, sigma, False) as psf:
            fi = psf.FisherMatrix(theta)
            mu = psf.ExpectedValue(theta)
            return mu, fi

    # Spots is an array with rows: [ x,y, sigmaX, sigmaY, intensity ]
    def Draw(self, img, spots, addSigma=0):
        spots = np.ascontiguousarray(spots, dtype=np.float32)
        nspots = spots.shape[0]
        assert spots.shape[1] == 5
        img = np.ascontiguousarray(img, dtype=np.float64)
        self._Gauss2D_Draw(img, img.shape[1], img.shape[0], spots, nspots, addSigma)
        return img

    def Plot(self, theta, sigma, imgw, sample=False):
        mu, fi = self.ComputeFisherMatrix(theta, sigma, imgw)
        mu = mu[0]
        if sample:
            mu = np.random.poisson(mu)

        plt.figure()
        plt.imshow(mu)

    def EstimateIBg(self, images, psfSigma, xy, useCuda=False):
        images = np.ascontiguousarray(images, dtype=np.float32)
        xy = np.ascontiguousarray(xy, dtype=np.float32)
        assert len(images.shape) == 3
        numspots = len(images)
        assert xy.shape == (numspots, 2)
        imgw = images.shape[1]
        result = np.zeros((numspots, 2), dtype=np.float32)
        crlb = np.zeros((numspots, 2), dtype=np.float32)
        if numspots > 0:
            self._EstimateIBg(images, result, crlb, numspots, xy, psfSigma, imgw, 100, useCuda)
        return result, crlb

#CDLL_EXPORT void Gauss2D_IntensityBg_CRLB(Vector2f* IBg_crlb,
#	int numspots, const Vector4f* xyIBg, float sigma, int imgw, bool cuda)
    def CRLB_IBg(self, xy, IBg, roisize, psfSigma, useCuda=False):
        numspots = len(xy)
        assert xy.shape == (numspots, 2)
        assert IBg.shape == (numspots, 2)
        xyIBg = np.zeros((numspots,4),dtype=np.float32)
        xyIBg[:,[0,1]] = xy
        xyIBg[:,[2,3]] = IBg
        crlb = np.zeros((numspots, 2), dtype=np.float32)
        if numspots > 0:
            self._Gauss2D_IntensityBg_CRLB(crlb, numspots, xyIBg, psfSigma, roisize, useCuda)
        return crlb
    

    def CreateImageQueue(
        self, w, h, psfSigma, roisize, maxSpotsPerFrame, spotDetectorConfig=None
    ) -> image_queue.ImageProcessingQueue:
        cfg = image_queue.ImageQueueConfig.make(w, h)

        # Interpret the config as threshold
        if isinstance(spotDetectorConfig, numbers.Number):
            spotDetectorConfig = spotdetect.SpotDetectorConfig(psfSigma, roisize, spotDetectorConfig)

        if spotDetectorConfig is None:
            spotDetectorConfig = spotdetect.SpotDetectorConfig(psfSigma, roisize)

        with spotDetectorConfig.CreateNativeFactory(self.lib) as sdf:
            inst = self._Gauss2D_CreateImageQueue(cfg, psfSigma, roisize, maxSpotsPerFrame, sdf.inst, None)

        return GaussianImageQueue(self.lib, inst)

    # CDLL_EXPORT PSF* Gauss_CreatePSF_XYIBg(int roisize, float sigma, bool cuda);
    def CreatePSF_XYIBg(self, roisize, sigma, cuda) -> psf.PSF:
        inst = self._Gauss_CreatePSF_XYIBg(roisize, sigma, cuda)
        return psf.PSF(self.lib, inst, roisize)

    def CreatePSF_XYI(self, roisize, sigma, cuda) -> psf.PSF:
        inst = self._Gauss_CreatePSF_XYI(roisize, sigma, cuda)
        return psf.PSF(self.lib, inst, roisize)

    # CDLL_EXPORT PSF* Gauss_CreatePSF_XYZIBg(int roisize, const Gauss3D_Calibration& calib, bool cuda);
    def CreatePSF_XYZIBg(self, roisize, calib: Gauss3D_Calibration, cuda) -> psf.PSF:
        inst = self._Gauss_CreatePSF_XYZIBg(roisize, calib, cuda)
        return psf.PSF(self.lib, inst, roisize)


# CDLL_EXPORT Gauss2DProcessingQueue* Gauss2D_CreateImageQueue(const ImageQueueConfig& config, float sigma, int imgw,
# 	int maxSpotsPerFrame, float spotDetectionThreshold, sCMOS_CalibrationTransform **p_calibTransform);
