"""

General purpose


CDLL_EXPORT void IPQ_Destroy(ImageProcessingQueueBase* q);
CDLL_EXPORT void IPQ_PushFrame(ImageProcessingQueueBase* q, int frame, float* imgptr);
CDLL_EXPORT void IPQ_Flush(ImageProcessingQueueBase* q);
CDLL_EXPORT int IPQ_GetQueueLength(ImageProcessingQueueBase* q, int* maxQueueLen = 0);



CDLL_EXPORT void IPQ_Start(ImageProcessingQueueBase* q);
CDLL_EXPORT void IPQ_Stop(ImageProcessingQueueBase* q);

CDLL_EXPORT int IPQ_GetFrameCount(ImageProcessingQueueBase* q);
CDLL_EXPORT int IPQ_GetResultCount(ImageProcessingQueueBase* q, int frame);
CDLL_EXPORT void IPQ_GetResults(ImageProcessingQueueBase* q, int frame, void* data, int nbytes);

"""
import ctypes as ct
import numpy as np
import numpy.ctypeslib as ctl


class ImageQueueConfig(ct.Structure):
    _fields_ = [
        ("width", ct.c_int32),
        ("height", ct.c_int32),
        ("maxQueueLen", ct.c_int32),
        ("historyLen", ct.c_int32),
        ("cudaStreamCount", ct.c_int32),
    ]

    def make(w, h, maxQueueLen=20, historyLen=1, cudaStreamCount=3):
        return ImageQueueConfig(w, h, maxQueueLen, historyLen, cudaStreamCount)


class ImageProcessingQueue:
    def __init__(self, smlmlib, queueInst):
        self.inst = queueInst

        lib = smlmlib.lib

        InstancePtrType = ct.c_void_p
        self._IPQ_Destroy = lib.IPQ_Destroy
        self._IPQ_Destroy.argtypes = [InstancePtrType]

        self._IPQ_GetStreamState = lib.IPQ_GetStreamState
        self._IPQ_GetStreamState.argtypes = [
            InstancePtrType,
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),
        ]

        self._IPQ_PushFrameF32 = lib.IPQ_PushFrameF32
        self._IPQ_PushFrameF32.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ct.c_void_p,
            ct.c_int32,
        ]

        self._IPQ_PushFrameU16 = lib.IPQ_PushFrameU16
        self._IPQ_PushFrameU16.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.uint16, flags="aligned, c_contiguous"),
            ct.c_void_p,
            ct.c_int32,
        ]

        self._IPQ_PushFrameU8 = lib.IPQ_PushFrameU8
        self._IPQ_PushFrameU8.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.uint8, flags="aligned, c_contiguous"),
            ct.c_void_p,
            ct.c_int32,
        ]

        self._IPQ_GetQueueLength = lib.IPQ_GetQueueLength
        self._IPQ_GetQueueLength.restype = ct.c_int32
        self._IPQ_GetQueueLength.argtypes = [InstancePtrType, ct.POINTER(ct.c_int32)]

        # CDLL_EXPORT int IPQ_GetFrameCount(ImageProcessingQueueBase* q);
        self._IPQ_GetResultFrameCount = lib.IPQ_GetResultFrameCount
        self._IPQ_GetResultFrameCount.restype = ct.c_int32
        self._IPQ_GetResultFrameCount.argtypes = [InstancePtrType]

        self._IPQ_GetFramesInMemory = lib.IPQ_GetFramesInMemory
        self._IPQ_GetFramesInMemory.restype = ct.c_int32
        self._IPQ_GetFramesInMemory.argtypes = [InstancePtrType]

        self._IPQ_IsIdle = lib.IPQ_IsIdle
        self._IPQ_IsIdle.restype = ct.c_int32
        self._IPQ_IsIdle.argtypes = [InstancePtrType]

        self._IPQ_Flush = lib.IPQ_Flush
        self._IPQ_Flush.argtypes = [InstancePtrType]

        self._IPQ_Start = lib.IPQ_Start
        self._IPQ_Start.argtypes = [InstancePtrType]

        self._IPQ_Stop = lib.IPQ_Stop
        self._IPQ_Stop.argtypes = [InstancePtrType]

        self._IPQ_SetFrameStep = lib.IPQ_SetFrameStep
        self._IPQ_SetFrameStep.argtypes = [InstancePtrType, ct.c_int32]

        # CDLL_EXPORT int IPQ_GetResultCount(ImageProcessingQueueBase* q, int frame);
        self._IPQ_GetResultCount = lib.IPQ_GetResultCount
        self._IPQ_GetResultCount.restype = ct.c_int32
        self._IPQ_GetResultCount.argtypes = [InstancePtrType]

        self._IPQ_GetConfig = lib.IPQ_GetConfig
        self._IPQ_GetConfig.argtypes = [InstancePtrType, ct.POINTER(ImageQueueConfig)]

        # CDLL_EXPORT void IPQ_GetResults(ImageProcessingQueueBase* q, int frame, void* data, int nbytes);
        self._IPQ_GetResults = lib.IPQ_GetResults
        self._IPQ_GetResults.argtypes = [InstancePtrType, ct.c_int32, ct.c_void_p, ct.c_int32]

        self.config = self.GetConfig()

    def Flush(self):
        self._IPQ_Flush(self.inst)

    def Start(self):
        self._IPQ_Start(self.inst)

    def Stop(self):
        self._IPQ_Stop(self.inst)

    def Destroy(self):
        if self.inst is not None:
            self._IPQ_Destroy(self.inst)
            self.inst = None

    def IsIdle(self):
        return self._IPQ_IsIdle(self.inst) != 0

    def GetFramesInMemory(self):
        return self._IPQ_GetFramesInMemory(self.inst)

    def GetResultFrameCount(self):
        return self._IPQ_GetResultFrameCount(self.inst)

    def GetResultCount(self, frameIdx):
        return self._IPQ_GetResultCount(self.inst, frameIdx)

    def GetConfig(self):
        cfg = ImageQueueConfig()
        self._IPQ_GetConfig(self.inst, cfg)
        return cfg

    def GetQueueLength(self):
        maxLen = ct.c_int32()
        len = self._IPQ_GetQueueLength(self.inst, maxLen)
        return len

    def SetFrameStep(self, frameStep=1):
        self._IPQ_SetFrameStep(self.inst, frameStep)

    def GetStreamState(self):
        frames = np.zeros(self.config.cudaStreamCount, dtype=np.int32)
        tasks = np.zeros(self.config.cudaStreamCount, dtype=np.int32)
        activeStreams = self._IPQ_GetStreamState(self.inst, frames, tasks)
        return frames, tasks, activeStreams

    def ReportState(self):
        fr, tsk, active = self.GetStreamState()
        print('Queue length: {0}, Frames processed: {1}. Current frames: {2}. Tasks: {3}. #ActiveStreams={4} '.format( 
                self.GetQueueLength(), self.GetResultFrameCount(), fr, tsk, active ))

    # Only to be used by subclasses of the queue system that know the proper type that goes in here
    def _GetResults(self, frame, results):
        self._IPQ_GetResults(self.inst, frame, results, ct.sizeof(results))

    def PushFrame(self, frameNum, img, tagData=None):
        if tagData is not None:
            tagDataPtr = tagData.ctypes.data_as(ct.c_void_p)
            tagDataSize = tagData.itemsize * tagData.size
        else:
            tagDataPtr = None
            tagDataSize = 0

        if img.dtype == np.uint8:
            self._IPQ_PushFrameU8(self.inst, frameNum, img, tagDataPtr, tagDataSize)
        elif img.dtype == np.uint16:
            self._IPQ_PushFrameU16(self.inst, frameNum, img, tagDataPtr, tagDataSize)
        else:
            self._IPQ_PushFrameF32(
                self.inst, frameNum, np.ascontiguousarray(img, dtype=np.float32), tagDataPtr, tagDataSize
            )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.Destroy()
