import ctypes as ct
import numpy as np
import numpy.ctypeslib as ctl
from .base import NullableFloatArrayType


class PSF:
    def __init__(self, smlmlib, psfInst, roisize):
        self.inst = psfInst
        self.roisize = roisize
        lib = smlmlib.lib

        InstancePtrType = ct.c_void_p

        self._PSF_Delete = lib.PSF_Delete
        self._PSF_Delete.argtypes = [InstancePtrType]

        self._PSF_ThetaSize = lib.PSF_ThetaSize
        self._PSF_ThetaSize.restype = ct.c_int32
        self._PSF_ThetaSize.argtypes = [InstancePtrType]

        self._PSF_ThetaFormat = lib.PSF_ThetaFormat
        self._PSF_ThetaFormat.restype = ct.c_char_p
        self._PSF_ThetaFormat.argtypes = [InstancePtrType]

        self._PSF_NumConstants = lib.PSF_NumConstants
        self._PSF_NumConstants.restype = ct.c_int32
        self._PSF_NumConstants.argtypes = [InstancePtrType]

        self._PSF_SampleIndexDims = lib.PSF_SampleIndexDims
        self._PSF_SampleIndexDims.restype = ct.c_int32
        self._PSF_SampleIndexDims.argtypes = [InstancePtrType]

        self._PSF_SampleCount = lib.PSF_SampleCount
        self._PSF_SampleCount.restype = ct.c_int32
        self._PSF_SampleCount.argtypes = [InstancePtrType]

        # CDLL_EXPORT void PSF_ComputeExpectedValue(PSF* psf, int numspots, const float* theta, const float* constants, const int* spotpos, float* ev);
        self._PSF_ComputeExpectedValue = lib.PSF_ComputeExpectedValue
        self._PSF_ComputeExpectedValue.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"), #theta 
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"), # constants
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # spotpos
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"), # ev
        ]

        # CDLL_EXPORT void PSF_ComputeFisherMatrix(PSF* psf, int numspots, const float* theta, const float* constants,const int* spotpos,  float* fi);
        self._PSF_ComputeFisherMatrix = lib.PSF_ComputeFisherMatrix
        self._PSF_ComputeFisherMatrix.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # spotpos
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"), #fi
        ]

        # CDLL_EXPORT void PSF_ComputeDerivatives(PSF* psf, int numspots, const float* theta, const float* constants, const int* spotpos, float* derivatives, float* ev);
        self._PSF_ComputeDerivatives = lib.PSF_ComputeDerivatives
        self._PSF_ComputeDerivatives.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # spotpos
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
        ]

        # CDLL_EXPORT void PSF_ComputeMLE(PSF* psf, int numspots, const float* sample, const float* constants,const int* spotpos, 
        # const float* initial, float* theta, int* iterations, int maxiterations, float levmarAlpha, float* trace, int traceBufLen);
        self._PSF_ComputeMLE = lib.PSF_ComputeMLE
        self._PSF_ComputeMLE.argtypes = [
            InstancePtrType,
            ct.c_int32,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # sample
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # constants
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # spotpos
            NullableFloatArrayType,  # initial
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # theta
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # iterations
            ct.c_int32,
            ct.c_float,
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),
            ct.c_int32,
        ]

        self.thetasize = self._PSF_ThetaSize(self.inst)
        self.numconst = self._PSF_NumConstants(self.inst)
        self.samplecount = self._PSF_SampleCount(self.inst)
        self.indexdims = self._PSF_SampleIndexDims(self.inst)

    def _checkparams(self, theta, constants, spotpos):
        theta = np.array(theta)
        if len(theta.shape) == 1:
            theta = [theta]
        theta = np.ascontiguousarray(theta, dtype=np.float32)
        numspots = len(theta)
        if constants is None:
            constants = np.zeros((numspots, self.numconst), dtype=np.float32)
        else:
            constants = np.ascontiguousarray(constants, dtype=np.float32)
            assert(constants.shape == [numspots, self.numconst])
        if spotpos is None:
            spotpos = np.zeros((numspots, self.indexdims), dtype=np.int32)
        else:
            spotpos = np.ascontiguousarray(spotpos, dtype=np.int32)
            assert(spotpos.shape == [numspots, self.indexdims])
        return theta,constants,spotpos

    def ExpectedValue(self, theta, constants=None, spotpos=None):
        theta,constants,spotpos=self._checkparams(theta,constants,spotpos)
        ev = np.zeros((len(theta), self.roisize, self.roisize), dtype=np.float32)
        self._PSF_ComputeExpectedValue(self.inst, len(theta), theta, constants, spotpos, ev)
        return ev

    def FisherMatrix(self, theta, constants=None, spotpos=None):
        theta,constants,spotpos=self._checkparams(theta,constants,spotpos)
        fi = np.zeros((len(theta), self.thetasize, self.thetasize), dtype=np.float32)
        self._PSF_ComputeFisherMatrix(self.inst, len(theta), theta, constants, spotpos, fi)
        return fi

    def CRLB(self, theta, constants=None, spotpos=None):
        fisher = self.FisherMatrix(theta, constants, spotpos)
        return np.array([np.sqrt(np.abs(np.diag(np.linalg.inv(f)))) for f in fisher])

    def Derivatives(self, theta, constants=None, spotpos=None):
        theta,constants,spotpos=self._checkparams(theta,constants,spotpos)
        deriv = np.zeros((len(theta), self.thetasize, self.roisize, self.roisize), dtype=np.float32)
        ev = np.zeros((len(theta), self.roisize, self.roisize), dtype=np.float32)
        self._PSF_ComputeDerivatives(self.inst, len(theta), theta, constants, spotpos, deriv, ev)
        return deriv, ev

    def ComputeMLE(self, imgs, initial=None, maxiterations=50, levmarInitialAlpha=0.1, constants=None, spotpos=None):
        numspots = len(imgs)
        imgs = np.ascontiguousarray(imgs, dtype=np.float32)
        traceBufLen = maxiterations
        iterations = np.zeros(numspots, np.int32)
        theta = np.zeros((numspots, self.thetasize), dtype=np.float32)
        trace = np.zeros((numspots, traceBufLen, self.thetasize), dtype=np.float32)

        theta,constants,spotpos = self._checkparams(theta,constants,spotpos)

        if initial is not None:
            initial = np.ascontiguousarray(initial, dtype=np.float32)
            assert len(initial) == numspots

        self._PSF_ComputeMLE(
            self.inst,
            numspots,
            imgs,
            constants,
            spotpos,
            initial,
            theta,
            iterations,
            maxiterations,
            levmarInitialAlpha,
            trace,
            traceBufLen,
        )

        traces = []
        for k in range(numspots):
            traces.append(trace[k, 0 : iterations[k] + 1, :])

        return theta, iterations, traces

    def Destroy(self):
        if self.inst is not None:
            self._PSF_Delete(self.inst)
            self.inst = None

    def NumConstants(self):
        return self._PSF_NumConstants(self.inst)

    def ThetaFormat(self):
        return self._PSF_ThetaFormat(self.inst)

    def ThetaSize(self):
        return self._PSF_ThetaSize(self.inst)

    def SampleCount(self):
        return self._PSF_SampleCount(self.inst)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.Destroy()
