# -*- coding: utf-8 -*-


import ctypes
import numpy as np
import matplotlib.pyplot as plt
import numpy.ctypeslib as ctl
import scipy.stats
from enum import Enum

from smlmlib import image_queue
from .base import SMLM, NullableFloatArrayType
from smlmlib import gaussian

Theta = ctypes.c_float * 4
FisherMatrix = ctypes.c_float * 16
EstimationResult = gaussian.EstimationResult
Modulation = ctypes.c_float * 4


class SILM_Mode(Enum):
    SIMPLE = 0
    SWITCH_FRAMES = 1
    SWITCH_FRAME_SUM = 2
    CHOP_ENDS = 3


class SILM_ASW_Params(ctypes.Structure):
    _fields_ = [
        ("imgw", ctypes.c_int32),
        ("numep", ctypes.c_int32),
        ("sigma", ctypes.c_float),
        ("levMarMaxIt", ctypes.c_int32),
        ("levMarLambdaStep", ctypes.c_float),
        ("mode", ctypes.c_int32),
    ]

    def make(imgw, numep, sigma, levMarIt=100, startLambdaStep=0.1, mode=SILM_Mode.SWITCH_FRAMES):
        return SILM_ASW_Params(imgw, numep, sigma, levMarIt, startLambdaStep, mode)



class SILM:
    def __init__(self, smlm):
        smlmlib = smlm.lib
        self.lib = smlm
        self._SILM_ASW_ComputeMLE = smlmlib.SILM_ASW_ComputeMLE
        self._SILM_ASW_ComputeMLE.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # img
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # modulation
            ctypes.POINTER(EstimationResult),  # results
            ctypes.c_int32,  # numspots
            ctypes.c_int32,  # numframes
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # startPatterns
            ctypes.POINTER(SILM_ASW_Params),  # p
            NullableFloatArrayType,  # initialValue
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # roipos
            ctypes.c_int32,  # flags
            NullableFloatArrayType,  # tracebuf
            ctypes.c_int32,  # tracebuflen per spot
        ]

        self._SILM_ASW_ComputeFisherMatrix = smlmlib.SILM_ASW_ComputeFisherMatrix
        self._SILM_ASW_ComputeFisherMatrix.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # mu
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # fi
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # phi
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # theta
            ctypes.c_int32,  # numspots
            ctypes.c_int32,  # numframes
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # startPatterns
            ctypes.POINTER(SILM_ASW_Params),
        ]

        # void SILM_GenerateFrames(const float * blinktrace, float *framedata,float* excitationPerSpot,
        # int numspots, int numframes, const float * phi, int startPattern, const SILM_Theta * theta, const SILM_ASW_Params & p)
        """
                self._SILM_GenerateFrames = smlmself.SILM_GenerateFrames
                self._SILM_GenerateFrames.argtypes =[
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # blinktrace
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # framedata
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # excitationPerSpot
                    ctypes.c_int32,                                             # numspots
                    ctypes.c_int32,                                             # numframes
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # phi
                    ctypes.c_int32,                                             # startPattern
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # theta
                    ctypes.POINTER(SILM_ASW_Params)                           # silmParams
                    ]

        #CDLL_EXPORT void SILM_GenerateBlinkingTraces(float T_on, float T_off, SILM_Theta* theta, int numframes,
        #const float* phi, SILM_ASW_Params& params, float fps, int numspots, float* trace);

                self._SILM_GenerateBlinkingTraces = smlmself.SILM_GenerateBlinkingTraces
                self._SILM_GenerateBlinkingTraces.argtypes = [
                    ctypes.c_float, #T_on
                    ctypes.c_float, #T_off
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # theta
                    ctypes.c_int32, # numframes
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous'), # phi
                    ctypes.POINTER(SILM_ASW_Params),                           # silmParams
                    ctypes.c_float, #fps
                    ctypes.c_int32, # numspots
                    ctl.ndpointer(np.float32, flags='aligned, c_contiguous') # traces
                    ]
                #(T_on, T_off, theta, numframes, phi, silm_p, fps, numspots, traces)
        """

        #        SILM_BlinkHMM_ComputeMLE(const float * frameData, const float * phi, SILM_EstimationResult * results,
        # 	int numspots, int numframes, int startPattern, const SILM_ASW_Params & p, bool useCuda,
        # SILM_Theta * optimizerTrace,
        # 	int traceBufLen, float * p_on_estim, const SILM_BlinkHMM_Params & blinkParams)


        self._SILM_CreateImageQueue = smlmlib.SILM_CreateImageQueue
        self._SILM_CreateImageQueue.argtypes = [
            ctypes.POINTER(SILM_ASW_Params),  # p
            ctypes.POINTER(image_queue.ImageQueueConfig),
            ctypes.c_void_p,  # spotdetectorfactory
            ctypes.c_int32,  # maxspotsperframe
            ctypes.c_float, # silmMinIntensity
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # SILM_modulation[numep]
        ]
        self._SILM_CreateImageQueue.restype = ctypes.c_void_p

        self._SILM_SignalCorrelation = smlmlib.SILM_SignalCorrelation
        self._SILM_SignalCorrelation.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # signalA
            ctypes.c_int32,  # numsigA
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # signalB
            ctypes.c_int32,  # numsigB
            ctypes.c_int32,  # siglen
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # output
            ctypes.c_int32,  # useConvolution (correlation or convolution?)
        ]
        
       # CDLL_EXPORT void SILM_DFT2D_Points(const Vector3f* xyI, int numpts, const Vector2f* k, 
       # int numk, Vector2f* output, bool useCuda);

        self._SILM_DFT2D_Points = smlmlib.SILM_DFT2D_Points
        self._SILM_DFT2D_Points.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # xyI
            ctypes.c_int32,  # numpts
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # k
            ctypes.c_int32,  # numk
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),  # output
            ctypes.c_bool # useCuda
        ]

        # CDLL_EXPORT void FFT(const cuFloatComplex* src, cuFloatComplex* dst, int batchsize, int siglen, int forward)
        self._FFT = smlmlib.FFT
        self._FFT.argtypes = [
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),  # src
            ctl.ndpointer(np.complex64, flags="aligned, c_contiguous"),  # dst
            ctypes.c_int32,  # batchsize
            ctypes.c_int32,  # numsigA
            ctypes.c_int32,  # forward
        ]

        self._SILM_ProjectPointData = smlmlib.SILM_ProjectPointData
        self._SILM_ProjectPointData.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # xyI
            ctypes.c_int32,  # numpts
            ctypes.c_int32,  # projectionWidth
            ctypes.c_float,  # scale
            ctypes.c_int32,  # numProjAngles
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # projectionAngles
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # output
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # output
        ]
        
        self._SILM_ASW_ComputeOnOffProb = smlmlib.SILM_ASW_ComputeOnOffProb 
        self._SILM_ASW_ComputeOnOffProb.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # rois [numspots,numframes,roisize,roisize]
             ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # mod[numep]
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # gaussFits [numspots]
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # IBg[out]
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # probOnOff[numspots,numframes,2]
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # crlbVariances[numspots,numframes,2]
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # expectedIntensities[numspots,numframes]
            ctypes.POINTER(SILM_ASW_Params),  # p
            ctypes.c_int32,  # numframes
            ctypes.c_int32,  # numspots
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # startPatterns[numspots]
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # roipos[numspots]
            ctypes.c_bool,  # useCuda
                ]
        

        

    # Convert an array of phases to an array of alternating XY modulation parameters
    def phase_to_mod(self, phases, omega, depth=1):
        mod = np.zeros((*phases.shape, 5), dtype=np.float32)
        mod[..., 0::2, 0] = omega  # kx
        mod[..., 1::2, 1] = omega  # ky
        mod[..., 2] = depth
        mod[..., 3] = phases
        mod[..., 4] = 1/len(mod)
        return mod

#CDLL_EXPORT void SILM_ASW_ComputeOnOffProb(const float* rois, 
#const SILM_Modulation* modulation, Vector4f* gaussFits, 
#	Vector2f* IBg, Vector2f* probOnOff, const SILM_ASW_Params& params, int numframes, 
#	int numspots, const int* startPatterns, const int2* roipos, bool useCuda)
    def SILM_ASW_ComputeOnOffProb(self, images, mod, xyIBg_gauss, silmParams: SILM_ASW_Params, 
                                  startPatterns, roipos, useCuda):
        mod = np.ascontiguousarray(mod, dtype=np.float32)
        images = np.ascontiguousarray(images, dtype=np.float32)
        xyIBg_gauss = np.ascontiguousarray(xyIBg_gauss,dtype=np.float32)
        numframes = images.shape[1]
        numspots = images.shape[0]
        
        probOnOff = np.zeros((numspots,numframes,2),dtype=np.float32)
        crlbVariances = np.zeros((numspots,numframes,2),dtype=np.float32)
        expectedIntensity = np.zeros((numspots,numframes),dtype=np.float32)
        IBg = np.zeros((numspots,numframes,2),dtype=np.float32)
        startPatterns = np.ascontiguousarray(startPatterns,dtype=np.int32)
        roipos = np.ascontiguousarray(roipos, dtype=np.int32)
        
        self._SILM_ASW_ComputeOnOffProb(
            images, mod, xyIBg_gauss, IBg, probOnOff, crlbVariances, expectedIntensity, silmParams, 
            numframes, numspots, startPatterns,roipos, useCuda)
        
        return probOnOff, IBg, crlbVariances, expectedIntensity
        

    # returns [expected values, fishermatrices]
    def SILM_ASW_ComputeFisherMatrix(self, mod, theta, numframes, startPatterns, silmParams : SILM_ASW_Params):
        mod = np.ascontiguousarray(mod, dtype=np.float32)
        theta = np.ascontiguousarray(theta, dtype=np.float32)
        assert mod.shape[1] == silmParams.numep
        assert mod.shape[2] == 5
        assert mod.shape[0] == len(theta)
        startPatterns = np.ascontiguousarray(startPatterns, dtype=np.int32)
        if len(theta.shape) == 1:
            theta = np.array([theta])

        assert len(startPatterns) == len(theta)
        numspots = theta.shape[0]

        assert theta.shape[1] == 4

        mu = np.zeros((numspots, numframes, silmParams.imgw, silmParams.imgw), dtype=np.float32)
        fi = np.zeros((numspots, 4, 4), dtype=np.float32)
        self._SILM_ASW_ComputeFisherMatrix(
            mu, fi, mod, theta, numspots, numframes, startPatterns, silmParams
        )
        return mu, fi


    # CDLL_EXPORT void SILM_ASW_ComputeMLE(const float * imageData, const float * backgroundMatrix, const float * phi,
    # 	SILM_EstimationResult *results, int numspots, const SILM_ASW_Params & p, const SILM_Theta* initialTheta = 0, bool cuda=false);

    # returns [estimated theta, EstimationResult's]
    def SILM_ASW_ComputeMLE(
        self, images, mod, p: SILM_ASW_Params, startPatterns, cuda=False, initialValues=None, roipos=None, traces=False
    ):
        assert len(images.shape) == 4
        numspots = images.shape[0]
        numframes = images.shape[1]
        imgw = images.shape[2]
        assert images.shape[2] == images.shape[3]
        mod = np.ascontiguousarray(mod, dtype=np.float32)
        assert mod.shape[0] == numspots
        assert mod.shape[1] == p.numep
        images = np.ascontiguousarray(images, dtype=np.float32)
        startPatterns = np.ascontiguousarray(startPatterns, dtype=np.int32)

        if roipos is None:
            roipos = np.zeros((numspots, 2), dtype=np.int32)

        roipos = np.ascontiguousarray(roipos, dtype=np.int32)

        if initialValues is not None:
            initialValues = np.array(initialValues, dtype=np.float32)
            assert len(initialValues.shape) == 2 and initialValues.shape[1] == 4

        if traces:
            traceBufLen = p.levMarMaxIt
            traces = np.zeros((numspots, traceBufLen, 4), dtype=np.float32)
        else:
            traceBufLen = 0
            traces = None

        flags = 0
        if cuda:
            flags += 1

        results = (EstimationResult * numspots)()
        self._SILM_ASW_ComputeMLE(
            images,
            mod,
            results,
            numspots,
            numframes,
            startPatterns,
            p,
            initialValues,
            roipos,
            flags,
            traces,
            traceBufLen,
        )

        results_list = []
        for i in range(numspots):
            r = results[i]
            if traces:
                r.trace = traces[i, 0 : results[i].iterations]
            results_list.append(r)
            # I need to evaluate this otherwise the trace disappears..
            # Am I not allowed to add extra fields to a ctypes.Structure?
            if traces:
                r.trace
            # print(f"SILM_ASW: spot {i} iterations={len(r.trace)}")

        return np.array([r.estimate for r in results]), results_list




    def Params(self, imgw, numep, sigma, levMarIt=100, startLambdaStep=0.1):
        return SILM_ASW_Params(imgw, numep, sigma, levMarIt, startLambdaStep)
    
    
    def SILM_DFT2D_Points(self, xyI, k, useCuda=True):
        xyI = np.ascontiguousarray(xyI, dtype=np.float32)
        numpts = len(xyI)
        k = np.ascontiguousarray(k, dtype=np.float32)
        output = np.zeros( len(k), dtype=np.complex64)
        self._SILM_DFT2D_Points(xyI, numpts, k, len(k), output, useCuda)
        return output

    def SILM_CreateImageQueue(self, w, h, numframes, silmParams, maxSpotsPerFrame, silmMinIntensity, spotDetectorConfig, mod):
        cfg = image_queue.ImageQueueConfig.make(w, h, maxQueueLen=50, historyLen=numframes)

        with spotDetectorConfig.CreateNativeFactory(self.lib) as sdf:
            inst = self._SILM_CreateImageQueue(silmParams, cfg, sdf.inst, 
                                               maxSpotsPerFrame, silmMinIntensity, np.ascontiguousarray( mod,dtype=np.float32))

        return gaussian.GaussianImageQueue(self.lib, inst)


    # ImageProcessingQueue<SILM_EstimationResult>* SILM_SetupProcessingQueue(const SILM_ASW_Params & p, const ImageQueueConfig & iqc,
    # 	float detectionTreshold, int maxSpotsPerFrame, SILM_BlinkHMM_Params & hmmParams, const float *fullframePhi)


    # CDLL_EXPORT void SILM_ProjectPointData(const Vector3f *xyI, int numpts, int projectionWidth,
    # 	float scale, int numProjAngles, const float *projectionAngles, float* output)
    def ProjectPoints(self, xyI, projectionWidth, scale, projectionAngles):
        numProjAngles = len(projectionAngles)
        assert xyI.shape[1] == 3
        xyI = np.ascontiguousarray(xyI, dtype=np.float32)
        output = np.zeros((numProjAngles, projectionWidth), dtype=np.float32)
        shifts = np.zeros((numProjAngles), dtype=np.float32)

        self._SILM_ProjectPointData(
            xyI,
            len(xyI),
            projectionWidth,
            scale,
            numProjAngles,
            np.array(projectionAngles, dtype=np.float32),
            output,
            shifts,
        )
        return output, shifts

    ##CDLL_EXPORT void FFT(const cuFloatComplex* src, cuFloatComplex* dst, int batchsize, int siglen, int forward)

    def FFT(self, src, forward=True):
        batchsize = len(src)
        src = np.ascontiguousarray(src, dtype=np.complex64)
        dst = np.zeros(src.shape, dtype=np.complex64)
        self._FFT(src, dst, batchsize, src.shape[1], forward)
        return dst
