# -*- coding: utf-8 -*-

import ctypes
from .base import SMLM
import numpy as np
import numpy.ctypeslib as ctl
from smlmlib import image_queue
import numbers


class SpotDetectorNativeFactory:
    def __init__(self, inst, destructor):
        self.inst = inst
        self.destructor = destructor

    def __enter__(self):
        return self

    def __exit__(self, *args):
        d = self.destructor
        d(self.inst)


class SpotDetectorConfig(ctypes.Structure):
    # detectionImage = uniform1 - uniform2
    # Selected spot locations = max(detectionImage, maxFilterSize) == detectionImage
    _fields_ = [
        ("uniformFilter1Size", ctypes.c_int32),
        ("uniformFilter2Size", ctypes.c_int32),
        ("maxFilterSize", ctypes.c_int32),
        ("roisize", ctypes.c_int32),  # Roisize is used to remove ROIs near the image border
        ("minIntensity", ctypes.c_float),  # Only spots where detectionImage > intensityThreshold are selected
        ("maxIntensity", ctypes.c_float),
    ]  # Only spots where detectionImage > intensityThreshold are selected

    def __init__(self, psfSigma, roisize, minIntensity=10, maxIntensity=1000):
        print(f"Spot detector psf sigma: {psfSigma}")
        self.uniformFilter1Size = int(psfSigma * 2 + 1)
        self.uniformFilter2Size = self.uniformFilter1Size * 2
        self.maxFilterSize = int(psfSigma * 3)
        self.roisize = roisize
        self.minIntensity = minIntensity
        self.maxIntensity = maxIntensity
        

    def print(self):
        print("Spot detector config: " +
              f"uniform filter 1: {self.uniformFilter1Size}"
              f"uniform filter 2: {self.uniformFilter2Size}"
              f"maxfilter: {self.maxFilterSize}"
              )        

    def CreateNativeFactory(self, smlmlib):
        m = SpotDetectionMethods(smlmlib)
        return SpotDetectorNativeFactory(m._SpotDetector_Configure(self), m._SpotDetector_DestroyFactory)


# Currently not implemented on C++ side
class GLRTSpotDetectorConfig(ctypes.Structure):
    _fields = [
        ("falsePositiveRate", ctypes.c_float),
        ("roisize", ctypes.c_int32),
    ]  # Roisize is used to remove ROIs near the image border

    def CreateNativeFactory(self, smlmlib):
        m = SpotDetectionMethods(smlmlib)
        return SpotDetectorNativeFactory(m._GLRTSpotDetector_Configure(self), m._SpotDetector_DestroyFactory)


class SpotDetectionMethods:
    def __init__(self, smlmlib):
        InstancePtrType = ctypes.c_void_p
        self.lib = smlmlib

        self._SpotDetector_Configure = smlmlib.lib.SpotDetector_Configure
        self._SpotDetector_Configure.argtypes = [ctypes.POINTER(SpotDetectorConfig)]
        self._SpotDetector_Configure.restype = ctypes.c_void_p

        self._SpotDetector_DestroyFactory = smlmlib.lib.SpotDetector_DestroyFactory
        self._SpotDetector_DestroyFactory.argtypes = [InstancePtrType]

#CDLL_EXPORT int SpotDetector_ProcessFrame(const float* frame, int width, int height,
#	int maxSpots, float* spotScores, Int2* cornerPos, float* rois, const SpotDetectorConfig & cfg)

        self._SpotDetector_ProcessFrame = smlmlib.lib.SpotDetector_ProcessFrame
        self._SpotDetector_ProcessFrame.argtypes = [
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # frame
            ctypes.c_int32,  # width
            ctypes.c_int32,  # height
            ctypes.c_int32,  # maxspots
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # spotscores
            ctl.ndpointer(np.int32, flags="aligned, c_contiguous"),  # cornerpos
            ctl.ndpointer(np.float32, flags="aligned, c_contiguous"),  # rois [output]
            ctypes.POINTER(SpotDetectorConfig)        
            ]
        self._SpotDetector_ProcessFrame.restype = ctypes.c_int32

    def ProcessFrame(self, image, spotDetectorConfig:SpotDetectorConfig, maxSpotsPerFrame):
        assert len(image.shape)==2
        h = image.shape[0]
        w = image.shape[1]

        image = np.ascontiguousarray(image, dtype=np.float32)

        scores = np.zeros(maxSpotsPerFrame, dtype=np.float32)
        roisize=spotDetectorConfig.roisize
        rois = np.zeros((maxSpotsPerFrame, roisize,roisize),dtype=np.float32)
        cornerpos = np.zeros((maxSpotsPerFrame, 2),dtype=np.int32)
        
        numspots = self._SpotDetector_ProcessFrame(image,w,h,
                                                   maxSpotsPerFrame,scores,cornerpos,rois,spotDetectorConfig)
        rois = rois[:numspots]
        scores = scores[:numspots]
        cornerpos = cornerpos[:numspots]

        return rois, cornerpos, scores
