# -*- coding: utf-8 -*-

import os
import pickle
import tifffile
import numpy as np
from smlmlib.gaussian import Gaussian
from smlmlib.base import SMLM
import smlmlib.spotdetect as spotdetect
import time
import sys
import smlmlib.postprocess as postprocess
import matplotlib.pyplot as plt
from . import drift_estimate
import scipy.spatial.ckdtree


class DataColumns:
    FRAME = 0
    X = 1
    Y = 2
    I = 3
    BG = 4
    CRLB_X = 5
    CRLB_Y = 6
    CRLB_I = 7
    CRLB_BG = 8
    ITERATIONS = 9
    SCORE = 10
    LL = 11
    ROIX = 12
    ROIY = 13
    NUMCOL = 14


class LocResultList:
    # 2D result array, with
    data = np.zeros((0, DataColumns.NUMCOL), dtype=np.float)
    # list of indices into the data array
    frames = []
    roisize = 0
    cfg = None
    raw_image_sum = None
    spot_drift = None
    drift = None
    area = None # x,y,w,h
    filename = None
    extraColumns = None
        
    def __init__(self, data=None, cfg=None, area=None, filename=None, extraColumns=None):
        if not (data is None):
            frame_indices = data[:,DataColumns.FRAME].astype(int)
            if len(frame_indices) == 0: 
                numFrames = 0
            else:
                numFrames = np.max(frame_indices)+1
            self.frames = [[] for i in range(numFrames)]
            for k in range(len(data)):
                self.frames[frame_indices[k]].append(k)
            for f in range(numFrames):
                self.frames[f] = np.array(self.frames[f], dtype=int)
            self.data = data * 1
            self.cfg = dict(cfg)
            self.area = area
            self.filename = filename
            self.imgshape = [area[3],area[2]]
            self.extraColumns = extraColumns

    def __len__(self):
        return len(self.frames)

    def get_area(self):
        if self.area is not None:
            return np.array(self.area)
        return np.array([0, 0, self.raw_image_sum.shape[1], self.raw_image_sum.shape[0]])
    
    def get_image_size(self):
        return self.get_area()[[2,3]]
    
    def get_positions(self):
        positions = []
        for f, indices in enumerate(self.frames):
            positions.append(self.data[indices, 1:5])
        return positions

    def filter(self, fn):
        newdata = []
        spotcount = 0
        frames = []
        selindices = []
        for f, indices in enumerate(self.frames):
            ok = fn(indices, self.data[indices, :])
            okcount = np.sum(ok)
            selindices.append(indices[ok])
            frames.append(np.arange(okcount) + spotcount)
            newdata.append(self.data[indices[ok], :])
            spotcount += okcount
        self.frames = frames
        nremoved = len(self.data) - spotcount
        self.data = np.concatenate(newdata)
        selindices = np.concatenate(selindices)
        if self.extraColumns is not None:
            for k in self.extraColumns.keys():
                print(f"filtering {k} len(seli)={len(selindices)} len(data)={len(self.data)} len(col)={len(self.extraColumns[k])}")
                self.extraColumns[k] = self.extraColumns[k][selindices]
        return nremoved

    def filter_nan(self):
        return self.filter(lambda i,d: ~np.isnan(np.sum(d[:, 1:5], 1)))

    def draw(self, scale, smlm, area=None, fixedSigma=None):
        if area == None:
            area = self.get_area()
        image = np.zeros((np.array([area[3], area[2]]) * scale).astype(int), dtype=np.float64)
        pts = 1 * self.data[:, [DataColumns.X, DataColumns.Y, DataColumns.CRLB_X, DataColumns.CRLB_Y, DataColumns.I]]
        pts[:, [0, 1]] -= [area[0], area[1]]
        pts *= np.tile([scale, scale, scale, scale, 1], (len(pts), 1))
        if fixedSigma is not None:
            pts[:, [2, 3]] = fixedSigma
        image = Gaussian(smlm).Draw(image, pts)
        return image

    # Reads all results from the queue object
    def fetch(self, q):
        cfg = q.GetConfig()

        self.imgshape = (cfg.height, cfg.width)

        nframes = q.GetResultFrameCount()

        result_list = []
        self.frames = []
        spot_count = 0
        for f in range(nframes):
            fr = q.GetFrameResults(f)
            if f % 5 == 0:
                sys.stdout.write(f"\rreading results {f/nframes*100:3.0f}% {f}/{nframes} (#spots={spot_count})")
                sys.stdout.flush()

            if len(fr) > 0:
                result = np.array( [
                        [
                            f,
                            r.estimate[0] + r.roiPosition[0],
                            r.estimate[1] + r.roiPosition[1],
                            r.estimate[2],
                            r.estimate[3],
                            *r.crlb,
                            r.iterations,
                            r.score,
                            r.loglikelihood,
                            *r.roiPosition,
                        ] for r in fr
                    ],
                    dtype=np.float32,
                )
            else:
                result = np.zeros((0, DataColumns.NUMCOL), dtype=np.float32)

            result_list.append(result)
            self.frames.append(np.arange(len(result)) + spot_count)
            spot_count += len(result)
        self.data = np.concatenate(result_list)
        sys.stdout.write(f"\rtotal spot count: {len(self.data)}\t\t\t\n")

    def spot_counts(self):
        return np.array([len(f) for f in self.frames])

    def roisize(self):
        return self.cfg["roisize"]

    def extract_rois(self, tiff_path, spot_indices):
        roisize = self.roisize()
        rois = np.zeros((len(spot_indices), roisize, roisize))
        # map framenumbers so we only load each tiff frame once
        frame_indices = {}
        for i, spot in enumerate(spot_indices):
            f = int(self.data[spot, DataColumns.FRAME])
            if f in frame_indices:
                frame_indices[f].append(i)
            else:
                frame_indices[f] = [i]

        with tifffile.TiffFile(tiff_path) as tif:
            for f, spots in frame_indices.items():
                p = tif.pages[f]
                img = (p.asarray().astype(np.float) - self.cfg["offset"]) / self.cfg["gain"]
                img[img<0] = 0
                for i in spots:
                    roix = int(self.data[spot_indices[i], DataColumns.ROIX])
                    roiy = int(self.data[spot_indices[i], DataColumns.ROIY])
                    rois[i] = img[roiy : roiy + roisize, roix : roix + roisize]

        return rois

    def frame_data(self, framenumbers):
        fd = []
        for f in framenumbers:
            fd.append(self.data[self.frames[f], :])
        return fd

    def drift_estimate_nn(self, smlm, searchDist=2, iterations=2, framebinning=1):
        xyI = self.data[:, [DataColumns.X, DataColumns.Y, DataColumns.I]]
        framenum = self.data[:, DataColumns.FRAME] // framebinning

        drift, matchResults = postprocess.PostProcess(smlm).NearestNeighborDriftEstimate(
            xyI, framenum, self.imgshape, searchDist, iterations
        )

        drift = np.insert(drift, 0, [0, 0], axis=0)
        return drift, matchResults

    def drift_correct_nn(self, smlm, searchDist=2, iterations=3, framebinning=5, plot=None):
        drift, matchResults = self.drift_estimate(smlm, searchDist, iterations, framebinning)
        drift = np.cumsum(drift, 0)
        self.drift = drift
        self.spot_drift = drift[self.get_frame_numbers() // framebinning]
        self.drift_mr = matchResults
        self.data[:, [DataColumns.X, DataColumns.Y]] -= self.spot_drift

        if plot is not None:
            self.plot_drift(plot)

        return drift, matchResults

    # Some simple testing seems to suggest not using result intensities works better.. probably depends on data though
    def drift_estimate_cc(self, smlm, framebinning=5, scale=6, maxdrift=3, useIntensities=False, smooth_sigma=0):
        xyI = self.get_xyI() * 1
        if not useIntensities:
            xyI[:, 2] = 1

        frame_num = self.get_frame_numbers() // framebinning
        drift = drift_estimate.estimate(xyI, frame_num, self.imgshape, smlm, scale=scale, maxdrift=maxdrift)
        drift = np.insert(drift, 0, [0, 0], axis=0)
        if smooth_sigma > 0:
            from . import running_mean

            filtered_drift = 1 * drift
            for ax in [0, 1]:
                filtered_drift[:, ax], _ = running_mean.running_mean(drift[:, ax], sigma=smooth_sigma)
            self.drift = filtered_drift
            self.raw_drift = drift
        else:
            self.drift = drift
            self.raw_drift = None

        return self.drift
    
    def drift_correct_from_file(self,drift_fn, drift_frames_per_frame):
        drift = np.loadtxt(drift_fn)
        print(f"drift file {drift_fn} as {len(drift)} frames")
        drift_indices = (np.arange(len(self.frames)) * drift_frames_per_frame).astype(int)
        drift_indices = np.clip(drift_indices, 0, len(drift)-1)
        self.drift = drift[drift_indices]
        self.spot_drift = self.drift[self.get_frame_numbers()]
        self.data[:, [DataColumns.X, DataColumns.Y]] -= self.spot_drift
        self.raw_drift = None
        return drift
        
    def apply_drift_correction(self, drift, framebinning):
        frame_num = self.get_frame_numbers() // framebinning
        self.drift = drift
        self.spot_drift = drift[frame_num]
        self.data[:, [DataColumns.X, DataColumns.Y]] -= self.spot_drift

    def plot_drift(self, title):
        fig = plt.figure()
        plt.plot(self.drift[:, 0], label="X drift")
        plt.plot(self.drift[:, 1], label="Y drift")
        if self.raw_drift is not None:
            plt.plot(self.raw_drift[:, 0], label="Unfiltered X")
            plt.plot(self.raw_drift[:, 1], label="Unfiltered Y")
        plt.xlabel("Timebins")
        plt.ylabel("Pixels")
        plt.legend()
        plt.title(title)

        if hasattr(self, "drift_mr"):
            plt.figure()
            plt.plot(self.drift_mr, label="Point match counts")
            plt.legend()
            plt.title(title)
        return fig

    def get_crlb(self):
        return self.data[:, [DataColumns.CRLB_X, DataColumns.CRLB_Y, DataColumns.CRLB_I, DataColumns.CRLB_BG]]

    def get_xy(self):
        return self.data[:, [DataColumns.X, DataColumns.Y]]

    def get_xyI(self):
        return self.data[:, [DataColumns.X, DataColumns.Y, DataColumns.I]]

    def get_bg(self):
        return self.data[:, DataColumns.BG]

    def get_xyIBg(self):
        return self.data[:, [DataColumns.X, DataColumns.Y, DataColumns.I, DataColumns.BG]]

    def get_roipos(self):
        return self.data[:, [DataColumns.ROIX, DataColumns.ROIY]].astype(np.int32)

    def get_frame_numbers(self):
        return self.data[:, DataColumns.FRAME].astype(np.int32)

    def get_scores(self):
        return self.data[:, DataColumns.SCORE]

    def process_images(self, cfg, image_iterator, totalframes, smlm):
        self.cfg = cfg
        img0 = next(image_iterator)

        width = img0.shape[1]
        height = img0.shape[0]

        self.imgshape = (height,width)
        self.area = np.array([0, 0, width, height])

        spotDetectConfig = spotdetect.SpotDetectorConfig(
            cfg["sigma"], cfg["roisize"], cfg["detectionMinIntensity"], cfg["detectionMaxIntensity"]
        )

        with Gaussian(smlm).CreateImageQueue(
            width, height, cfg["sigma"], cfg["roisize"], cfg["maxSpotsPerFrame"], spotDetectConfig
        ) as q:
            q.Start()

            raw_image_sum = np.zeros((height, width), dtype=np.float32)
            t0 = time.time()

            q.PushFrame(0, img0)
            numframes = 1

            for index, img in enumerate(image_iterator):
                f = index + 1

                raw_image_sum += img
                while q.GetQueueLength() > 50:
                    time.sleep(0.05)

                q.PushFrame(f, img)
                if numframes > 1 and numframes % 20 == 0:
                    t = time.time()
                    fps = numframes / (t - t0)
                    remaining = (totalframes - numframes) / fps
                    sys.stdout.write(
                        f"\rprocessed {numframes}/{totalframes} frames. Queued: {q.GetQueueLength()}. Time remaining: {remaining:4.0f}s"
                    )
                    sys.stdout.flush()
                numframes += 1

            # waiting for IsIdle is currently unreliable. Sometimes IsIdle=true when not all frames have arrived in the result container
            #            while not q.IsIdle():

            while q.GetResultFrameCount() < numframes:
                #                print(f'Frames processed: {q.GetResultFrameCount()}/{numframes}')
                time.sleep(0.5)

            dt = time.time() - t0
            print(f"\r{numframes} frames in {dt:3.0f} seconds. {numframes/dt:.2f} fps\t\t\t\t\t\t\t\t\t")

            self.raw_image_sum = raw_image_sum
            self.fetch(q)

    def save_csv_with_drift(self, filename):
        if self.spot_drift is None:
            drift = np.zeros((len(self.data), 2))
        else:
            drift = self.spot_drift
        cols = [
            DataColumns.FRAME,
            DataColumns.X,
            DataColumns.Y,
            DataColumns.I,
            DataColumns.BG,
            DataColumns.CRLB_X,
            DataColumns.CRLB_Y,
            DataColumns.CRLB_I,
            DataColumns.CRLB_BG,
            DataColumns.ROIX,
            DataColumns.ROIY,
            DataColumns.ITERATIONS,
            DataColumns.LL,
        ]
        d = np.zeros((len(self.data), len(cols) + 2))
        d[:, : len(cols)] = self.data[:, cols]
        d[:, [-2, -1]] = drift
        np.savetxt(filename, d, delimiter="\t", fmt="%10.5f")
        return d

    def save_csv(self, filename):
        np.savetxt(filename, self.data, delimiter="\t", fmt="%10.5f")

    def view(self):
        import pptk

        def cloud(xyI, col):
            xyz = np.zeros((len(xyI), 3))
            xyz[:, [0, 1]] = xyI[:, 0:2]
            rgb = np.zeros((len(xyI), 3))
            rgb[:, col] = 1
            return xyz, rgb

        a = cloud(self.get_xyI(), 0)

        v = pptk.viewer(a[0], a[1])
        return v

    # Process a TIFF file using the Gaussian image queue, detecting spots and doing localization per frame
    def read_tiff(self, filename, cfg, smlm):
        self.cfg = cfg
        self.filename = filename
        sumframes = cfg["sumframes"] if "sumframes" in cfg else 1
        
        if sumframes > 1:
            print(f"Processing frames in blocks of {sumframes}:")

        with tifffile.TiffFile(filename) as tif:
            p0 = tif.pages[0]
            startframe = 0
            if 'skipfirst' in cfg:
                startframe = cfg['skipfirst']
                
            numframes = (len(tif.pages) - startframe) // sumframes

            def generator():
                imgsum = np.zeros((p0.shape), dtype=np.float32)
                for f in np.arange(startframe, len(tif.pages)):
                    page = tif.pages[f]
                    if "maxframes" in cfg and f == cfg["maxframes"]:
                        raise StopIteration()

                    img = (page.asarray().astype(np.float) - cfg["offset"]) / cfg["gain"]
                    img[img<0] = 0
                    imgsum += img

                    if f % sumframes == sumframes - 1:
                        yield imgsum
                        imgsum = np.zeros((p0.shape), dtype=np.float32)
                        
            self.process_images(cfg, generator(), numframes, smlm)

    def in_roi(self, roi_xy, roi_size=10):
        roi_xy = np.array(roi_xy)
        x = self.data[:, DataColumns.X]
        y = self.data[:, DataColumns.Y]

        indices = np.where(
                (x >= roi_xy[0] - roi_size/2) & (x <= roi_xy[0] + roi_size/2) & 
                (y >= roi_xy[1] - roi_size/2) & (y <= roi_xy[1] + roi_size/2))[0]

        area = [roi_xy[0]-roi_size/2, roi_xy[1]-roi_size/2, roi_size, roi_size]
        r = LocResultList(self.data[indices,:], self.cfg, area)
        return r
    
    def count_neighbors(self, searchdist):
        """
        Return the number of neighbors for each localization
        """
        kdtree = scipy.spatial.ckdtree.cKDTree(self.get_xy())
        pairs = kdtree.query_pairs(searchdist) # output_type='ndarray')
        
        counts = np.zeros((len(self.data)),dtype=np.int)
        for p in pairs:
            counts[p[0]] += 1
        
        return counts
    
    def save_picasso_hdf5(self, fn=None, extra_columns=False):
        import h5py 
        import yaml
        
        if fn is None:
            fn = os.path.splitext(self.filename)[0] + "_g2d.hdf5"
        
        print(f"Saving hdf5 to {fn}")
        with h5py.File(fn, 'w') as f:
            dtype = [('frame', '<u4'), 
                     ('x', '<f4'), ('y', '<f4'), 
                     ('photons', '<f4'), 
                     ('sx', '<f4'), ('sy', '<f4'), 
                     ('bg', '<f4'), 
                     ('lpx', '<f4'), ('lpy', '<f4'), 
                     ('lI', '<f4'), ('lbg', '<f4'), 
                     ('ellipticity', '<f4'), 
                     ('net_gradient', '<f4')]
            
            if extra_columns and self.extraColumns is not None:
                for k in self.extraColumns.keys():
                    dtype.append((k,self.extraColumns[k].dtype,self.extraColumns[k].shape[1:]))
            
            locs = f.create_dataset('locs', shape=(len(self.data),), dtype=dtype)
            locs['frame'] = self.get_frame_numbers()
            locs['x'] = self.get_xyI()[:,0]
            locs['y'] = self.get_xyI()[:,1]
            locs['photons'] = self.get_xyI()[:,2]
            locs['sx'] = self.cfg['sigma']
            locs['sy'] = self.cfg['sigma']
            locs['bg'] = self.data[:, DataColumns.BG]
            locs['lpx'] = self.data[:, DataColumns.CRLB_X]
            locs['lpy'] = self.data[:, DataColumns.CRLB_Y]
            locs['lI'] = self.data[:, DataColumns.CRLB_I]
            locs['lbg'] = self.data[:, DataColumns.CRLB_BG]
            locs['net_gradient'] = self.data[:, DataColumns.SCORE]
            
#            locs['ellipticity']
            
            if extra_columns and self.extraColumns is not None:
                for k in self.extraColumns.keys():
                    locs[k] = self.extraColumns[k] 
                    
            info =  {'Byte Order': '<',
                     'Camera': 'Dont know' ,
                     'Data Type': 'uint16',
                     'File': self.filename,
                     'Frames': len(self.frames),
                     'Width': self.get_area()[2],
                     'Height': self.get_area()[3]}

            info_fn = os.path.splitext(fn)[0] + ".yaml" 
            with open(info_fn, "w") as file:
                yaml.dump(info, file, default_flow_style=False) 

        return fn
    # def export_rois(peaklist)
    #   , width=6, zoom=20, outdir=os.path.split(fn)[0]+'aois/')


def load(path, cfg, smlm, rebuild=False, tag="g2d_result") -> LocResultList:
    dir, file = os.path.split(path)
    ext = f".{tag}.pickle"
    if path.endswith(ext):
        with open(path, "rb") as f:
            result, loaded_cfg = pickle.load(f)
            return result

    file, _ = os.path.splitext(file)
    path_noext, _ = os.path.splitext(path)

    resultsfn = path_noext + ext
    if not rebuild and os.path.exists(resultsfn):
        with open(resultsfn, "rb") as f:
            result, loaded_cfg = pickle.load(f)
            if loaded_cfg == cfg:
                print(f"{path} contains {len(result.data)} spots.")
                result.filename = path
                return result

    result = LocResultList()
    result.read_tiff(path, cfg, smlm)

    with open(resultsfn, "wb") as f:
        pickle.dump((result, cfg), f)

    return result


def read_tiff_frames(filename, frameNumbers):
    with tifffile.TiffFile(filename) as tif:
        print(f"opened {filename}")

        h = tif.pages[0].shape[0]
        w = tif.pages[0].shape[1]

        imgs = np.zeros((len(frameNumbers), h, w), dtype=np.uint16)
        for i, f in enumerate(frameNumbers):
            imgs[i] = tif.pages[f].asarray()

        return imgs


def process_movie(frames, cfg, smlm):
    r = LocResultList()
    r.process_images(cfg, iter(frames), len(frames), smlm)
    return r


# Run spot detection and fitting on a single frame
def process_single_frame(img, cfg, smlm):
    return process_movie([img], cfg, smlm)


def draw_spots(img, x, y, I, sigmaX, sigmaY, scale, smlm):
    # Spots is an array with rows: [ x,y, sigmaX, sigmaY, intensity ]
    img = np.ascontiguousarray(img, dtype=np.float32)

    spots = np.zeros((len(x), 5), dtype=np.float32)
    spots[:, 0] = x * scale
    spots[:, 1] = y * scale
    spots[:, 2] = sigmaX * scale
    spots[:, 3] = sigmaY * scale
    spots[:, 4] = I

    return Gaussian(smlm).Draw(img, spots)


def binlocalizations(imgshape, xyI, scale=4):
    image = np.zeros((imgshape[0] * scale, imgshape[1] * scale), dtype=np.float32)
    shape = image.shape

    x = (xyI[:, 0] * scale + 0.5).astype(int)
    y = (xyI[:, 1] * scale + 0.5).astype(int)
    x = np.clip(x, 0, shape[1] - 1)
    y = np.clip(y, 0, shape[0] - 1)
    image[y, x] += xyI[:, 2]

    return image


def to_array(results):
    return results.data

def localize(tiffpath, roisize, sigma, detectionThreshold=20, gain=1, offset=0,sumframes=None,label=None):
    cfg = {
       'roisize':roisize,
       'maxSpotsPerFrame':2000,
       'detectionMinIntensity':detectionThreshold,
       'detectionMaxIntensity':1e6,
       'offset':offset,
       'gain':gain,
       'sigma': sigma
    }
    
    if sumframes:
        cfg['sumframes'] = sumframes
    
    with SMLM() as smlm:
        results = load(tiffpath, cfg, smlm) 
        results.filter_nan()

        if label is None:
            results.save_picasso_hdf5()
        else:
            results.save_picasso_hdf5(os.path.splitext(tiffpath)[0]+label+".hdf5")
        return results
