"""
SIMFLUX CLI

This interface allows you to process a TIFF file and do the following:
    - Estimate illumination patterns. 
        The pattern estimator assumes there are X and Y patterns but they can be in any order and number.
    - Export HDF5 localization files to view with picasso render 
        (https://github.com/jungmannlab/picasso)
        These files will be stored in the "results" folder that will be created in the same dir as the tiff file.

"""
import os
import sys
sys.path.append("python")
sys.path.append("python/simflux")

import argparse
import numpy as np
import python.simflux.process_tiff as process_tiff

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process TIFF file with SIMFLUX")
    parser.add_argument("input", help="Input TIFF file (.tif)")
    parser.add_argument("--patterns", default=6, help="Set number of modulation patterns. Default 6")
    parser.add_argument("--sigma", default=2, help="PSF width [pixels]. Default 2")
    parser.add_argument("--offset", default=0, help="Camera calibration offset. Default 0")
    parser.add_argument("--gain", default=1, help="Camera calibration gain. Default 1")
    parser.add_argument("--roi", default=10, help="Gaussian fit ROI size. Default 10 pixels")
    parser.add_argument("--threshold", default=10, help="Detection threshold. Defaults to 10.")
#    parser.add_argument("--driftbins", default=100, help="Timebin size for drift correction")    
 #   parser.add_argument("--rerun", default=False, action="store_true", help="Ignore cached results")
    parser.add_argument("--pixelsize", default=65, help="Set pixelsize [nm]. Default 65 nm.")
    parser.add_argument("--minpitch", default=200, help="Set smallest possible pitch [nm]. Default 200 nm")
    parser.add_argument("--maxpitch", default=250, help="Set largest possible pitch [nm]. Default 250 nm")
    parser.add_argument("--minfilter", default=30, help="Set SIMFLUX on/off transition filter [photons], defaults to 30")
    parser.add_argument("--fixdepth", default=None, help="Fix modulation depth instead of estimating. Disabled by default")
    parser.add_argument("--fixshift", default=None, help="Fix phase shifts instead of estimating, in [degrees]. Disabled by default")
    parser.add_argument(
        "-n", "--num-frames", type=int, default=None, help="Number of frames to process. Leave blank to process all"
    )
    
    if len(sys.argv) > 1:
        args = parser.parse_args()
    
        cfg = { 
           'sigma':float(args.sigma),
           'roisize':int(args.roi),
           'maxSpotsPerFrame':2000,
           'detectionMinIntensity':float(args.threshold),
           'detectionMaxIntensity':1e6,
           'offset':float(args.offset),
           'gain':float(args.gain)
           }
        
        if args.num_frames:
            cfg['maxframes'] = args.num_frames
        
        
        dataname = os.path.split(args.input)[1]
        
        pixelsize = float(args.pixelsize)
        maxfreq = 2*np.pi / (float(args.minpitch) / pixelsize)
        minfreq = 2*np.pi / (float(args.maxpitch) / pixelsize)        
                
        fix_depths = None
        if args.fixdepth is not None:
            fix_depths = float(args.fixdepth)

        fixshift = None
        if args.fixshift is not None:
            fixshift = float(args.fixshift)
        
        mod = np.zeros((int(args.patterns),5))
        process_tiff.process(args.input, cfg, mod, 
                             pixelsize=pixelsize,
                             plot_ffts=True, 
                             freq_minmax = [minfreq, maxfreq],
                             minfilter = args.minfilter,
                             fix_depths = fix_depths,
                             fix_phase_shifts = fixshift,
                             )
    else:
        parser.print_help()       
                
            
            
            
            
            