// Model-independent Fisher matrix calculation and Levenberg-Marquardt optimizer (https://www.nature.com/articles/nmeth0510-338)
// All estimators here are for Poisson-distributed samples.
#pragma once

#include "SolveMatrix.h"
#include "Vector.h"
#include <math_constants.h>


template<typename Theta>
struct FisherMatrixType {
	const static int K = Theta::K;
	typedef Vector<typename Theta::TElem, K*K> type;
};

template<typename T>
class SampleOffset_None
{
public:
	PLL_DEVHOST T Apply(T mu, int sampleIndex) const { return mu; }
};

template<typename T>
class SampleOffset_sCMOS
{
public:
	PLL_DEVHOST	T Apply(T mu, int sampleIndex) const {
		return mu + d_vargain2[sampleIndex];
	}
	const float* d_vargain2; // var / g^2
};

template<typename Theta>
struct EstimationResult
{
	template<typename TLevMarResult>
	PLL_DEVHOST EstimationResult(Theta initialValue, TLevMarResult levMarResult, const Theta& crlb, float spotScores=0.0f) :
		estimate(levMarResult.estimate), initialValue(initialValue), crlb(crlb),
		iterations(levMarResult.iterations), spotScores(spotScores) {
	}
	PLL_DEVHOST EstimationResult() { }

	Theta estimate;
	Theta initialValue;
	Theta crlb;
	int iterations;
	Int2 roiPosition;
	float spotScores;
	float loglikelihood; // sum(x log(mu) - mu)
};




template<typename TModel, typename Theta, typename TSampleOffset>
PLL_DEVHOST auto ComputeFisherMatrix(const TModel& model, const TSampleOffset &offset, const Theta& theta, typename Theta::TElem* expectedValue=nullptr)
	-> typename FisherMatrixType<Theta>::type
{
	const int K = model.K;
	typedef typename Theta::TElem T;
	typename FisherMatrixType<Theta>::type fi;

	for (int i = 0; i < K*K; i++)
		fi[i] = 0;
	
	model.ComputeDerivatives([&](int smpIndex, T mu, const T* jacobian) {
		mu = offset.Apply(mu, smpIndex);
		if(expectedValue)
			expectedValue[smpIndex] = mu;

		float inv_mu_c = 1.0f / (mu > 1e-8f ? mu : 1e-8f);
		for (int i = 0; i < K; i++) {
			for (int j = i; j < K; j++) {
				const T fi_ij = jacobian[i] * jacobian[j] * inv_mu_c;
				fi[K*i + j] += fi_ij;
			}
		};
	}, theta);
	// fill below diagonal
	for (int i = 1; i < K; i++)
		for (int j = 0; j < i; j++)
			fi[K*i + j] = fi[K*j + i];

	return fi;
}

template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST void ComputeLevMarAlphaBeta(T* __restrict alpha, T* __restrict beta, const typename TModel::Theta& theta,
	const T* __restrict sample, const TModel& model, const TSampleOffset &offset)
{
	const int K = model.K;
	model.ComputeDerivatives([&](int smpIndex, T mu, const T* jacobian) {
		T sampleValue = sample[smpIndex];
		mu = offset.Apply(mu, smpIndex);

		float mu_c = mu > 1e-6f ? mu : 1e-6f;
		float invmu = 1.0f / mu_c;
		T x_f2 = sampleValue * invmu*invmu;

		for (int i = 0; i < K; i++)
			for (int j = i; j < K; j++)
				alpha[K*i + j] += jacobian[i] * jacobian[j] * x_f2;

		T beta_factor = 1 - sampleValue * invmu;
		for (int i = 0; i < K; i++) {
			beta[i] -= beta_factor * jacobian[i];
		}
	}, theta);

	// fill below diagonal
	for (int i = 1; i < K; i++)
		for (int j = 0; j < i; j++)
			alpha[K*i + j] = alpha[K*j + i];
}



template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST T ComputeLogLikelihood(const typename TModel::Theta& theta, const T* sample, const TModel& model, const TSampleOffset &offset)
{
	T LL = 0;
	const int K = model.K;
	model.ComputeDerivatives([&](int idx, T mu, const T (&jacobian)[K]) {
		T sampleValue = sample[idx];
		mu = offset.Apply(mu, idx);
		float mu_clamped = mu > 1e-8f ? mu : 1e-8f;
		LL += sampleValue * log(mu_clamped) - mu_clamped;
	}, theta);
	return LL;
}


template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST T ComputeChiMLE(const typename TModel::Theta& theta, const T* sample, const TModel& model, const TSampleOffset &offset)
{
	T LL = 0;
	const int K = model.K;
	model.ComputeDerivatives([&](int idx, T mu, const T (&jacobian)[K]) {
		T sampleValue = sample[idx];
		mu = offset.Apply(mu, idx);
		if (sampleValue > 1e-8)
			LL += (mu - sampleValue) - sampleValue * log(mu / sampleValue);
	}, theta);
	return LL;
}


template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST void ComputeExpectedValue(const typename TModel::Theta& theta, const TModel& model, const TSampleOffset &offset, T* expectedValue)
{
	const int K = model.K;
	model.ComputeDerivatives([&](int smpIndex, T mu, const T(&jacobian)[K]) {
		expectedValue[smpIndex] = offset.Apply(mu,smpIndex);
	}, theta);
}

template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST void ComputeDerivatives(const typename TModel::Theta& theta, const TModel& model, const TSampleOffset &offset, T* derivatives, T* expectedValue = 0)
{
	const int K = model.K;
	const int img_size = model.SampleCount();
	model.ComputeDerivatives([&](int smpIndex, T mu, const T(&jacobian)[K]) {
		for (int i = 0; i < K; i++) {
			derivatives[smpIndex + i * img_size] = jacobian[i];
		}
		if(expectedValue) expectedValue[smpIndex] = mu;
	}, theta);
}



template<typename TParams >
struct OptimizerResult
{
	TParams estimate;
	int iterations;
	TParams initialValue;
};

template<typename TModel, typename TSampleOffset>
class LevMarOptimizer {
public:
	typedef typename TModel::T T;
	typedef typename TModel::Theta Theta;

	typename TModel::Theta theta, lastTheta;

	PLL_DEVHOST LevMarOptimizer(Theta initialValue) : 
		theta(initialValue), lastTheta(initialValue)
	{}
	PLL_DEVHOST LevMarOptimizer() {}

	PLL_DEVHOST void Init(Theta t) {
		theta = t;
		lastTheta = t;
	}

	PLL_DEVHOST bool Step(const T* sample, int iteration, const TModel& model, float lambdaStep, TSampleOffset sampleOffset) {
		const int K = model.K;
		T alpha[K*K] = {};
		T LU[K*K] = {};
		T beta[K] = {};
		T thetaStep[K];

		ComputeLevMarAlphaBeta(alpha, beta, theta, sample, model, sampleOffset);

		for (int k = 0; k < K; k++)
			alpha[k*K + k] *= 1 + lambdaStep;

		if (!Cholesky(K, alpha, LU))
			return false;
		if (!SolveCholesky(LU, beta, thetaStep))
			return false;

		bool smallDeltas = true;
		for (int k = 0; k < K; k++) {
			T delta = theta[k] - lastTheta[k];
			if (abs(delta) > model.StopLimit(k)) {
				smallDeltas = false;
				break;
			}
		}

		lastTheta = theta;
		for (int k = 0; k < K; k++)
			theta[k] += thetaStep[k];
		model.CheckLimits(theta, lastTheta);
		if (iteration > 0 && smallDeltas)
			return false;

		return true;
	}
};


template<typename TModel, typename TSampleOffset>
PLL_DEVHOST OptimizerResult <typename TModel::Theta > LevMarOptimize(const typename TModel::T* sample, typename TModel::Theta initialValue, 
	const TModel& model, const TSampleOffset &offset, int maxIterations, typename TModel::Theta* trace=0, int traceBufLen=0, float startLambdaStep=0.1f)
{
	LevMarOptimizer<TModel, TSampleOffset> optimizer(initialValue);

	int i;
	for (i = 0; i < maxIterations; i++) {
		if (i < traceBufLen)
			trace[i] = optimizer.theta;

		if (!optimizer.Step(sample, i,model, startLambdaStep, offset))
			break;
	}
	OptimizerResult<typename TModel::Theta> lr;
	lr.iterations = i;
	lr.estimate = optimizer.theta;
	lr.initialValue = initialValue;
	return lr;
}

template<typename T, typename TModel, typename TSampleOffset>
PLL_DEVHOST OptimizerResult<typename TModel::Theta> 
ComputeMLE(const T* sample, const TModel& model, const TSampleOffset &sampleOffset, int maxIterations, 
	const typename TModel::Theta *initial=0, typename TModel::Theta* trace=0, int traceBufLen=0, float startLambdaStep=0.1f)
{
	typedef typename TModel::Theta Theta;
	OptimizerResult<Theta> r;
	Theta initialValue;

	if (!initial) {
		initialValue = model.ComputeInitialEstimate(sample, 0);
		r = LevMarOptimize(sample, initialValue, model, sampleOffset, maxIterations,
			trace, traceBufLen, startLambdaStep);

		if (model.NumStartPos > 1) {
			auto bestLikelihood = ComputeLogLikelihood(r.estimate, sample, model, sampleOffset);
			for (int i = 1; i < model.NumStartPos; i++) {
				Theta iv_new = model.ComputeInitialEstimate(sample, i);
				OptimizerResult<Theta> r_new = LevMarOptimize(sample, iv_new, model, sampleOffset, maxIterations, 
					trace, traceBufLen, startLambdaStep);
				auto ll = ComputeLogLikelihood(r_new.estimate, sample, model, sampleOffset);
				if (ll > bestLikelihood) {
					bestLikelihood = ll;
					r = r_new;
					initialValue = iv_new;
				}
			}
		}
	}
	else {
		initialValue = *initial;
		r = LevMarOptimize(sample, initialValue, model, sampleOffset, maxIterations, 
			trace, traceBufLen, startLambdaStep);
	}
	return r;
}

// TModel needs to implement ComputeSecondDerivatives() for this to work
template<typename TModel, typename TSampleOffset>
PLL_DEVHOST OptimizerResult<typename TModel::Theta> NewtonRaphson(const typename TModel::T* __restrict sample, typename TModel::Theta initialValue,
	TModel& model, const TSampleOffset &offset, int maxIterations, typename TModel::Theta* trace = 0, int traceBufLen = 0, float stepFactor=0.8f)
{
	const int K = TModel::Theta::K;
	typedef typename TModel::T T;
	typename TModel::Theta theta = initialValue, lastTheta;

	T num[K];
	T denom[K];
	int i;
	for (i = 0; i < maxIterations; i++) {
		if (i < traceBufLen)
			trace[i] = theta;

		for (int j = 0; j < K; j++)
			num[j] = denom[j] = 0.0f;

		model.ComputeSecondDerivatives([&](int smpIndex, T mu, const T(&firstOrder)[K], const T(&secondOrder)[K]) {
			T x = sample[smpIndex];
			T mu_clamped = mu > 1e-8f ? mu : 1e-8f;
			T invMu = 1.0f / mu_clamped;
			T m = x * invMu - 1;
			for (int j = 0; j < K; j++) {
				num[j] += firstOrder[j] * m;
				denom[j] += secondOrder[j] * m - firstOrder[j] * firstOrder[j] * x*invMu*invMu;
			}
		}, theta);
		if (i > 0) {
			bool smallDeltas = true;
			for (int k = 0; k < K; k++) {
				T delta = theta[k] - lastTheta[k];
				if (abs(delta) > model.StopLimit(k)) {
					smallDeltas = false;
					break;
				}
			}
			if (smallDeltas)
				break;
		}

		lastTheta = theta;
		typename TModel::Theta step;
		for (int j = 0; j < K; j++)
			step[j] = num[j] / denom[j];

#ifndef __CUDA_ARCH__
		//DebugPrintf("iteration %d: ",i);
		//PrintVector(step);
#endif

		theta -= step*stepFactor;
		model.CheckLimits(theta, lastTheta);
	}
	OptimizerResult<typename TModel::Theta> lr;
	lr.iterations = i;
	lr.estimate = theta;
	return lr;
}


template<typename T, int KK> PLL_DEVHOST Vector<T, CompileTimeSqrt(KK)> ComputeCRLB(const Vector<T, KK>& fisher)
{
	const int K = CompileTimeSqrt(KK);
	Vector<T,KK> inv;
	Vector<T, K> crlb;
	if (InvertMatrix<T,KK>(fisher, inv))
	{
		for (int i = 0; i < K; i++)
			crlb[i] = sqrt(fabs(inv[i*(K + 1)]));
	}
	else {
		for (int i = 0; i < K; i++)
			crlb[i] = INFINITY;
	}
	return crlb;
}

// LM optimization at CPU side with a dynamic theta size
template<typename TModel, typename TSampleOffset>
class LevMarOptimizerDynamic {
public:
	typedef typename TModel::T T;
	typedef typename TModel::Theta Theta;

	typename TModel::Theta theta, lastTheta;
	const TModel& model;
	std::vector<T> alpha, LU, beta, thetaStep, solveTemp;
	int iteration = 0;

	LevMarOptimizerDynamic(const TModel& model, Theta initialValue) :
		theta(initialValue), lastTheta(initialValue), model(model), alpha(model.K*model.K), 
		LU(model.K*model.K), beta(model.K), thetaStep(model.K), solveTemp(model.K)
	{}

	void ComputeLevMarAlphaBeta(const Theta& theta, const T* sample, const TSampleOffset &offset)
	{
		const int K = model.K;
		for (int i = 0; i < K*K; i++)
			alpha[i] = 0.0f;
		for (int i = 0; i < K; i++)
			beta[i] = 0.0f;

		model.ComputeDerivatives([&](int smpIndex, T mu, const T* jacobian) {
			T sampleValue = sample[smpIndex];
			mu = offset.Apply(mu, smpIndex);

			float mu_c = mu > 1e-6f ? mu : 1e-6f;
			float invmu = 1.0f / mu_c;
			T x_f2 = sampleValue * invmu*invmu;

			for (int i = 0; i < K; i++)
				for (int j = i; j < K; j++)
					alpha[K*i + j] += jacobian[i] * jacobian[j] * x_f2;

			T beta_factor = 1 - sampleValue * invmu;
			for (int i = 0; i < K; i++) {
				beta[i] -= beta_factor * jacobian[i];
			}
		}, theta);

		// fill below diagonal
		for (int i = 1; i < K; i++)
			for (int j = 0; j < i; j++)
				alpha[K*i + j] = alpha[K*j + i];
	}


	bool Step(float* alpha, const float* beta, float lambdaStep) {
		const int K = model.K;

		for (int k = 0; k < K; k++)
			alpha[k*K + k] *= 1 + lambdaStep;

		Cholesky(K, &alpha[0], &LU[0]);
		SolveCholesky(K, &LU[0], &beta[0], &thetaStep[0], &solveTemp[0]);

		bool smallDeltas = true;
		for (int k = 0; k < K; k++) {
			T delta = theta[k] - lastTheta[k];
			if (abs(delta) > model.StopLimit(k)) {
				smallDeltas = false;
				break;
			}
		}

		if (iteration > 0 && smallDeltas)
			return false;

		lastTheta = theta;
		for (int k = 0; k < K; k++)
			theta[k] += thetaStep[k];
		model.CheckLimits(theta, lastTheta);
		iteration++;
		return true;
	}

	bool Step(const T* sample, float lambdaStep, TSampleOffset sampleOffset) {
		ComputeLevMarAlphaBeta(theta, sample, sampleOffset);
		return Step(&alpha[0], &beta[0], lambdaStep);
	}

};
