
#include "MemLeakDebug.h"
#include "StringUtils.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <chrono>
#include <memory>
#include <cassert>

#include "ImageProcessingQueue.h"
#include "ContainerUtils.h"
#include "ThreadUtils.h"

ImageQueueConfig::ImageQueueConfig(int w, int h, int maxFrameQueueLen, int historyLen, int cudaStreams)
	: imageWidth(w), imageHeight(h), maxQueueLen(maxFrameQueueLen), historyLength(historyLen), cudaStreamCount(cudaStreams)
{
}


ImageProcessingQueueBase::ImageProcessingQueueBase(const ImageQueueConfig& config, ResultContainerBase& results) :
	config(config), stopProcessing(false), calibration(0)
{
	activeStreams = 0;
	processingThread = 0;
	frameCopyThread = 0;
	currentCopyFrame = -1;
	frameStep = 1;

	streams.resize(config.cudaStreamCount);
	for (int i = 0; i < config.cudaStreamCount; i++) {
		auto& sd = streams[i];
		cudaStreamCreate(&sd.stream);
		// otherwise cudaStreamQuery will fail
		EmptyKernel(sd.stream);

		cudaEventCreate(&sd.startEvent);
		cudaEventCreate(&sd.stopEvent);
	}

	cudaStreamCreate(&frameCopyStream);
#ifdef _DEBUG
	verbose = false;
#else
	verbose = false;
#endif
}

ImageProcessingQueueBase::~ImageProcessingQueueBase()
{
	Stop();

	if (calibration)
		delete calibration;

	for (auto t : tasks) 
		delete t.task;

	for (auto s : streams) {
		cudaEventDestroy(s.startEvent);
		cudaEventDestroy(s.stopEvent);
		cudaStreamDestroy(s.stream);
	}

	cudaStreamDestroy(frameCopyStream);
}



DLL_EXPORT void ImageProcessingQueueBase::SetCalibration(IImageProcessor * calib)
{
	calibration = calib;
}

ImageProcessingQueueBase::Frame::Frame() {
	frame = -1;
	src = 0;
	processed = false;
	datatype = FDT_Float;
	perFrameData = nullptr;
}
ImageProcessingQueueBase::Frame::~Frame() {
	if (src) delete[] src;
}

// Default implementation does not parse tag data, up to subclasses of ImageProcessingQueueBase
void ImageProcessingQueueBase::PushFrame(int frame, const void* data, FrameDataType type, void* frameTagData = 0, int tagDataSize = 0)
{
	PushFrame(frame, data, type, (IPerFrameData*)nullptr);
}

int ImageProcessingQueueBase::GetStreamState(int * framePerStream, int * streamTaskIndex)
{
	LockedAction(streamStateMutex, [&]() {
		for (int i = 0; i < streams.size(); i++) {
			framePerStream[i] = streams[i].currentFrame;
			streamTaskIndex[i] = streams[i].taskIndex;
		}
	});
	return ActiveStreams();
}

DLL_EXPORT void ImageProcessingQueueBase::PushFrame(int frameIdx, const void * data, FrameDataType type, IPerFrameData* perFrameData)
{
	auto it = LockedFunction(framesMutex, [&]() {
		return frames.insert(std::pair<int, Frame>(frameIdx, Frame())); 
	});
	assert(it.second);

	Frame& fr = it.first->second;

	fr.processed = false;
	fr.datatype = type;
	fr.frame = frameIdx;
	fr.perFrameData = perFrameData;
	int elemsize;
	if (type == FDT_Float) elemsize = 4;
	if (type == FDT_Short) elemsize = 2;
	if (type == FDT_Byte) elemsize = 1;

	int len = elemsize * config.imageWidth*config.imageHeight;
	fr.src = new uint8_t[len];
	memcpy(fr.src, data, len);

	LockedAction(frameCopyMutex, [&]() { frameCopyQueue.push_back(frameIdx); });

	if (verbose)
		DebugPrintf("pushed frame %d\n", frameIdx);
}



void ImageProcessingQueueBase::Stop()
{
	stopProcessing = true;
	if (processingThread) {
		processingThread->join();
		delete processingThread;
	}

	if (frameCopyThread) {
		frameCopyThread->join();
		delete frameCopyThread;
	}
}

void ImageProcessingQueueBase::AddTask(IComputeTask * task)
{
	Task t = {};
	t.task = task;
	tasks.push_back(t);
}

DLL_EXPORT int ImageProcessingQueueBase::NumTasks()
{
	return (int)tasks.size();
}

DLL_EXPORT float ImageProcessingQueueBase::GetAverageTaskRunTime(int taskIndex)
{
	return (float)(tasks[taskIndex].timeSpent / tasks[taskIndex].runs);
}

IComputeTask* ImageProcessingQueueBase::GetTask(int i)
{
	return tasks[i].task;
}

const DeviceImage<float>* ImageProcessingQueueBase::GetFrame(int frameIdx)
{
	return LockedFunction(framesMutex, [&]() -> const DeviceImage<float>*{
		auto it = frames.find(frameIdx);
		if (it != frames.end()) {
			return &it->second.d_img;
		}
		return 0;
	});
}


const DeviceImage<float>& ImageProcessingQueueBase::GetImage(int streamIndex)
{
	auto img = GetFrame(streams[streamIndex].currentFrame);
	assert(img);
	return *img;
}

void ImageProcessingQueueBase::Start()
{
	if (processingThread)
		return;

	processingThread = new std::thread([&]() {
		ProcessingThreadMain(); 
	});

	frameCopyThread = new std::thread([&]() {
		FrameCopyThreadMain();
	});
}

void ImageProcessingQueueBase::FrameCopyThreadMain()
{
	while (!stopProcessing) {
		// is the current task finished?
		if (currentCopyFrame >= 0 && cudaStreamQuery(frameCopyStream) == cudaSuccess)
		{
			LockedAction(frameProcessMutex, [&]() { 
				if (currentCopyFrame % frameStep == 0) {
					frameProcessQueue.push_back(currentCopyFrame);
					if (verbose)
						DebugPrintf("copied frame %d, rdy for processing\n", currentCopyFrame);
				}
			});
			currentCopyFrame = -1;
		}

		// is there stuff to do, and is the stream idle?
		if (currentCopyFrame < 0)
		{
			int newFrame = -1;
			LockedAction(frameCopyMutex, [&]() {
				if (!frameCopyQueue.empty()) {
					newFrame = frameCopyQueue.front();
					frameCopyQueue.pop_front();
				}});

			if (newFrame >= 0) {
				decltype(frames.begin()) f;
				auto hasFrame = LockedFunction(framesMutex, [&]() { 
					f = frames.find(newFrame); 
					return f != frames.end(); 
				});
				if (hasFrame) {
					CopyFrameData(f->second);
					currentCopyFrame = newFrame;
				}
			}
		}

		if (currentCopyFrame < 0) {
			std::this_thread::sleep_for(std::chrono::milliseconds(1));
		}
	}
}

void ImageProcessingQueueBase::CopyFrameData(Frame& f)
{
	int n = config.imageWidth*config.imageHeight;
	if (f.h_img.size() != n)
		f.h_img.Init(n);
	auto copy = [&](auto d) {		
		for (int i = 0; i < n; i++) 
			f.h_img[i] = d[i];
	};
	if (f.datatype == FDT_Float) copy((float*)f.src);
	if (f.datatype == FDT_Byte) copy((uint8_t*)f.src);
	if (f.datatype == FDT_Short) copy((uint16_t*)f.src);

	if (f.d_img.NumPixels() != n)
		f.d_img.Init(config.ImageSize());

	f.d_img.CopyFromHost(f.h_img.data(), frameCopyStream);
	if (calibration)
	{
		calibration->ProcessImage(f.d_img, frameCopyStream);
	}

	if (verbose)
		DebugPrintf("Copied frame %d data.\n", f.frame);
}

bool ImageProcessingQueueBase::AllStreamsIdle()
{
	return activeStreams == 0 && currentCopyFrame < 0;
}

void ImageProcessingQueueBase::ProcessingThreadMain()
{
	int lastFrame = -1;
	int processedUpTo = -1;

	auto frameDone = [&](int streamIndex) {
		StreamData& sd = streams[streamIndex];
		LockedAction(framesMutex, [&]() {
			frames[sd.currentFrame].processed = true;

			do {
				auto next = frames.find(processedUpTo + 1);
				if (next == frames.end() || !next->second.processed)
					break;
				processedUpTo++;
			} while (true);

			int deleteUpTo = processedUpTo - config.historyLength + 1;
			while (lastFrame < deleteUpTo) {
				lastFrame++;
				frames.erase(frames.find(lastFrame));
			}
		});
		sd.currentFrame = -1;
	};

	auto executeTask = [&](int streamIndex) {
		auto& sd = streams[streamIndex];
		cudaEventRecord(sd.startEvent, sd.stream);
		if (verbose)
			DebugPrintf("starting task %d on frame %d, stream %d\n", sd.taskIndex, sd.currentFrame, streamIndex);
		if (!tasks[sd.taskIndex].task->Execute(this, streamIndex)) {
			activeStreams--;
			frameDone(streamIndex);
		}
		cudaEventRecord(sd.stopEvent, sd.stream);
	};

	auto assignFrame = [&](int streamIndex) {
		int f = -1;
		LockedAction(frameProcessMutex, [&]() {
			if (!frameProcessQueue.empty()) {
				activeStreams++;
				f = frameProcessQueue.front();
				frameProcessQueue.pop_front();
			}
		});

		if (f >= 0) {
			auto& sd = streams[streamIndex];
			sd.taskIndex = 0;
			sd.currentFrame = f;
			executeTask(streamIndex);
		}
	};

	while (!stopProcessing)
	{
		int idleCount = 0;

		LockedAction(streamStateMutex, [&]() {
			for (int streamIndex = 0; streamIndex < streams.size(); streamIndex++) {
				StreamData& sd = streams[streamIndex];

				// If no frame assigned or the last task was completed
				if (sd.currentFrame < 0 || cudaStreamQuery(sd.stream) == cudaSuccess)
				{
					// stream has finished processing a task
					if (sd.currentFrame >= 0)
					{
						Task& t = tasks[sd.taskIndex];
						if (verbose)
							DebugPrintf("completed task %d of frame %d on stream %d\n", sd.taskIndex, sd.currentFrame, streamIndex);
						float ms;
						cudaEventElapsedTime(&ms, sd.startEvent, sd.stopEvent);
						t.timeSpent += ms;
						t.runs++;
						bool abort = !t.task->Completed(this, streamIndex);
						sd.taskIndex++;

						if (abort || sd.taskIndex == tasks.size()) {
							frameDone(streamIndex);
							activeStreams--;
						}
					}

					// Is there a frame assigned: execute the next task
					if (sd.currentFrame >= 0) {
						executeTask(streamIndex);
					}
					else {
						assignFrame(streamIndex);
					}

					if (sd.currentFrame < 0) {
						idleCount++;
					}
				}
			}
		});
		if (idleCount == 0)
		{
			std::this_thread::sleep_for(std::chrono::milliseconds(1));
		}
	}
}

int ImageProcessingQueueBase::GetQueueLength(int* maxQueueLen)
{
	if (maxQueueLen) *maxQueueLen = config.maxQueueLen;
	return LockedFunction(frameProcessMutex, [&]() {return (int)frameProcessQueue.size(); }) +
		LockedFunction(frameCopyMutex, [&]() { return (int)frameCopyQueue.size(); });
}

void ImageProcessingQueueBase::Flush()
{
	while (!IsIdle()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(5));
	}
}

bool ImageProcessingQueueBase::IsIdle()
{
	return GetQueueLength() == 0 && AllStreamsIdle();
}

int ImageProcessingQueueBase::ActiveStreams()
{
	return activeStreams;
}

int ImageProcessingQueueBase::GetNumFrames()
{
	return LockedFunction(framesMutex, [&]() {return (int)frames.size(); });
}

DLL_EXPORT void ImageProcessingQueueBase::DeleteResults(int startFrame, int endFrameIncl)
{
	GetResultContainer()->DeleteResults(startFrame, endFrameIncl);
}


ResultContainerBase::~ResultContainerBase()
{
}

std::unique_ptr<std::lock_guard<std::mutex>> ResultContainerBase::GetResultLock()
{
	return std::move(std::make_unique<std::lock_guard<std::mutex>>(resultWriteMutex));
}

CDLL_EXPORT int IPQ_IsIdle(ImageProcessingQueueBase *q)
{
	return q->IsIdle() ? 1 : 0;
}

CDLL_EXPORT void IPQ_SetFrameStep(ImageProcessingQueueBase* q, int frameStep)
{
	q->SetFrameStep(frameStep);
}

CDLL_EXPORT void IPQ_Destroy(ImageProcessingQueueBase * q)
{
	delete q;
}

CDLL_EXPORT void IPQ_PushFrameF32(ImageProcessingQueueBase * q, int frame,const float * imgptr, uint8_t *tagData, int tagDataSize)
{
	q->PushFrame(frame, imgptr, ImageProcessingQueueBase::FDT_Float, tagData, tagDataSize);
}

CDLL_EXPORT void IPQ_PushFrameU16(ImageProcessingQueueBase * q, int frame, const uint16_t* imgptr, uint8_t *tagData, int tagDataSize)
{
	q->PushFrame(frame, imgptr, ImageProcessingQueueBase::FDT_Short, tagData, tagDataSize);
}

CDLL_EXPORT void IPQ_PushFrameU8(ImageProcessingQueueBase * q, int frame, const uint8_t * imgptr, uint8_t *tagData, int tagDataSize)
{
	q->PushFrame(frame, imgptr, ImageProcessingQueueBase::FDT_Byte, tagData, tagDataSize);
}



CDLL_EXPORT void IPQ_Flush(ImageProcessingQueueBase * q)
{
	q->Flush();
}

CDLL_EXPORT int IPQ_GetQueueLength(ImageProcessingQueueBase * q, int * maxQueueLen)
{
	int l = q->GetQueueLength(maxQueueLen);
	return l;
}


CDLL_EXPORT void IPQ_Start(ImageProcessingQueueBase * q)
{
	q->Start();
}

CDLL_EXPORT void IPQ_Stop(ImageProcessingQueueBase * q)
{
	q->Stop();
}

CDLL_EXPORT int IPQ_GetResultFrameCount(ImageProcessingQueueBase * q)
{
	return q->GetResultContainer()->GetFrameCount();
}

CDLL_EXPORT int IPQ_GetFramesInMemory(ImageProcessingQueueBase * q)
{
	return q->GetNumFrames();
}

CDLL_EXPORT int IPQ_GetResultCount(ImageProcessingQueueBase * q, int frame)
{
	return q->GetResultContainer()->GetResultCount(frame);
}

CDLL_EXPORT void IPQ_DeleteResults(ImageProcessingQueueBase* q, int startFrame, int endFrameIncl)
{
	q->GetResultContainer()->DeleteResults(startFrame, endFrameIncl);
}

CDLL_EXPORT void IPQ_GetConfig(ImageProcessingQueueBase* q, ImageQueueConfig& iqc)
{
	iqc = q->GetConfig();
}

CDLL_EXPORT int IPQ_GetStreamState(ImageProcessingQueueBase * q, int * frames, int * tasks)
{
	return q->GetStreamState(frames, tasks);
}

CDLL_EXPORT void IPQ_GetResults(ImageProcessingQueueBase * q, int frame, void * data, int nbytes)
{
	q->GetResultContainer()->CopyResults(frame, data, nbytes);
}

CDLL_EXPORT void ctypes_test(ImageProcessingQueueBase* inst, int f, float* a)
{

}

