#pragma once

#include <map>
#include <list>
#include <mutex>
#include <deque>
#include <unordered_map>
#include <atomic>

#include "CudaUtils.h"
#include "ContainerUtils.h"
#include "DLLMacros.h"



class IComputeTask;

struct ImageQueueConfig
{
	DLL_EXPORT ImageQueueConfig(int w=1024, int h=512, int maxFrameQueueLen=20, int historyLen=1, int cudaStreams=3);

	int imageWidth, imageHeight;
	int maxQueueLen;
	int historyLength;  // how many frames do we keep in memory for IComputeTask's to read
	int cudaStreamCount;

	int2 ImageSize() const { return int2{ imageWidth,imageHeight }; }
};

class ResultContainerBase
{
public:
	virtual ~ResultContainerBase();

	virtual int GetFrameCount() = 0;
	virtual int GetResultCount(int frame) = 0;
	virtual void CopyResults(int frame, void* dstbuf, size_t nbytes) = 0;
	virtual void DeleteResults(int startFrame, int endFrameIncl) = 0;

	DLL_EXPORT std::unique_ptr<std::lock_guard<std::mutex> > GetResultLock();

protected:
	std::mutex resultWriteMutex;
};

template<typename TResult>
class ResultContainer : public ResultContainerBase
{
public:
	struct FrameResults
	{
		std::vector<TResult> results;
	};

	~ResultContainer() { 
	}
	int GetFrameCount() {
		auto l = GetResultLock();
		return (int)frames.size(); 
	}
	int GetResultCount(int f) {
		auto l = GetResultLock();
		auto it = frames.find(f);
		if (it == frames.end())
			return -1;
		return (int)it->second.results.size();
	}
	void AddResults(int f, std::vector<TResult>&& results)
	{
		auto l = GetResultLock();
		frames[f] = { results };
	}
	const FrameResults* GetResults(int f)
	{
		auto l = GetResultLock();
		auto it = frames.find(f);
		if (it == frames.end())
			return 0;
		return &it->second;
	}
	void CopyResults(int f, void *dst, size_t nbytes)
	{
		auto r = GetResults(f);
		if (r) {
			if (r->results.size() * sizeof(TResult) != nbytes)
				throw std::exception("invalid buffer size");
			for (size_t i = 0; i < r->results.size(); i++)
				((TResult*)dst)[i] = r->results[i];
		}
	}
	void DeleteResults(int startFrame, int endFrameIncl)
	{
		auto l = GetResultLock();
		for (int i = startFrame; i <= endFrameIncl; i++)
		{
			auto it = frames.find(i);
			if (it != frames.end()) frames.erase(it);
		}
	}


protected:
	std::map<int, FrameResults> frames;
};

// Image source keeps track of 1 image per stream
class IImageSource
{
public:
	virtual const DeviceImage<float>& GetImage(int streamIndex) = 0;
	virtual int2 GetImageSize() = 0;
};

// Frame source keeps track of a indexable set of frames. Returns zero for non-existent frames
class IFrameSource
{
public:
	virtual const DeviceImage<float>* GetFrame(int frame) = 0;
	virtual int2 GetImageSize() = 0;
};


class IImageProcessor
{
public:
	virtual void ProcessImage(DeviceImage<float>& img, cudaStream_t stream) = 0;
	virtual void Completed() = 0;
	virtual ~IImageProcessor() {}
};

class IComputeTaskManager {
public:
	virtual int GetFrameIndex(int streamIndex) = 0;// TODO: remove this so compute tasks can span data from multiple frames
	virtual cudaStream_t GetStream(int streamIndex) = 0;
};

class IPerFrameData {
	virtual ~IPerFrameData() = 0;
};

class ImageProcessingQueueBase : public IFrameSource, public IImageSource, public IComputeTaskManager
{
protected:
	struct Frame;

public:
	ImageProcessingQueueBase(const ImageQueueConfig& config, ResultContainerBase& results);
	virtual ~ImageProcessingQueueBase();

	// ImageProcessingQueue will delete the calibration object
	DLL_EXPORT void SetCalibration(IImageProcessor* calib);

	enum FrameDataType { FDT_Float=0, FDT_Short=1, FDT_Byte =2};

	void PushFrame(int frame, const float* imageData) { PushFrame(frame, imageData, FDT_Float); }
	void PushFrame(int frame, const uint16_t* imageData) { PushFrame(frame, imageData, FDT_Short); }
	void PushFrame(int frame, const uint8_t* imageData) { PushFrame(frame, imageData, FDT_Byte); }

	DLL_EXPORT int GetStreamState(int* framePerStream, int* streamTaskIndex);
	DLL_EXPORT void PushFrame(int frame, const void* data, FrameDataType type, IPerFrameData* perFrameData=0);

	// Subclasses can implement the virtual function below, and call the PushFrame with queue-specific perFrameData
	virtual void PushFrame(int frame, const void* data, FrameDataType type, void* frameTagData, int tagDataSize);

	DLL_EXPORT void Flush();
	DLL_EXPORT int GetQueueLength(int* maxQueueLen = 0);
	DLL_EXPORT bool IsIdle();
	DLL_EXPORT int ActiveStreams();
	DLL_EXPORT int GetNumFrames();

	DLL_EXPORT void DeleteResults(int startFrame, int endFrameIncl);

	DLL_EXPORT void Start();
	DLL_EXPORT void Stop();

	DLL_EXPORT void AddTask(IComputeTask* task);
	DLL_EXPORT int NumTasks();
	DLL_EXPORT float GetAverageTaskRunTime(int taskIndex);
	DLL_EXPORT IComputeTask* GetTask(int i);

	// IComputeTaskManager
	cudaStream_t GetStream(int idx) { return streams[idx].stream; }

	// IFrameSource
	// Not thread safe! Can only be called from frame processing thread
	const DeviceImage<float>* GetFrame(int frameIdx);  

	// IImageSource
	// Not thread safe! Can only be called from frame processing thread
	const DeviceImage<float>& GetImage(int stream);

	int2 GetImageSize() { return config.ImageSize(); }
	int GetFrameIndex(int streamIdx) { return streams[streamIdx].currentFrame; }
	void SetFrameStep(int frameStep) { this->frameStep = frameStep; }

	virtual ResultContainerBase* GetResultContainer() = 0;

	ImageQueueConfig GetConfig() { return config; }

protected:

	struct Frame {
		Frame();
		~Frame();

		int frame;
		uint8_t* src;
		PinnedArray<float> h_img;
		DeviceImage<float> d_img;
		bool processed;
		FrameDataType datatype;
		IPerFrameData* perFrameData;
	};

	IImageProcessor* calibration;
	ImageQueueConfig config;

	std::mutex frameCopyMutex;
	std::deque<int> frameCopyQueue;

	std::mutex frameProcessMutex;
	std::deque<int> frameProcessQueue;

	std::mutex framesMutex;
	std::unordered_map<int, Frame> frames;

	struct StreamData
	{
		StreamData() { stream = 0; currentFrame = -1; }
		cudaStream_t stream;
		int currentFrame;
		int taskIndex;

		cudaEvent_t startEvent, stopEvent; // measuring current task runtime
	};

	std::vector<StreamData> streams;
	std::mutex streamStateMutex;
	cudaStream_t frameCopyStream;

	struct Task {
		Task() { task = 0; timeSpent = 0.0; runs = 0; }

		IComputeTask *task;
		double timeSpent;
		int runs;
	};

	std::vector<Task> tasks;
	std::thread* processingThread, *frameCopyThread;
	std::atomic<int> activeStreams;
	std::atomic<int> currentCopyFrame;
	volatile bool stopProcessing;

	bool verbose;
	int frameStep; // only process frames that have frame%frameStep==0

	void ProcessingThreadMain();
	void FrameCopyThreadMain();
	void CopyFrameData(Frame& f);
	bool AllStreamsIdle();
};

template<typename TResult>
class ImageProcessingQueue : public ImageProcessingQueueBase {
public:
	ImageProcessingQueue(const ImageQueueConfig& config) :
		ImageProcessingQueueBase(config, resultContainer) {}

	ResultContainerBase* GetResultContainer() { return &resultContainer; }

	ResultContainer<TResult> resultContainer;
};


class IComputeTask {
public:
	IComputeTask() {}
	virtual ~IComputeTask() {}

	// If false, skip the rest of the tasks
	virtual bool Execute(IComputeTaskManager* tm, int streamIndex) = 0;

	// If false, skip the rest of the tasks
	virtual bool Completed(IComputeTaskManager* tm, int streamIndex) { return true; }
	virtual const char* GetTaskName() { return "Unnamed task"; }
};


CDLL_EXPORT void IPQ_Destroy(ImageProcessingQueueBase* q);
CDLL_EXPORT void IPQ_PushFrameF32(ImageProcessingQueueBase* q, int frame, const float* imgptr, uint8_t *tagData, int tagDataSize);
CDLL_EXPORT void IPQ_PushFrameU16(ImageProcessingQueueBase* q, int frame, const uint16_t* imgptr, uint8_t *tagData, int tagDataSize);
CDLL_EXPORT void IPQ_PushFrameU8(ImageProcessingQueueBase* q, int frame, const uint8_t* imgptr, uint8_t *tagData, int tagDataSize);
CDLL_EXPORT void IPQ_Flush(ImageProcessingQueueBase* q);
CDLL_EXPORT int IPQ_GetQueueLength(ImageProcessingQueueBase* q, int* maxQueueLen = 0);

CDLL_EXPORT void IPQ_Start(ImageProcessingQueueBase* q);
CDLL_EXPORT void IPQ_Stop(ImageProcessingQueueBase* q);

CDLL_EXPORT int IPQ_GetFramesInMemory(ImageProcessingQueueBase* q);
CDLL_EXPORT int IPQ_GetResultFrameCount(ImageProcessingQueueBase* q);
CDLL_EXPORT int IPQ_GetResultCount(ImageProcessingQueueBase* q, int frame);
CDLL_EXPORT void IPQ_GetResults(ImageProcessingQueueBase* q, int frame, void* data, int nbytes);
CDLL_EXPORT void IPQ_DeleteResults(ImageProcessingQueueBase* q, int startFrame, int endFrameIncl);
CDLL_EXPORT void IPQ_GetConfig(ImageProcessingQueueBase* q, ImageQueueConfig& iqc);
CDLL_EXPORT int IPQ_GetStreamState(ImageProcessingQueueBase* q, int *frames, int *tasks);
CDLL_EXPORT int IPQ_IsIdle(ImageProcessingQueueBase *q);
CDLL_EXPORT void IPQ_SetFrameStep(ImageProcessingQueueBase* q, int frameStep);
