#pragma once

#include "Vector.h"
#include "MathUtils.h"
#include "palala.h"
#include "GaussianPSF.h"

typedef Vector5f Theta_XYZIBg;
typedef Vector4f Theta_XYIBg;
typedef Vector5f Theta_XYIBgSigma;




template<typename T>
PLL_DEVHOST Vector3f ComputeCOM(const T* sample, int w)
{
	// Compute initial value
	T sumX = 0, sumY = 0, sum = 0;
	for (int y = 0; y < w; y++)
		for (int x = 0; x < w; x++) {
			T v = sample[y * w + x];
			sumX += x * v;
			sumY += y * v;
			sum += v;
		}

	float comx = sumX / sum;
	float comy = sumY / sum;
	return { comx,comy, sum };
}


template<typename TNumber>
struct Gauss2D_Model_XYIBg
{
	typedef typename TNumber T;
	typedef Vector4f Theta;

	enum { K = Theta::K };
	enum { NumStartPos = 1 };
	enum { NumConstants = 0 };

	T sigma;
	int imgw;

	static const char* ThetaFormat() { return "x,y,I,bg"; }

	PLL_DEVHOST int SampleCount() const { return imgw * imgw; }

	static PLL_DEVHOST T StopLimit(int k)
	{
		const float deltaStopLimit[]={ 1e-4f, 1e-4f, 1e-1f,1e-4f };
		return deltaStopLimit[k];
	}

	PLL_DEVHOST Gauss2D_Model_XYIBg(int roisize, float sigma, const float* unused_constants=0) : imgw(roisize), sigma(sigma) {}

	PLL_DEVHOST void CheckLimits(Theta& t, const Theta& last) const {
		t.elem[0] = clamp(t.elem[0], 2.0f, imgw - 1.0f);
		t.elem[1] = clamp(t.elem[1], 2.0f, imgw - 1.0f);

		if (t.elem[2] < 10.0f) t.elem[2] = 10.0f;
		if (t.elem[3] < 0.0f) t.elem[3] = 0.0f;
		//	if (t.elem[3] > t.elem[2]) t.elem[3] = t.elem[2];
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeDerivatives(TCallback cb, const Theta& theta) const
	{
		const T OneOverSqrt2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * sigma);
		const T OneOverSqrt2Sigma = 1.0f / (sqrtf(2) * sigma);

		T tx = theta[0];
		T ty = theta[1];
		T tI = theta[2];
		T tbg = theta[3];

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			T dEy = OneOverSqrt2PiSigma * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				T dEx = OneOverSqrt2PiSigma * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));

				T mu = tbg + tI * DeltaX * DeltaY;
				T dmu_dx = tI * dEx * DeltaY;
				T dmu_dy = tI * dEy * DeltaX;
				T dmu_dI0 = DeltaX * DeltaY;
				T dmu_dIbg = 1;
				const T jacobian[] = { dmu_dx, dmu_dy, dmu_dI0, dmu_dIbg };
				cb(y*imgw + x, mu, jacobian);
			}
		}
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeSecondDerivatives(TCallback cb, Theta theta) const
	{
		const T OneOverSqrt2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * sigma);
		const T OneOverSqrt2Sigma = 1.0f / (sqrtf(2) * sigma);
		const T OneOverSqrt2PiSigma3 = 1.0f / (sqrtf(2 * MATH_PI) * sigma*sigma*sigma);

		T tx = theta[0];
		T ty = theta[1];
		T tI = theta[2];
		T tbg = theta[3];

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			T dEy = tI * OneOverSqrt2PiSigma * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
			T dEy2 = tI * OneOverSqrt2PiSigma3 * ((y - ty - 0.5f) * exp(-Yexp1 * Yexp1) - (y - ty + 0.5f) * exp(-Yexp0 * Yexp0));
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				T dEx = tI * OneOverSqrt2PiSigma * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));
				T dEx2 = tI * OneOverSqrt2PiSigma3 * ((x - tx - 0.5f) * exp(-Xexp1 * Xexp1) - (x - tx + 0.5f) * exp(-Xexp0 * Xexp0));

				T mu = tbg + tI * DeltaX * DeltaY;
				T dmu_dx = dEx * DeltaY;
				T dmu_dy = dEy * DeltaX;
				T dmu_dI0 = DeltaX * DeltaY;

				T dmu2_dx2 = dEx2 * DeltaY;
				T dmu2_dy2 = dEy2 * DeltaX;

				const T firstOrder[] = { dmu_dx, dmu_dy, dmu_dI0, 1 };
				const T secondOrder[] = { dmu2_dx2, dmu2_dy2, 0, 0 };
				cb(y*imgw + x, mu, firstOrder, secondOrder);
			}
		}
	}


	PLL_DEVHOST Theta ComputeInitialEstimate(const T* sample, int index=0) const
	{
		Vector3f com = ComputeCOM(sample, imgw);
		return { com[0],com[1],com[2] * 0.9f,com[2] * 0.1f / (imgw*imgw) };
	}
};


template<typename TNumber>
struct Gauss2D_Model_XYI
{
	typedef typename TNumber T;
	typedef Vector3f Theta;

	enum { K = Theta::K };
	enum { NumStartPos = 1 };
	enum { NumConstants = 1 };

	T sigma, bg;
	int imgw;

	static const char* ThetaFormat() { return "x,y,I"; }

	PLL_DEVHOST int SampleCount() const { return imgw * imgw; }

	static PLL_DEVHOST T StopLimit(int k)
	{
		const float deltaStopLimit[] = { 1e-4f, 1e-4f, 1e-1f,1e-4f };
		return deltaStopLimit[k];
	}

	PLL_DEVHOST Gauss2D_Model_XYI(int roisize, float sigma, const float* bg) : imgw(roisize), sigma(sigma), bg(*bg) {}

	PLL_DEVHOST void CheckLimits(Theta& t, const Theta& last) const {
		t.elem[0] = clamp(t.elem[0], 2.0f, imgw - 1.0f);
		t.elem[1] = clamp(t.elem[1], 2.0f, imgw - 1.0f);

		if (t.elem[2] < 10.0f) t.elem[2] = 10.0f;
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeDerivatives(TCallback cb, const Theta& theta) const
	{
		const T OneOverSqrt2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * sigma);
		const T OneOverSqrt2Sigma = 1.0f / (sqrtf(2) * sigma);

		T tx = theta[0];
		T ty = theta[1];
		T tI = theta[2];

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			T dEy = OneOverSqrt2PiSigma * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				T dEx = OneOverSqrt2PiSigma * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));

				T mu = bg + tI * DeltaX * DeltaY;
				T dmu_dx = tI * dEx * DeltaY;
				T dmu_dy = tI * dEy * DeltaX;
				T dmu_dI0 = DeltaX * DeltaY;
				const T jacobian[] = { dmu_dx, dmu_dy, dmu_dI0 };
				cb(y*imgw + x, mu, jacobian);
			}
		}
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeSecondDerivatives(TCallback cb, Theta theta) const
	{
		const T OneOverSqrt2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * sigma);
		const T OneOverSqrt2Sigma = 1.0f / (sqrtf(2) * sigma);
		const T OneOverSqrt2PiSigma3 = 1.0f / (sqrtf(2 * MATH_PI) * sigma*sigma*sigma);

		T tx = theta[0];
		T ty = theta[1];
		T tI = theta[2];

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			T dEy = tI * OneOverSqrt2PiSigma * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
			T dEy2 = tI * OneOverSqrt2PiSigma3 * ((y - ty - 0.5f) * exp(-Yexp1 * Yexp1) - (y - ty + 0.5f) * exp(-Yexp0 * Yexp0));
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				T dEx = tI * OneOverSqrt2PiSigma * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));
				T dEx2 = tI * OneOverSqrt2PiSigma3 * ((x - tx - 0.5f) * exp(-Xexp1 * Xexp1) - (x - tx + 0.5f) * exp(-Xexp0 * Xexp0));

				T mu = bg + tI * DeltaX * DeltaY;
				T dmu_dx = dEx * DeltaY;
				T dmu_dy = dEy * DeltaX;
				T dmu_dI0 = DeltaX * DeltaY;

				T dmu2_dx2 = dEx2 * DeltaY;
				T dmu2_dy2 = dEy2 * DeltaX;

				const T firstOrder[] = { dmu_dx, dmu_dy, dmu_dI0 };
				const T secondOrder[] = { dmu2_dx2, dmu2_dy2, 0 };
				cb(y*imgw + x, mu, firstOrder, secondOrder);
			}
		}
	}


	PLL_DEVHOST Theta ComputeInitialEstimate(const T* sample, int index = 0) const
	{
		return ComputeCOM(sample, imgw);
	}
};


// A model that only fits signal vs background for a predefined PSF
template<typename TNumber>
struct IntensityBgModel
{
	typedef typename TNumber T;
	typedef Vector2f Theta;
	enum { K = Theta::K };

	int imgw;
	const float* psf; // precomputed [imgw*imgw]

	PLL_DEVHOST int SampleCount() const { return imgw * imgw; }

	// Precompute the integral of the gaussian over the pixel area, as this will stay constant during optimization
	PLL_DEVHOST static void ComputeGaussianPSF(float sigma, float tx, float ty, int imgw, float* psf)
	{
		const T OneOverSqrt2Sigma = 1.0f / (sqrtf(2) * sigma);

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				psf[y*imgw + x] = DeltaX * DeltaY;
			}
		}
	}

	PLL_DEVHOST T StopLimit(int k) const
	{
		const float deltaStopLimit[]={ 1e-2f,1e-4f };
		return deltaStopLimit[k];
	}

	PLL_DEVHOST IntensityBgModel(int roisize, const float* psf_space) : imgw(roisize), psf(psf_space) {}

	PLL_DEVHOST void CheckLimits(Theta& t, const Theta& last) const 
	{
		if (t.elem[0] < 1e-5f) t.elem[0] = 1e-5f;
		if (t.elem[1] < 1e-8f) t.elem[1] = 1e-8f;
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeDerivatives(TCallback cb, Theta theta) const
	{
		for (int y = 0; y < imgw; y++)
		{
			for (int x = 0; x < imgw; x++)
			{
				T mu = theta[0] * psf[y*imgw + x] + theta[1];

				const T firstOrder[] = { psf[y*imgw + x], 1 };
				cb(y*imgw + x, mu, firstOrder);
			}
		}
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeSecondDerivatives(TCallback cb, Theta theta) const
	{
		for (int y = 0; y < imgw; y++)
		{
			for (int x = 0; x < imgw; x++)
			{
				T mu = theta[0] * psf[y*imgw + x] + theta[1];

				const T firstOrder[] = { psf[y*imgw + x], 1 };
				const T secondOrder[] = { 0, 0 };
				cb(y*imgw + x, mu, firstOrder, secondOrder);
			}
		}
	}
};


template<typename TNumber>
struct Gauss2D_Model_XYZIBg
{
	typedef typename TNumber T;
	typedef Theta_XYZIBg Theta;
	typedef Gauss3D_Calibration Calibration;
	enum { K = Theta::K };
	enum { NumStartPos = 2 };
	static const char* ThetaFormat() { return "x,y,z,I,bg"; }
	enum { NumConstants = 0 };

	int imgw;
	Calibration calibration;

	PLL_DEVHOST Gauss2D_Model_XYZIBg(int roisize, const Calibration& calibration, const float* unused_constants=0) :
		imgw(roisize), calibration(calibration) {}

	PLL_DEVHOST int SampleCount() const { return imgw * imgw; }

	static PLL_DEVHOST T StopLimit(int k)
	{
		const float deltaStopLimit[] = { 1e-4f, 1e-4f, 1e-1f, 1e-4f, 1e-4f };
		return deltaStopLimit[k];
	}

	PLL_DEVHOST void CheckLimits(Theta& t, const Theta& last) const
	{
		// TODO: Get from calibration
		const T max_z = 0.3f;

		t.elem[0] = clamp(t.elem[0], 2.0f, imgw - 1.0f); // x
		t.elem[1] = clamp(t.elem[1], 2.0f, imgw - 1.0f); // y
		t.elem[2] = clamp(t.elem[2], -max_z, max_z); // z
		t.elem[3] = fmaxf(t.elem[3], 10.f); //I
		t.elem[4] = fmaxf(t.elem[4], 0.0f); // bg
	}

	PLL_DEVHOST Theta ComputeInitialEstimate(const T* sample, int index = 0) const
	{
		// Compute initial value
		T sumX = 0, sumY = 0, sum = 0;
		T bg = std::numeric_limits<T>::max();

		for (int y = 0; y < imgw; y++)
			for (int x = 0; x < imgw; x++) {
				T v = sample[y * imgw + x];
				sumX += x * v;
				sumY += y * v;
				sum += v;
				if (v < bg) bg = v;
			}

		float comx = sumX / sum;
		float comy = sumY / sum;

		Theta estim;
		estim[0] = comx;
		estim[1] = comy;
		estim[2] = index == 0 ? -0.3f : 0.3f; // Two different starting z
		estim[3] = sum - bg * imgw * imgw; // I
		estim[4] = bg; // bg
		return estim;
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeDerivatives(TCallback cb, const Theta& theta) const
	{
		T tx = theta[0];
		T ty = theta[1];
		T tz = theta[2];
		T tI = theta[3];
		T tbg = theta[4];

		T s0_x = calibration.x[0];
		T gamma_x = calibration.x[1];
		T d_x = calibration.x[2];
		T A_x = calibration.x[3];

		T s0_y = calibration.y[0];
		T gamma_y = calibration.y[1];
		T d_y = calibration.y[2];
		T A_y = calibration.y[3];

		T d_y2 = d_y * d_y;
		T d_x2 = d_x * d_x;

		const T tzx = tz - gamma_x; const T tzx2 = tzx * tzx; const T tzx3 = tzx2 * tzx;
		const T tzy = tz - gamma_y; const T tzy2 = tzy * tzy; const T tzy3 = tzy2 * tzy;

		const T sigma_x = s0_x * sqrt(1.0f + tzx2 / d_x2 + A_x * tzx3 / d_x2);
		const T sigma_y = s0_y * sqrt(1.0f + tzy2 / d_y2 + A_y * tzy3 / d_y2);

		const T OneOverSqrt2PiSigma_x = 1.0f / (sqrtf(2 * MATH_PI) * sigma_x);
		const T OneOverSqrt2Sigma_x = 1.0f / (sqrtf(2) * sigma_x);
		const T OneOverSqrt2PiSigma_y = 1.0f / (sqrtf(2 * MATH_PI) * sigma_y);
		const T OneOverSqrt2Sigma_y = 1.0f / (sqrtf(2) * sigma_y);

		for (int y = 0; y < imgw; y++)
		{
			T Yexp0 = (y - ty + .5f) * OneOverSqrt2Sigma_y;
			T Yexp1 = (y - ty - .5f) * OneOverSqrt2Sigma_y;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			T dEy = OneOverSqrt2PiSigma_y * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
			for (int x = 0; x < imgw; x++)
			{
				T Xexp0 = (x - tx + .5f) * OneOverSqrt2Sigma_x;
				T Xexp1 = (x - tx - .5f) * OneOverSqrt2Sigma_x;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
				T dEx = OneOverSqrt2PiSigma_x * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));

				T mu = tbg + tI * DeltaX * DeltaY;
				T dmu_dx = tI * dEx * DeltaY;
				T dmu_dy = tI * dEy * DeltaX;

				T G21x = 1 / (sqrt(2.0f * MATH_PI) * sigma_x * sigma_x) * (
					(x - tx - 0.5f) * exp(-(x - tx - 0.5f)*(x - tx - 0.5f) / (2.0f * sigma_x * sigma_x)) -
					(x - tx + 0.5f) * exp(-(x - tx + 0.5f)*(x - tx + 0.5f) / (2.0f * sigma_x * sigma_x)));
				T G21y = 1 / (sqrt(2.0f * MATH_PI) * sigma_y * sigma_y) * (
					(y - ty - 0.5f) * exp(-(y - ty - 0.5f)*(y - ty - 0.5f) / (2.0f * sigma_y * sigma_y)) -
					(y - ty + 0.5f) * exp(-(y - ty + 0.5f)*(y - ty + 0.5f) / (2.0f * sigma_y * sigma_y)));

				T dMuSigmaX = tI * DeltaY * G21x;
				T dMuSigmaY = tI * DeltaX * G21y;

				T dSigmaXThetaZ = s0_x * (2 * tzx / d_x2 + A_x * 3 * tzx2 / d_x2) /
					(2 * sqrt(1 + tzx2 / d_x2 + A_x * tzx3 / d_x2));
				T dSigmaYThetaZ = s0_y * (2 * tzy / d_y2 + A_y * 3 * tzy2 / d_y2) /
					(2 * sqrt(1 + tzy2 / d_y2 + A_y * tzy3 / d_y2));

				T dmu_dz = dMuSigmaX * dSigmaXThetaZ + dMuSigmaY * dSigmaYThetaZ;

				T dmu_dI0 = DeltaX * DeltaY;
				T dmu_dIbg = 1;
				const T jacobian[] = { dmu_dx, dmu_dy, dmu_dz, dmu_dI0, dmu_dIbg };
				cb(y*imgw + x, mu, jacobian);
			}
		}
	}
};



