/*

GaussianPSF_XYIBg:		Sigma
GaussianPSF_XYIBgSigma	No calib.
GaussianPSF_XYZIBg		Z calibration
CubicSplinePSF			Spline coefficients
BSplinePSF				Spline coefficients


CalibrationData
PSFImpl< Model, Calibration >

*/
#pragma once

#include <driver_types.h>
#include "DLLMacros.h"
#include "Vector.h"
#include "ContainerUtils.h"

template<typename T>
class DeviceVector;

class PSFBase
{
	std::string thetaFormat; 
	std::vector<int> sampleSize; // for example: [frames, height, width] or [height, width]
	int sampleCount;
	int numConstants;
	int thetaSize;
public:
	PSFBase(const std::vector<int>& sampleSize, int numConst, int thetaSize, const char* thetaFormat);
	virtual ~PSFBase();

	const char* ThetaFormat() { return thetaFormat.c_str(); }// Comma separated like "x,y,I,bg
	int ThetaSize() { return thetaSize;  }
	int SampleCount() { return sampleCount; }
	int NumConstants() { return numConstants; } // Parameters that vary per spot, but are not estimated (like roix,roiy)
	int SampleIndexDims() {	return (int)sampleSize.size(); } // Number of dimensions for a sample index (x,y or frame,x,y)
	int SampleSize(int dim) { return sampleSize[dim]; } // 0 <= dim < SampleIndexDims()
	std::vector<int> SampleSize() { return sampleSize; }
};

// Abstract PSF Model. All pointers point to host memory
class PSF : public PSFBase
{
public:
	PSF(const std::vector<int>& sampleSize, int numConst, int thetaSize, const char* thetaFormat) : PSFBase(sampleSize, numConst, thetaSize, thetaFormat) {}

	virtual void FisherMatrix(const float* theta, const float* h_const, const int* spot_pos, int numspots, float* d_FI) = 0;

	// d_image[numspots, SampleCount()], d_theta[numspots, ThetaSize()]
	virtual void ExpectedValue(float* expectedvalue, const float* theta, const float* _const, const int* spot_pos, int numspots) = 0;
	// d_deriv[numspots, ThetaSize(), SampleCount()], d_expectedvalue[numspots, SampleCount()], d_theta[numspots, ThetaSize()]
	virtual void Derivatives(float* deriv, float *expectedvalue, const float* theta, const float* _const, const int* spot_pos, int numspots) = 0;

	// d_sample[numspots, SampleCount()], d_initial[numspots, ThetaSize()], d_theta[numspots, ThetaSize()], d_iterations[numspots]
	virtual void MLE(const float* sample, const float* d_const, const int* spot_pos, const float* initial, float* theta, int* iterations, int numspots,
		int maxiterations, float levmarInitialAlpha, float* trace, int traceBufLen) = 0;

	template<typename T>
	struct VectorType {
		typedef std::vector<T> type;
	};
};


// Abstract PSF Model. All pointers with d_ prefix point to cuda memory. 
// Default implementation uses Derivatives() to implement MLE() and FisherMatrix()
class CUDA_PSF : public PSFBase
{
public:
	CUDA_PSF(const std::vector<int>& sampleSize, int numConst, int thetaSize, const char* thetaFormat);
	~CUDA_PSF();

	virtual void FisherMatrix(const float* d_theta, const float* d_const, const int* d_spot_pos, int numspots, float* d_FI, cudaStream_t stream) = 0;
	// d_image[numspots, SampleCount()], d_theta[numspots, ThetaSize()]
	virtual void ExpectedValue(float* expectedvalue, const float* d_theta, const float* d_const, const int* d_spot_pos, int numspots, cudaStream_t stream) = 0;
	// d_deriv[numspots, ThetaSize(), SampleCount()], d_expectedvalue[numspots, SampleCount()], d_theta[numspots, ThetaSize()]
	virtual void Derivatives(float* deriv, float *expectedvalue, const float* theta, const float* d_const, const int* d_spot_pos, int numspots, cudaStream_t stream) = 0;

	// d_sample[numspots, SampleCount()], d_params[numspots, numparams], d_initial[numspots, ThetaSize()], d_theta[numspots, ThetaSize()], d_iterations[numspots]
	virtual void MLE(const float* d_sample, const float* d_const, const int* d_spot_pos, const float* d_initial, float* d_theta, int* d_iterations,
		int numspots, int maxiterations, float levmarInitialAlpha, float* d_trace, int traceBufLen, cudaStream_t stream);

	template<typename T>
	struct VectorType {
		typedef DeviceVector<T> type;
	};

	virtual void SetMaxSpots(int maxspots);

protected:
	int maxspots;
	class DeviceMemoryBuffers;
	std::unique_ptr<DeviceMemoryBuffers> buffers;
	void InitBuffers();
};

CDLL_EXPORT PSF* PSF_WrapCUDA_PSF(CUDA_PSF* cuda_psf);

// Ignore the psf->Derivatives and implement Derivatives(), FisherMatrix() and MLE() using numerically computed derivatives
// Epsilon has size [ThetaSize()] and sets the numerical derivative step size for each theta element.
CDLL_EXPORT CUDA_PSF* CUDA_PSF_NumericalDerivatives(CUDA_PSF* psf, const float* epsilon);

// C/Python API - All pointers are host memory
CDLL_EXPORT void PSF_Delete(PSF* psf);
CDLL_EXPORT const char *PSF_ThetaFormat(PSF* psf);
CDLL_EXPORT int PSF_ThetaSize(PSF* psf);
CDLL_EXPORT int PSF_SampleCount(PSF* psf);
CDLL_EXPORT int PSF_NumConstants(PSF* psf);
CDLL_EXPORT void PSF_ComputeExpectedValue(PSF* psf, int numspots, const float* theta, const float* constants, const int* spotpos, float* ev);
//CDLL_EXPORT void PSF_ComputeInitialEstimate(PSF* psf, int numspots, const float* sample, const float* constants, float* theta);
CDLL_EXPORT void PSF_ComputeMLE(PSF* psf, int numspots, const float* sample, const float* constants, const int* spotpos, const float* initial, float* theta,
	int* iterations, int maxiterations, float levmarAlpha, float* trace, int traceBufLen);
CDLL_EXPORT void PSF_ComputeFisherMatrix(PSF* psf, int numspots, const float* theta, const float* constants, const int* spotpos, float* fi);
CDLL_EXPORT void PSF_ComputeDerivatives(PSF* psf, int numspots, const float* theta, const float* constants, const int* spotpos, float* derivatives, float* ev);

// Utils that use the PSF objects
CDLL_EXPORT void PSF_DrawSpots(PSF * psf, const float * theta, const float* constants, const Int2* roipos, int numspots);

