#include "PSFImpl.h"
#include "CudaUtils.h"


CDLL_EXPORT PSF* PSF_WrapCUDA_PSF(CUDA_PSF* psf)
{
	return new CUDA_PSF_Wrapper(psf);
}

class CUDA_PSF::DeviceMemoryBuffers {
public:
	DeviceMemoryBuffers(int numspots, int smpcount, int thetasize) :
		derivatives(numspots*smpcount*thetasize), expectedValue(numspots*smpcount)
	{}
	DeviceVector<float> derivatives, expectedValue;
};

void CUDA_PSF::InitBuffers()
{
	buffers.reset(new DeviceMemoryBuffers(maxspots, SampleCount(), ThetaSize()));
}

void CUDA_PSF::SetMaxSpots(int maxspots)
{
	this->maxspots = maxspots;
	InitBuffers();
}

CUDA_PSF::CUDA_PSF(const std::vector<int>& sampleSize, int numConst, int thetaSize, const char* thetaFormat) 
	: PSFBase(sampleSize, numConst, thetaSize, thetaFormat),  maxspots(0) {}

CUDA_PSF::~CUDA_PSF()
{}

void CUDA_PSF::MLE(const float * d_sample, const float * d_const, const int* d_spot_pos, const float * d_initial, float * d_theta, int * d_iterations,
	int numspots, int maxiterations, float levmarInitialAlpha, float * d_trace, int traceBufLen, cudaStream_t stream) 
{
	if (buffers.get() == 0)
		InitBuffers();

	float* d_deriv = buffers->derivatives.data();
	float* d_ev = buffers->expectedValue.data();

	for (int i = 0; i < maxiterations; i++)
	{
		Derivatives(d_deriv, d_ev, d_theta, d_const, d_spot_pos, numspots, stream);

		LaunchKernel(numspots, [=]__device__(int i) {
			
		}, 0, stream);
	}
}



CUDA_PSF_Wrapper::CUDA_PSF_Wrapper(CUDA_PSF * cudaPSF) : 
	PSF(cudaPSF->SampleSize(), cudaPSF->NumConstants(), cudaPSF->ThetaSize(), cudaPSF->ThetaFormat()), psf(cudaPSF)
{}


CUDA_PSF_Wrapper::~CUDA_PSF_Wrapper()
{
	delete psf;
}


void CUDA_PSF_Wrapper::FisherMatrix(const float * h_theta, const float* h_const, const int* spot_pos, int numspots, float * h_FI)
{
	DeviceVector<float> d_theta(numspots*ThetaSize(), h_theta);
	DeviceVector<float> d_FI(numspots*ThetaSize()*ThetaSize());
	DeviceVector<float> d_const(numspots*NumConstants(), h_const);
	DeviceVector<int> d_spot_pos(numspots*SampleIndexDims(), spot_pos);
	psf->FisherMatrix(d_theta.ptr(), d_const.ptr(), d_spot_pos.ptr(), numspots, d_FI.ptr(), 0);
	d_FI.CopyToHost(h_FI);
}


void CUDA_PSF_Wrapper::ExpectedValue(float * h_expectedvalue, const float * h_theta, const float* h_const, const int* spot_pos, int numspots)
{
	DeviceVector<float> d_ev(numspots*SampleCount());
	DeviceVector<float> d_theta(numspots*ThetaSize(), h_theta);
	DeviceVector<float> d_const(numspots*NumConstants(), h_const);
	DeviceVector<int> d_spot_pos(numspots*SampleIndexDims(), spot_pos);
	psf->ExpectedValue(d_ev.ptr(), d_theta.ptr(), d_const.ptr(), d_spot_pos.ptr(), numspots, 0);
	d_ev.CopyToHost(h_expectedvalue);
}

void CUDA_PSF_Wrapper::Derivatives(float * h_deriv, float * h_expectedvalue, const float * h_theta, const float* h_const, const int* spot_pos, int numspots)
{
	DeviceVector<float> d_theta(numspots*ThetaSize(), h_theta);
	DeviceVector<float> d_ev(numspots*SampleCount());
	DeviceVector<float> d_deriv(numspots*ThetaSize()*psf->SampleCount());
	DeviceVector<float> d_const(numspots*NumConstants(), h_const);
	DeviceVector<int> d_spot_pos(numspots*SampleIndexDims(), spot_pos);
	psf->Derivatives(d_deriv.ptr(), d_ev.ptr(), d_theta.ptr(), d_const.ptr(), d_spot_pos.ptr(), numspots, 0);
	d_ev.CopyToHost(h_expectedvalue);
	d_deriv.CopyToHost(h_deriv);
}

void CUDA_PSF_Wrapper::MLE(const float * h_sample, const float* h_const, const int* spot_pos, const float * h_initial, float * h_theta, int * h_iterations,
	int numspots, int maxiterations, float levmarInitialAlpha, float* h_trace, int traceBufLen)
{
	DeviceVector<float> d_smp(numspots*SampleCount(), h_sample);
	DeviceVector<float> d_initial(h_initial ? numspots * ThetaSize() : 0, h_initial);
	DeviceVector<float> d_theta(numspots*ThetaSize());
	DeviceVector<int> d_iterations(numspots);
	DeviceVector<float> d_trace(numspots*traceBufLen*ThetaSize());
	DeviceVector<float> d_const(numspots*NumConstants(), h_const);
	DeviceVector<int> d_spot_pos(numspots*SampleIndexDims(), spot_pos);
	psf->MLE(d_smp.ptr(), d_const.ptr(), d_spot_pos.ptr(), d_initial.ptr(), d_theta.ptr(), d_iterations.ptr(), numspots, maxiterations, levmarInitialAlpha, d_trace.ptr(), traceBufLen, 0);
	d_theta.CopyToHost(h_theta);
	if (h_trace) d_trace.CopyToHost(h_trace);
	if (h_iterations) d_iterations.CopyToHost(h_iterations);
}


