#pragma once

#include <streambuf>
#include "PSF.h"
#include "Estimation.h"
#include "ContainerUtils.h"
#include "ThreadUtils.h"

template<typename TModel, typename TCalibration, typename TSampleOffset>
class CUDA_PSFImpl : public CUDA_PSF
{
public:
	typedef typename TCalibration Calibration;
	typedef TModel Model;
	typedef TSampleOffset SampleOffset;
	typedef typename Model::Theta Theta;

	SampleOffset sampleOffset;
	int roisize;

	CUDA_PSFImpl(const TSampleOffset& sampleOffset, int roisize) : CUDA_PSF({ roisize,roisize }, TModel::NumConstants, TModel::K, TModel::ThetaFormat()),
		sampleOffset(sampleOffset), roisize(roisize)
	{}

	virtual const TCalibration& GetCalibration() = 0;

	void FisherMatrix(const float* d_theta, const float *d_const, const int* d_spot_pos, int numspots, float* d_FI, cudaStream_t stream) override
	{
		const Theta* theta = (const Theta*)d_theta;
		TCalibration calib = GetCalibration();
		SampleOffset ofs = sampleOffset;
		int roisize=this->roisize;
		auto *fi = (typename FisherMatrixType<Theta>::type*)d_FI;
		LaunchKernel(numspots, [=]__device__(int i) {
			TModel model(roisize, calib, &d_const[i*TModel::NumConstants]);
			fi[i] = ComputeFisherMatrix(model, ofs, theta[i]);
		}, 0, stream);
	}

	void ExpectedValue(float* d_image, const float* d_theta, const float *d_const, const int* d_spot_pos, int numspots, cudaStream_t stream) override
	{
		const Theta* theta = (const Theta*)d_theta;
		TCalibration calib = GetCalibration();
		SampleOffset ofs = sampleOffset;
		int roisize = this->roisize;
		int sc = SampleCount();
		LaunchKernel(numspots, [=]__device__(int i) {
			TModel model(roisize, calib, &d_const[i*TModel::NumConstants]);
			ComputeExpectedValue(theta[i], model, ofs, &d_image[sc*i] );
		}, 0, stream);
	}

	void Derivatives(float* d_deriv, float *d_expectedvalue, const float* d_theta, const float *d_const, const int* d_spot_pos, int numspots, cudaStream_t stream) override
	{
		const Theta* theta = (const Theta*)d_theta;
		TCalibration calib = GetCalibration();
		SampleOffset ofs = sampleOffset;
		int smpcount = SampleCount();
		int roisize = this->roisize;
		int K = ThetaSize();
		LaunchKernel(numspots,  [=]__device__(int i) {
			TModel model(roisize, calib, &d_const[i*TModel::NumConstants]);
			ComputeDerivatives(theta[i], model, ofs, &d_deriv[i*smpcount*K], &d_expectedvalue[i*smpcount]);
		}, 0, stream);
	}

	void MLE(const float* d_sample, const float *d_const, const int* d_spot_pos, const float* d_initial, float* d_theta, int* d_iterations,
		int numspots, int maxiterations, float levmarInitialAlpha, float* d_trace, int traceBufLen, cudaStream_t stream)  override
	{
		Theta* theta = (Theta*)d_theta;
		Theta* trace = (Theta*)d_trace;
		const Theta* initial = (const Theta*)d_initial;
		TCalibration calib = GetCalibration();
		int smpcount = SampleCount();
		int roisize = this->roisize;
		SampleOffset ofs = sampleOffset;
		LaunchKernel(numspots, [=]__device__(int i) {
			TModel model(roisize, calib, &d_const[i*TModel::NumConstants]);
			auto r = ComputeMLE(&d_sample[i*smpcount], model, ofs, maxiterations, initial ? &initial[i] : 0, 
				&trace[traceBufLen*i], traceBufLen, levmarInitialAlpha);
			if (d_iterations) d_iterations[i] = r.iterations;
			theta[i] = r.estimate;
		}, 0, stream);
	}
};

template<typename TModel, typename TCalibration, typename TSampleOffset>
class PSFImpl : public PSF
{
public:
	typedef typename TCalibration Calibration;
	typedef typename TModel Model;
	typedef typename TSampleOffset SampleOffset;

	SampleOffset sampleOffset;
	int roisize;
	typedef typename Model::Theta Theta;

	PSFImpl(const TSampleOffset& sampleOffset, int roisize) : PSF({ roisize,roisize }, TModel::NumConstants, TModel::K, TModel::ThetaFormat()), 
		sampleOffset(sampleOffset),roisize(roisize)
	{}

	virtual const TCalibration& GetCalibration() = 0;


	// Inherited via PSF
	virtual void FisherMatrix(const float *theta, const float *h_const, const int* spot_pos, int numspots, float * FI) override
	{
		const Theta* theta_ = (const Theta*)theta;
		const TCalibration& calib = GetCalibration();
		auto* fi = (typename FisherMatrixType<Theta>::type*)FI;
		ParallelFor(numspots, [&](int i) {
			TModel model(roisize, calib, &h_const[i*TModel::NumConstants]);
			fi[i] = ComputeFisherMatrix(model, sampleOffset, theta_[i]);
		});
	}

	virtual void ExpectedValue(float * expectedvalue, const float * theta, const float * h_const, const int* spot_pos, int numspots)  override
	{
		const Theta* theta_ = (const Theta*)theta;
		const TCalibration& calib = GetCalibration();
		ParallelFor(numspots, [&](int i) {
			TModel model(roisize, calib, &h_const[i*TModel::NumConstants]);
			ComputeExpectedValue(theta_[i], model, sampleOffset, &expectedvalue[model.SampleCount() *i]);
		});
	}
	virtual void Derivatives(float * deriv, float * expectedvalue, const float * theta, const float *h_const, const int* spot_pos, int numspots) override
	{
		const Theta* theta_ = (const Theta*)theta;
		const TCalibration& calib = GetCalibration();
		ParallelFor(numspots, [&](int i) {
			TModel model(roisize, calib, &h_const[i*TModel::NumConstants]);
			ComputeDerivatives(theta_[i], model, sampleOffset, &deriv[model.SampleCount()*Theta::K*i], &expectedvalue[model.SampleCount()*i]);
		});
	}
	virtual void MLE(const float * samples, const float *h_const, const int* spot_pos, const float * initial, float * theta,
		int * iterations, int numspots, int maxiterations, float levmarInitialAlpha, float * trace, int traceBufLen) override
	{
		Theta* trace_ = (Theta*)trace;
		Theta* theta_ = (Theta*)theta;
		const Theta* initial_ = (const Theta*)initial;
		const TCalibration& calib = GetCalibration();
		int smpcount = SampleCount();

		ParallelFor(numspots, [&](int i) {
			TModel model(roisize, calib, &h_const[i*TModel::NumConstants]);
			auto r = ComputeMLE(&samples[i*smpcount], model, sampleOffset, maxiterations, 
				initial_ ? &initial_[i] : (Theta*)0, &trace_[traceBufLen*i], traceBufLen, levmarInitialAlpha);
			if (iterations) iterations[i] = r.iterations;
			theta_[i] = r.estimate;
		});
	}
};


// PSF Implementation for models with a plain-old-data (POD) calibration type (no [cuda] memory management needed)
template<typename BasePSFImpl>
class SimpleCalibrationPSF : public BasePSFImpl
{
public:
	typedef typename BasePSFImpl::Calibration Calibration;
	typedef typename BasePSFImpl::SampleOffset SampleOffset;
	Calibration calib;

	const Calibration& GetCalibration() override { return calib; }

	SimpleCalibrationPSF(const Calibration& calib, const SampleOffset& smpofs, int roisize) :
		BasePSFImpl(smpofs, roisize), calib(calib)
	{}
};

// Wraps a CUDA_PSF into a host-memory PSF
class CUDA_PSF_Wrapper : public PSF {
	CUDA_PSF* psf;
public:
	CUDA_PSF_Wrapper(CUDA_PSF* psf);
	~CUDA_PSF_Wrapper();

	// Inherited via PSF
	virtual void FisherMatrix(const float * h_theta, const float* h_const, const int* spot_pos, int numspots, float * h_FI) override;
	virtual void ExpectedValue(float * h_expectedvalue, const float * h_theta, const float* h_const, const int* spot_pos, int numspots) override;
	virtual void Derivatives(float * h_deriv, float * h_expectedvalue, const float * h_theta, const float* h_const, const int* spot_pos, int numspots) override;
	virtual void MLE(const float * h_sample, const float* h_const, const int* spot_pos, const float * h_initial, float * h_theta, int * h_iterations,
		int numspots, int maxiterations, float levmarInitialAlpha, float* d_trace, int traceBufLen) override;
};

