
#include "MemLeakDebug.h"

#include "palala.h"
#include "SILM.h"
#include "SolveMatrix.h"

#include "MathUtils.h"
#include "PSFModels/GaussianPSF.h"

#include "Estimation.h"
#include "StringUtils.h"
#include "palala.h"

#include "CameraCalibration.h"

#include "PSFModels/GaussianPSFModels.h"
#include "RandomDistributions.h"

#include "ExcitationModel.h"

#include "FFT.h"

#include "ImageProcessingQueue.h"
#include "ExcitationModel.h"
#include "SpotDetection/SpotDetector.h"

#include "DebugImageCallback.h"

#pragma warning(disable : 4503) // decorated name length exceeded, name was truncated


#include "SILM_Models.h"


struct SILM_SampleIndex
{
	int f, x, y;
};


template<typename T>
PLL_DEVHOST T squared(T x) { return x * x; }

inline PLL_DEVHOST Vector4f Compute2x2Inverse(const Vector4f& m)
{
	float d = 1.0f / (m[0] * m[3] - m[1] * m[2]);
	return { d*m[3], -d * m[1], -d * m[2], d*m[0] };
}

PLL_DEVHOST float NormalDistrProb(float smp, float mu, float variance)
{
	float a = mu - smp;
	return exp(-a * a / (2 * variance)) / sqrt(2 * MATH_PI*variance);
}



PLL_DEVHOST float LogNormalDistrProb(float smp, float mu, float variance)
{
	float a = mu - smp;
	return -a * a / (2 * variance) - log(sqrt(2 * MATH_PI*variance));
}


CDLL_EXPORT void SILM_ASW_ComputeOnOffProb(const float* rois, const SILM_Modulation* modulation, Vector4f* gaussFits, 
	Vector2f* IBg, Vector2f* probOnOff, Vector2f* crlbVariance, float* expectedIntensities, const SILM_ASW_Params& params, int numframes,
	int numspots, const int* startPatterns, const int2* roipos, bool useCuda)
{
	SILM_ASW_Params p = params;
	std::vector<float> psf_space(p.roisize*p.roisize*numspots);
	int maxIterations = 100;
	int imgsize = p.roisize*p.roisize;

	parallel_for_uc(numspots, useCuda, 
	PALALA(int i, const float* rois, const SILM_Modulation* modulation, const int* startPatterns, 
		Vector2f* crlbVariance, float* expectedIntensities, float* psf_space,
		const Vector4f *gaussFits, Vector2f* IBg, Vector2f* probOnOff, const int2* roipos) {
		//int numep, float sigma, const SILM_Modulation* modulation, int2 roiPos
		FixedFunctionExcitationPatternModel epModel(p.numepp, p.psfSigma, modulation, roipos[i]);
		float *psf = &psf_space[imgsize*i];
		Vector4f gaussFit = gaussFits[i];
		SampleOffset_None<float> sampleOffset;
		IntensityBgModel<float>::ComputeGaussianPSF(p.psfSigma, gaussFit[0], gaussFit[1], p.roisize, psf);
		IntensityBgModel<float> IBg_model(p.roisize, psf);
		SampleOffset_None<float> smpofs;

		float silmPerPatternBg = gaussFit[3] * (p.numepp / (float)numframes) / p.numepp;
		Vector2f ibg_off{ 1e-4f, silmPerPatternBg };
		float variance_off = Compute2x2Inverse(ComputeFisherMatrix(IBg_model, smpofs, ibg_off))[0];

		int e = startPatterns[i];
		for (int f = 0; f < numframes; f++) {
			float q, dqdx, dqdy;
			epModel.ExcitationPattern(q, dqdx, dqdy, e, { gaussFit[0], gaussFit[1] });

			const float* spot_img = &rois[imgsize*numframes*i + imgsize * f];
			float sum = 0.0f;
			for (int j = 0; j < imgsize; j++)
				sum += spot_img[j];
			auto r = LevMarOptimize(spot_img, { 1,0 }, IBg_model, sampleOffset, maxIterations);
			//auto r = NewtonRaphson(spot_img, { fmaxf(10, sum-silmPerPatternBg*imgsize),silmPerPatternBg }, IBg_model, sampleOffset, maxIterations);
			IBg[i*numframes + f] = r.estimate;
#ifndef __CUDA_ARCH__
			DebugPrintf("IBg[%d,%d]={%.4f,%.2f}. x=%.4f, y=%.4f, sum=%.4f\n", 
				i,f, r.estimate[0],r.estimate[1], gaussFit[0],gaussFit[1],sum);
#endif

			// Theta_I*Q_f
			Vector2f expected_ibg{ gaussFit[2] * q, silmPerPatternBg };

			float variance_on = Compute2x2Inverse(ComputeFisherMatrix(IBg_model, smpofs, expected_ibg))[0];

			expectedIntensities[i*numframes + f] = expected_ibg[0];
			crlbVariance[i*numframes + f] = { variance_on,variance_off };

			float on_prob = LogNormalDistrProb(r.estimate[0], expected_ibg[0], variance_on);
			float off_prob = LogNormalDistrProb(r.estimate[0], 0, variance_off);
			probOnOff[i*numframes + f] = { on_prob,off_prob};

			e++;
			if (e == p.numepp) e = 0;
		}

	}, const_array(rois, numspots*imgsize*numframes),
		const_array(modulation, p.numepp),
		const_array(startPatterns, numspots),
		out_array(crlbVariance, numspots*numframes),
		out_array(expectedIntensities, numspots*numframes),
		psf_space,
		const_array(gaussFits, numspots),
		out_array(IBg, numspots*numframes),
		out_array(probOnOff, numspots*numframes),
		const_array(roipos, numspots));
//	PALALA(int i, const float* rois, const SILM_Modulation* modulation, const int* startPatterns, float* psf_space,
	//	const Vector4f *gaussFits, Vector2f* IBg, Vector2f* probOnOff, const Int2* roipos)

}

CDLL_EXPORT void SILM_ASW_ComputeMLE(const float * imageData, const SILM_Modulation* modulation,
	Gauss2D_EstimationResult *results, int numspots, int numframes, const int *startPatterns, const SILM_ASW_Params& silm_p,
	const SILM_Theta* initialValues, const int2* roipos, int flags, SILM_Theta* optimizerTrace, int traceBufferLength)
{
	SILM_ASW_Params p = silm_p;
	int imgsize = p.roisize * p.roisize;
	bool cuda = flags & SILM_MLE_CUDA;

	std::vector<float> imageSums(imgsize*numspots);

	parallel_for_uc(numspots, cuda,
		PALALA(int i, const float* imageData, const SILM_Modulation* mod, Gauss2D_EstimationResult* results, 
			const int* startPatterns, const int2* roipos, float* imgsums, const SILM_Theta* initialValues, SILM_Theta* trace)
	{
		const SILM_Modulation* spot_modulation = &mod[p.numepp * i];
		const float* spot_img = &imageData[p.roisize*p.roisize*numframes*i];

		SampleOffset_None<float> sampleOffset;
		SILM_Theta initialValue;
		if (!initialValues) {
			float* sum = &imgsums[imgsize*i];
			ComputeImageSums(spot_img, sum, p.roisize, p.roisize, numframes);
			auto com = ComputeCOM(sum, p.roisize);
			initialValue = { com[0],com[1],com[2] * 0.9f,com[3] * 0.1f / (p.roisize*p.roisize) };

			Gauss2D_Model_XYIBg<float> gaussModel(p.roisize, p.psfSigma);
			initialValue = LevMarOptimize(sum, initialValue, gaussModel, sampleOffset, 10).estimate;
		}
		else
			initialValue = initialValues[i];

		FixedFunctionExcitationPatternModel epModel(p.numepp, p.psfSigma, spot_modulation, roipos[i]);
		SILM_Model<FixedFunctionExcitationPatternModel>
			model(p, startPatterns[i], numframes, epModel);

		OptimizerResult<SILM_Theta> r = LevMarOptimize(spot_img, initialValue, model,
			sampleOffset, p.maxLevMarIterations, &trace[traceBufferLength*i], traceBufferLength, p.levMarLambdaStep);

		auto fi = ComputeFisherMatrix(model, SampleOffset_None<float>(), r.estimate);
		results[i] = Gauss2D_EstimationResult(initialValue, r, ComputeCRLB(fi));

	}, const_array(imageData, numspots*imgsize*numframes),
		const_array(modulation, p.numepp*numspots),
		out_array(results, numspots),
		const_array(startPatterns, numspots),
		const_array(roipos, numspots),
		imageSums,
		const_array(initialValues, initialValues ? numspots : 0),
		out_array(optimizerTrace, numspots*traceBufferLength));
}



CDLL_EXPORT void SILM_ASW_ComputeFisherMatrix(float * expectedValue, Gauss2D_FisherMatrix* fiMatrix, const SILM_Modulation * mod,
	const SILM_Theta * theta, int numspots, int nframes, const int *startPatterns, const SILM_ASW_Params& p)
{
	int nspotpixels = p.roisize*p.roisize*nframes;
	for (int i = 0; i < numspots; i++) {
		auto* spot_mod = &mod[p.numepp*i];

		FixedFunctionExcitationPatternModel epModel(p.numepp, p.psfSigma, spot_mod, int2());
		SILM_Model<FixedFunctionExcitationPatternModel>
			model(p, startPatterns[i], nframes, epModel);

		fiMatrix[i] = ComputeFisherMatrix(model, SampleOffset_None<float>(), theta[i], &expectedValue[i*nspotpixels]);
	}
}

    
CDLL_EXPORT void SILM_ASW_ComputeBackgroundMatrix(float* bgMatrix, int numepp, int width, int height, float psfSigma, float omega, float patternDepth, const float* phi)
{
	for (int e = 0; e < numepp; e++) {
		for (int y = 0; y < height; y++)
			for (int x = 0; x < width; x++) {
				float& bg = bgMatrix[e*width*height + width * y + x];
				//ps/M+2*m*exp(-(1/2)*omega^2*sigma^2)*sin(omega*xc-phi)*sin((1/2)*omega*ps)/(omega*M)
				int c = (e < numepp * 0.5f) ? x : y;
				bg = ( 1.0f + 2*patternDepth / omega * expf(-omega * omega * psfSigma*psfSigma * 0.5f) *
					sinf(omega*c - phi[e])*sinf(omega / 2)) / numepp;
			}
	}
}







CDLL_EXPORT void SILM_ProjectPointData(const Vector3f *xyI, int numpts, int projectionWidth, 
	float scale, int numProjAngles, const float *projectionAngles, float* output, float* shifts)
{
	int pw = projectionWidth;
	parallel_for_cpu(numProjAngles, [&](int p) {
		float* proj = &output[pw*p];
		for (int i = 0; i < pw; i++)
			proj[i] = 0.0f;
		float kx = cos(projectionAngles[p]);
		float ky = sin(projectionAngles[p]);

		double moment = 0.0;
		for (int i = 0; i < numpts; i++)
			moment += kx * xyI[i][0] + ky * xyI[i][1];
		float shift = pw / 2 - scale * float(moment) / numpts;
		if(shifts) shifts[p] = shift;

		for (int i = 0; i < numpts; i++) {
			float coord = kx * xyI[i][0] + ky * xyI[i][1];
			int index = int(scale * coord + shift + 0.5f);
			if (index < 0 || index> pw - 1)
				continue;
			proj[index] += xyI[i][2];
		}
	});
}


CDLL_EXPORT void SILM_SignalCorrelation(const float* signalsA, int numsigA, 
										const float* signalsB, int numsigB, int siglen, float *output, int useConvolution)
{
	DeviceVector<float> d_output(numsigA*numsigB);

	std::vector<cuFloatComplex> a_complex(siglen*numsigA);
	for (size_t i = 0; i < a_complex.size(); i++) a_complex[i] = make_cuFloatComplex(signalsA[i], 0.0f);
	std::vector<cuFloatComplex> b_complex(siglen*numsigB);
	for (size_t i = 0; i < b_complex.size(); i++) b_complex[i] = make_cuFloatComplex(signalsB[i], 0.0f);

	// Compute the signal A FFTs
	DeviceVector<cuFloatComplex> d_a(siglen * numsigA, a_complex.data());
	DeviceVector<cuFloatComplex> d_b(siglen * numsigB, b_complex.data());

	// Real to complex is 
	
	DeviceVector<cuFloatComplex> a_fft(siglen*numsigA);
	DeviceVector<cuFloatComplex> b_fft(siglen*numsigB);

	FFTPlan1D(siglen, numsigA, true).Transform(d_a.ptr(), a_fft.ptr());
	FFTPlan1D(siglen, numsigB, true).Transform(d_b.ptr(), b_fft.ptr());

#if 0
	ShowDebugImage(siglen, numsigA, 1,
		Transform(a_fft.ToVector(), [=](cuFloatComplex v) { return v.x; }).data(), "real(a_fft)");
	ShowDebugImage(siglen, numsigA, 1,
		Transform(a_fft.ToVector(), [=](cuFloatComplex v) { return v.y; }).data(), "imag(a_fft)");
#endif

#if 0
	ShowDebugImage(siglen, numsigB, 1,
		Transform(b_fft.ToVector(), [=](cuFloatComplex v) { return v.x; }).data(), "real(b_fft)");
	ShowDebugImage(siglen, numsigB, 1,
		Transform(b_fft.ToVector(), [=](cuFloatComplex v) { return v.y; }).data(), "imag(b_fft)");
#endif

	// Do all the work
	// [numSignals, numFreq]: for every direction a list of correlations with each frequency
	cuFloatComplex* d_a_fft = a_fft.ptr();
	cuFloatComplex* d_b_fft = b_fft.ptr();

	DeviceVector<cuFloatComplex> conv_fft(siglen*numsigB);
	DeviceVector<cuFloatComplex> conv(siglen*numsigB);

	cuFloatComplex* d_conv_fft = conv_fft.ptr();
	cuFloatComplex* d_conv = conv.ptr();
	float* out = d_output.ptr();
	FFTPlan1D ifft_plan(siglen, numsigB, false);
	float f = 1.0f / siglen;
	for (int sig = 0; sig < numsigA; sig++) {
		if (useConvolution) {
			LaunchKernel(siglen, numsigB, [=]__device__(int x, int b) {
				d_conv_fft[siglen*b + x] = cuCmulf(d_b_fft[b * siglen + x] * f, d_a_fft[sig * siglen + x] * f);
			});
		}
		else {
			LaunchKernel(siglen, numsigB, [=]__device__(int x, int b) {
				d_conv_fft[siglen*b + x] = cuCmulf(d_b_fft[b * siglen + x] * f, cuConjf(d_a_fft[sig * siglen + x] * f));
			});
		}
		ifft_plan.Transform(d_conv_fft, d_conv); // inverse transform

		const float f = 1.0f/siglen;
		LaunchKernel(numsigB, [=]__device__(int b) {
			float sum = 0.0f;
			for (int x = 0; x < siglen; x++) {
				cuFloatComplex v = d_conv[b*siglen + x]*f;
				sum += (v * cuConjf(v)).x;
			}
			out[sig * numsigB + b] = sum;
		});
	}

	d_output.CopyToHost(output, false);
}


CDLL_EXPORT void SILM_CorrelatePhaseAndFreq(const Vector2f* pts, int numpts,
	const float* freq_list, int numfreq, const float* phase_list, int numphases, float* output)
{
	parallel_for_cuda(numfreq, numphases,PALALA(int f, int p, const Vector2f *pts, const float* freq_list, const float* phase_list, float* output) {
		float freq = freq_list[f];
		float phase = phase_list[p];
		float sum = 0.0f, c=0.0f;
		for (int i = 0; i < numpts; i++)
		{
			float x = pts[i][0];
			float y = pts[i][1];

			float v = y * sin(x*freq - phase) - c;
			float t = sum + v;
			c = (t - sum) - v;
			sum = t;
		}
		output[f*numphases + p] = sum;
	}, const_array(pts, numpts), const_array(freq_list, numfreq), const_array(phase_list, numphases), out_array(output, numphases*numfreq));
}




class SILM_EstimationTask : public IComputeTask
{
public:
	typedef Gauss2D_EstimationResult Result;

	SILM_EstimationTask(const ImageQueueConfig& iqc, ISpotImageSource* sourceData, 
		const SILM_ASW_Params& silmParams, float silmMinIntensity,
		ResultContainer<Result>& resultContainer, std::vector<SILM_Modulation> modulation)
		: targetContainer(resultContainer),
		spotImageSource(sourceData),
		maxSpots(sourceData->MaxSpotCount()),
		params(silmParams),
		silmMinIntensity(silmMinIntensity),
		streamData(iqc.cudaStreamCount),
		numFrames(iqc.historyLength),
		modulation(modulation)
	{
		for (int i = 0; i < streamData.size(); i++) {
			auto& sd = streamData[i];
			sd.results.Init(maxSpots);
			sd.h_results.Init(maxSpots);
			sd.frameIBg.Init(maxSpots*numFrames);
			sd.frameIBg_crlb.Init(maxSpots*numFrames);
			int imgsize = silmParams.roisize*silmParams.roisize;
			sd.summedROIs.Init(maxSpots*imgsize);
		}
	}

	SILM_ASW_Params params;
	ResultContainer<Result>& targetContainer;
	ISpotImageSource* spotImageSource;
	int maxIterations = 100;
	int maxSpots;
	int numFrames;
	float silmMinIntensity;
	DeviceVector<SILM_Modulation> modulation;

	struct StreamData
	{
		int numSpots;
		DeviceVector<Result> results;
		PinnedArray<Result> h_results;

		DeviceVector<float> summedROIs, gaussPSFs;
		DeviceVector<Vector2f> frameIBg, frameIBg_crlb;
	};

	std::vector<StreamData> streamData;

	const char* GetTaskName()
	{
		return "2D SILM Estimator";
	}

	bool Execute(IComputeTaskManager* tm, int streamIndex)
	{
		StreamData& sd = streamData[streamIndex];
		const float* imageData = spotImageSource->GetSpotImages(sd.numSpots, streamIndex);
		Result* d_results = sd.results.ptr();

		SILM_ASW_Params silm_p = params;
		int roisize = spotImageSource->ROISize();
		cudaStream_t stream = tm->GetStream(streamIndex);

		const float *d_spotScores = spotImageSource->GetSpotLocationSource().GetSpotList(streamIndex).spotScores;
		const int2* roiPositions = spotImageSource->GetROICornerPositions(streamIndex);
		int imgsize = silm_p.roisize*silm_p.roisize;
		int numframes = this->numFrames;

		float* summedROIs = sd.summedROIs.data();
		float* gaussPSFs = sd.gaussPSFs.data();
		Vector2f* frameIBg = sd.frameIBg.data();
		Vector2f* frameIBg_crlb = sd.frameIBg_crlb.data();

		int startPattern = tm->GetFrameIndex(streamIndex) % modulation.size();

		const SILM_Modulation* d_modulation = modulation.ptr();
		float silmMinIntensity = this->silmMinIntensity;

		if (sd.numSpots == 0)
			return true;

		LaunchKernel(sd.numSpots, [=] __device__(int i) {
			int2 roiPos = roiPositions[i];
			FixedFunctionExcitationPatternModel epModel(silm_p.numepp, silm_p.psfSigma, &d_modulation[i*silm_p.numepp], roiPos);
			SampleOffset_None<float> smpofs;

			const float* src = &imageData[imgsize*numframes*i];
			float* summedROI = &summedROIs[imgsize*i];
			ComputeImageSums(src, &summedROI[i], silm_p.roisize, silm_p.roisize, numframes);

			Gauss2D_Model_XYIBg<float> gaussModel(silm_p.roisize, silm_p.psfSigma);
			auto gauss2DResult = ComputeMLE(summedROI, gaussModel, smpofs, 15);

			float* psf = &gaussPSFs[imgsize*i];
			IntensityBgModel<float>::ComputeGaussianPSF(silm_p.psfSigma, gauss2DResult.estimate[0], gauss2DResult.estimate[1], silm_p.roisize, psf);
			IntensityBgModel<float> intensityBgModel(silm_p.roisize, psf);

			bool discard = false;
			for (int f = 0; f < numframes; f++) {
				auto r = NewtonRaphson(&src[imgsize*f], { gauss2DResult.estimate[2]/numframes, 1.0f }, intensityBgModel, smpofs, 15);
				frameIBg[i*numframes + f] = r.estimate;
				//frameIBg_crlb[i*numframes + f] = ComputeCRLB(ComputeFisherMatrix(intensityBgModel, smpofs, r.estimate));

				if (r.estimate[0] < silmMinIntensity) {
					discard = true;
					break;
				}
			}

			if (!discard) {
				SILM_Model<FixedFunctionExcitationPatternModel> model(silm_p, startPattern, numframes, epModel);

				OptimizerResult<SILM_Theta> r = LevMarOptimize(src, gauss2DResult.estimate, model,
					smpofs, silm_p.maxLevMarIterations, 0, 0, silm_p.levMarLambdaStep);

				auto fi = ComputeFisherMatrix(model, SampleOffset_None<float>(), r.estimate);
				auto result = Gauss2D_EstimationResult(gauss2DResult.estimate, r, ComputeCRLB(fi), d_spotScores[i]);
				result.roiPosition = { roiPos.x, roiPos.y };

				d_results[i] = result;
			}
			else {
				d_results[i].iterations = 0;
			}
		}, 0, stream);

		sd.results.CopyToHost(sd.h_results.data(), true, stream);
		return true;
	}

	bool Completed(IComputeTaskManager* tm, int streamIndex)
	{
		auto& sd = this->streamData[streamIndex];

		// Count spots that have passed silmMinIntensity
		int count = 0;
		for (int i = 0; i < sd.numSpots; i++) {
			if (streamData[streamIndex].h_results[i].iterations) count++;
		}

		// copy results to result container
		std::vector<Result> results(count);
		int index = 0;
		for (int i = 0; i < sd.numSpots; i++) {
			if (streamData[streamIndex].h_results[i].iterations)
				results[index++] = streamData[streamIndex].h_results[i];
		}

		int frame = tm->GetFrameIndex(streamIndex);
		targetContainer.AddResults(frame, std::move(results));
		return true;
	}
};


CDLL_EXPORT ImageProcessingQueue<Gauss2D_EstimationResult>* SILM_CreateImageQueue(const SILM_ASW_Params & p, 
	const ImageQueueConfig & iqc, ISpotDetectorFactory* spotDetection, int maxSpotsPerFrame, 
	float silmMinIntensity, const SILM_Modulation*modulation)
{
	auto *q = new ImageProcessingQueue<Gauss2D_EstimationResult>(iqc);

	auto* frameSummer = new FrameSumTask(iqc.cudaStreamCount, iqc.historyLength, true, *q);
	// Run spot detection on the summed frames
	auto* spotDetector = spotDetection->CreateInstance(iqc, frameSummer);
	// Extract the ROI movies
	auto* extractor = new SpotImageExtractorTask(iqc.cudaStreamCount, *q, *spotDetector, maxSpotsPerFrame, p.roisize, iqc.historyLength);
	// Run SILM on extracted ROI movies
	auto* localizer = new SILM_EstimationTask(iqc, extractor, p, silmMinIntensity, q->resultContainer, 
		std::vector<SILM_Modulation>(modulation, modulation + p.numepp));

	q->AddTask(frameSummer);
	q->AddTask(spotDetector);
	q->AddTask(extractor);
	q->AddTask(localizer);

	return q;
}


CDLL_EXPORT void SILM_DFT2D_Points(const Vector3f* xyI, int numpts, const Vector2f* k, int numk, Vector2f* output, bool useCuda)
{
	parallel_for_uc(numk, useCuda, PALALA(int i, const Vector3f* xyI, const Vector2f* k, Vector2f* output) {
		Vector2f k_ = k[i];

		// Use Kahan Sum for reduced errors
		Vector2f sum, c;

		for (int j = 0; j < numpts; j++) {
			float p = xyI[j][0] * k_[0] + xyI[j][1] * k_[1];
			float I = xyI[j][2];
			Vector2f input = { cos(p) * I, sin(p) * I };
			Vector2f y = input - c;
			Vector2f t = sum + y;
			c = (t - sum) - y;
			sum = t;
		}
		output[i] = sum;
	}, const_array(xyI, numpts),
		const_array(k, numk),
		out_array(output, numk));
}
