#pragma once
#include "CudaUtils.h"

#pragma warning(disable : 4503) // decorated name length exceeded, name was truncated

PLL_DEVHOST inline void ComputeImageSums(const float* frames, float* dst, int w, int h, int numframes)
{
	for (int i = 0; i < w*h; i++)
		dst[i] = 0;
	for (int f = 0; f < numframes; f++)
	{
		for (int i = 0; i < w*h; i++)
			dst[i] += frames[f*w*h + i];
	}
}

template<typename TExcitationPatternModel>
struct SILM_Model
{
	typedef float T;
	typedef SILM_Theta Theta;
	enum { K = Theta::K };

	PLL_DEVHOST SILM_Model(const SILM_ASW_Params& p, int startPattern,
		int numframes, const TExcitationPatternModel& excitationPatternModel)
		: params(p), startPattern(startPattern), numframes(numframes), epModel(excitationPatternModel)
	{}

	const SILM_ASW_Params& params;
	int startPattern, numframes; // ep = (frameNum + startPattern) % params.numepp
	const TExcitationPatternModel& epModel;

	PLL_DEVHOST int SampleCount() const { return params.roisize * params.roisize * numframes; }

	static PLL_DEVHOST T StopLimit(int k)
	{
		const float deltaStopLimit[] = { 1e-5f, 1e-5f, 1e-2f,1e-4f };
		return deltaStopLimit[k];
	}

	PLL_DEVHOST void CheckLimits(Theta& t, const Theta& last) const {
		t.elem[0] = clamp(t.elem[0], 2.0f, params.roisize - 1.0f);
		t.elem[1] = clamp(t.elem[1], 2.0f, params.roisize - 1.0f);
		t.elem[2] = fmaxf(t.elem[2], 10.0f);
		t.elem[3] = fmaxf(t.elem[3], 0.0f);
	}

	template<typename TCallback>
	PLL_DEVHOST void ComputeDerivatives(TCallback cb, SILM_Theta theta) const
	{
		int w = params.roisize;
		T _1oSq2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * params.psfSigma);
		T _1oSq2Sigma = 1.0f / (sqrtf(2) * params.psfSigma);

		T thetaX = theta[0], thetaY = theta[1], thetaI = theta[2], thetaBg = theta[3];

		for (int f = 0; f < numframes; f++) {
			int e = (startPattern + f) % params.numepp;

			// compute Q, dQ/dx, dQ/dy
			float Q, dQdx, dQdy;
			epModel.ExcitationPattern(Q, dQdx, dQdy, e, { thetaX,thetaY });

			for (int y = 0; y < w; y++) {
				// compute Ey,dEy/dy
				T Yexp0 = (y - thetaY + .5f) * _1oSq2Sigma;
				T Yexp1 = (y - thetaY - .5f) * _1oSq2Sigma;
				T Ey = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
				T dEy = _1oSq2PiSigma * (exp(-Yexp1 * Yexp1) - exp(-Yexp0 * Yexp0));
				for (int x = 0; x < w; x++) {
					// compute Ex,dEx/dx
					T Xexp0 = (x - thetaX + .5f) * _1oSq2Sigma;
					T Xexp1 = (x - thetaX - .5f) * _1oSq2Sigma;
					T Ex = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);
					T dEx = _1oSq2PiSigma * (exp(-Xexp1 * Xexp1) - exp(-Xexp0 * Xexp0));

					const float bg = epModel.BackgroundPattern(e, x, y);

					// combine
					const float Exy = Ex * Ey;
					const float mu = thetaI * (Q*Exy) + thetaBg * bg;

					const float dmu_x = thetaI * (dQdx * Exy + Q * dEx * Ey);
					const float dmu_y = thetaI * (dQdy * Exy + Q * Ex * dEy);
					const float dmu_I = Q * Exy;
					const float dmu_bg = bg;

					const float jacobian[] = { dmu_x, dmu_y,dmu_I,dmu_bg };

					cb(f*w*w + y * w + x, mu, jacobian);
				}
			}
		}
	}
};




template<typename T>
PLL_DEVHOST typename Vector<T, 4> SILM_ComputeCOMEstimate(const T* sample, int w, int numframes)
{
	int imgsize = w * w;

	// Compute initial value
	T sumX = 0, sumY = 0, sum = 0;
	for (int f = 0; f < numframes; f++)
		for (int y = 0; y < w; y++)
			for (int x = 0; x < w; x++) {
				T v = 0;
				v += sample[f*imgsize + y * w + x];
				sumX += x * v;
				sumY += y * v;
				sum += v;
			}

	float comx = sumX / sum;
	float comy = sumY / sum;

	typename Vector<T, 4> estim;
	estim[0] = comx;
	estim[1] = comy;
	estim[2] = sum * 0.9f;
	estim[3] = sum * 0.1f / (w*w);
	return estim;
}
