#include "MemLeakDebug.h"

#include "Vector.h"
#include "CudaUtils.h"
#include "GLRTSpotDetector.h"
#include "SpotDetector.h"
#include "Estimation.h"
#include "MathUtils.h"

#include "PSFModels/GaussianPSFModels.h"

GLRTSpotDetector::GLRTSpotDetector(int nStreams, IImageSource* src, float psfSigma, int maxSpotImages, float falsePositiveRateTarget) :
	psfSigma(psfSigma),
	maxSpotImages(maxSpotImages),
	streamData(nStreams),
	fdrTarget(falsePositiveRateTarget),
	src(src)
{
	for (auto& sd : streamData) {
		sd.numspots = 0;
		sd.spotImages.Init(roisize*roisize*maxSpotImages);
	}

	roisize = int(3 * (2 * psfSigma + 1));

	int2 imageSize = src->GetImageSize();
	offset = int2{ roisize / 2, roisize / 2 };
	croppedSize = int2{ imageSize.x - roisize, imageSize.y - roisize };

	testStatistic.Init(croppedSize);
}

GLRTSpotDetector::~GLRTSpotDetector()
{}

struct GLRT_Model
{
	typedef int2 SampleIndex;
	typedef float T;

	float sigma;
	int roisize, imgPitch;
	int2 roipos;

	PLL_DEVHOST float StartLambdaStep() const { return 0.1f; }

	PLL_DEVHOST int SampleElementIndex(int2 i) const
	{
		return (i.y + roipos.y)*imgPitch + i.x + roipos.x;
	}
};

struct GLRT_H1_Model : public GLRT_Model
{
	PLL_DEVHOST GLRT_H1_Model(const GLRT_Model& org) { *(GLRT_Model*)this = org; }

	typedef Vector2f Theta;
	PLL_DEVHOST Theta Process(Theta t, const T * sample, int it, T LL) const {
		if (t.elem[0] < 0.0f) t.elem[0] = 0.0f;
		if (t.elem[1] < 1e-9f) t.elem[1] = 1e-9f;
		return t;
	}

	template<typename TCallback>
	PLL_DEVHOST void ProcessSamples(TCallback cb, Theta theta) const
	{
		T _1oSq2PiSigma = 1.0f / (sqrtf(2 * MATH_PI) * sigma);
		T _1oSq2Sigma = 1.0f / (sqrtf(2) * sigma);

		const T tx = roisize/2;
		const T ty = roisize/2;
		T tI = theta[0];
		T tbg = theta[1];

		for (int y = 0; y < roisize; y++)
		{
			T Yexp0 = (y - ty + .5f) * _1oSq2Sigma;
			T Yexp1 = (y - ty - .5f) * _1oSq2Sigma;
			T DeltaY = 0.5f * erf(Yexp0) - 0.5f * erf(Yexp1);
			for (int x = 0; x < roisize; x++)
			{
				T Xexp0 = (x - tx + .5f) * _1oSq2Sigma;
				T Xexp1 = (x - tx - .5f) * _1oSq2Sigma;
				T DeltaX = 0.5f * erf(Xexp0) - 0.5f * erf(Xexp1);

				T mu = tbg + tI * DeltaX * DeltaY;
				T dmu_dI0 = DeltaX * DeltaY;
				T dmu_dIbg = 1;
				const T jacobian[] = { dmu_dI0, dmu_dIbg };
				cb(int2{ x,y }, mu, jacobian);
			}
		}
	}
};


bool GLRTSpotDetector::Execute(IComputeTaskManager* tm, int streamIndex)
{
	auto& img = src->GetImage(streamIndex);

	LaunchKernel(testStatistic.width, testStatistic.height, [=]__device__(int x, int y) {
		
		

	}, 0, tm->GetStream(streamIndex));
	return true;
}

bool GLRTSpotDetector::Completed(IComputeTaskManager* tm, int streamIndex)
{
	return false;
}

const int * GLRTSpotDetector::GetSpotIndexListDevicePtr(int streamIndex, int & numIndices)
{
	numIndices = 0;
	return 0;
}


