#include "MemLeakDebug.h"

#include <cassert>
#include "CudaUtils.h"

#include "ImageFilters.h"

#include <thrust/functional.h>
#include <cub/cub.cuh>
#include "Vector.h"
#include "SpotDetector.h"

#include "DebugImageCallback.h"

// DLL API
CDLL_EXPORT ISpotDetectorFactory* SpotDetector_Configure(const SpotDetectorConfig& config)
{
	return new SpotDetectorTask::Factory{ config };
}
CDLL_EXPORT void SpotDetector_DestroyFactory(ISpotDetectorFactory* factory)
{
	delete factory;
}

ISpotDetectorTask * SpotDetectorTask::Factory::CreateInstance(const ImageQueueConfig & iqc, IImageSource * imgSource)
{
	return new SpotDetectorTask(iqc, *imgSource, config);
}


SpotDetectorTask::SpotDetectorTask(const ImageQueueConfig& iqc, IImageSource& imgSource, const SpotDetectorConfig& sdcfg) :
	config(sdcfg),
	streamData(iqc.cudaStreamCount),
	imgSource(imgSource)
{
	for (auto& sd : streamData) {
		sd.instance.Init(iqc.ImageSize());
	}
}

SpotDetectorTask::~SpotDetectorTask()
{}

SpotLocationList SpotDetectorTask::GetSpotList(int streamIndex)
{
	SpotLocationList list;

	list.numSpots = streamData[streamIndex].instance.numspots;
	list.spotIndices = streamData[streamIndex].instance.selectedIndices.ptr();
	list.spotScores = streamData[streamIndex].instance.d_spotValidity.ptr();

	return list;
}




struct non_negative
{
	__host__ __device__
		bool operator()(const int &x)
	{
		return x >= 0;
	}
};

void SpotDetector::Init(int2 imgsize)
{
	temp.Init(imgsize);
	filtered1.Init(imgsize);
	filtered2.Init(imgsize);
	maxFiltered.Init(imgsize);
	int numpixels = imgsize.x*imgsize.y;
	indices.Init(numpixels);
	selectedIndices.Init(numpixels);
	spotScores.Init(numpixels);
	d_spotValidity.Init(numpixels);
	numFoundSpots.Init(1);
	numspots = 0;
}

void SpotDetector::Detect(const DeviceImage<float>& srcImage, cudaStream_t stream, const SpotDetectorConfig& config)
{
	UniformFilter2D(srcImage, temp, filtered1, config.uniformFilter1Size, stream);
	UniformFilter2D(srcImage, temp, filtered2, config.uniformFilter2Size, stream);

	ApplyBinaryOperator2D(filtered1, filtered2, filtered1, thrust::minus<float>(), stream);
	ComparisonFilter2D(filtered1, temp, maxFiltered, config.maxFilterSize, thrust::maximum<float>(), stream);

	// convert to indices
	float* d_filtered1 = filtered1.data;
	float* d_max = maxFiltered.data;
	int pitch = filtered1.pitch / sizeof(float);
	int *d_indices = indices.ptr();
	int w = filtered1.width;
	int h = filtered1.height;
	float minIntensity = config.minIntensity, maxIntensity = config.maxIntensity;
	int roisize = config.roisize;

	LaunchKernel(w, h, [=]__device__(int x, int y) {
		float a = d_filtered1[y*pitch + x];
		float b = d_max[y*pitch + x];
		bool isMax = a == b
			&& x > roisize / 2
			&& y > roisize / 2
			&& x + (roisize - roisize / 2) < w - 1
			&& y + (roisize - roisize / 2) < h - 1;

		if (isMax && b > minIntensity && b < maxIntensity)
			d_indices[y*w + x] = y * w + x;
		else
			d_indices[y*w + x] = -1;
	}, 0, stream);

	if (!partitionTempStorage.ptr())
	{
		size_t tempBytes;
		CUDAErrorCheck(cub::DevicePartition::If(0, tempBytes, indices.ptr(), selectedIndices.ptr(), numFoundSpots.ptr(), (int)indices.size(), non_negative(), stream));
		partitionTempStorage.Init(tempBytes);
	}

	size_t tmpsize = partitionTempStorage.size();
	CUDAErrorCheck(cub::DevicePartition::If(partitionTempStorage.ptr(), tmpsize, indices.ptr(),
		selectedIndices.ptr(), numFoundSpots.ptr(), (int)indices.size(), non_negative(), stream));

	float* d_spotValidity = this->d_spotValidity.ptr();
	const int* d_selIndices = selectedIndices.ptr();
	LaunchKernel(w*h, [=]__device__(int i) {
		int index = d_selIndices[i];
		if (index >= 0) {
			int srcy = index / w;
			int srcx = index % w; // this is really slow..
			float spotValue = d_max[pitch*srcy + srcx];
			d_spotValidity[i] = spotValue;
		}
	}, 0, stream);
	this->d_spotValidity.CopyToHost(spotScores.data(), true, stream);

	numFoundSpots.CopyToHost(&numspots, true, stream);
}

bool SpotDetectorTask::Execute(IComputeTaskManager* tm, int streamIndex)
{
	auto& sd = streamData[streamIndex];
	const DeviceImage<float>& srcImage = imgSource.GetImage(streamIndex);
	cudaStream_t stream = tm->GetStream(streamIndex);
	sd.instance.Detect(srcImage, stream, config);
	return true;
}

bool SpotDetectorTask::Completed(IComputeTaskManager*tm, int streamIndex)
{
	auto& sd = streamData[streamIndex];

//	for (int i = 0; i < sd.numspots; i++) {
//		DebugPrintf("spot[%d] validity = %f\n", i, sd.spotValidity[i]);
//	}

	// since the stream is idle, numspots is now synchronized
	return true;
	//	return sd.numspots > 0;
}

const char * SpotDetectorTask::GetTaskName()
{
	return "SSA Spot Detector";
}



const int2 * SpotImageExtractorTask::GetROICornerPositions(int streamIndex)
{
	return streamData[streamIndex].spotROICornerPositions.ptr();
}

const float* SpotImageExtractorTask::GetSpotImages(int & numspots, int streamindex)
{
	numspots = streamData[streamindex].numSpots;
	return streamData[streamindex].spotImages.ptr();
}

const char * SpotImageExtractorTask::GetTaskName()
{
	return "Spot Image Extractor";
}

SpotImageExtractorTask::SpotImageExtractorTask(int nStreams, IFrameSource& src, ISpotLocationListSource& listSource, int maxSpots, int roisize, int numframes)
	: listSource(listSource), maxSpots(maxSpots), roisize(roisize), numframes(numframes), imgSize(src.GetImageSize()), src(src)
{
	streamData.resize(nStreams);
	for (int i = 0; i < streamData.size(); i++) {
		streamData[i].spotImages.Init(roisize*roisize*maxSpots*numframes);
		streamData[i].spotROICornerPositions.Init(maxSpots);
	}
}


bool SpotImageExtractorTask::Execute(IComputeTaskManager*tm, int streamIndex)
{
	auto& sd = streamData[streamIndex];
	auto spotList = listSource.GetSpotList(streamIndex);
	sd.numSpots = spotList.numSpots;

	if (!sd.numSpots)
		return true;

	int startFrame = tm->GetFrameIndex(streamIndex) - numframes + 1;
	if (startFrame < 0) {
		sd.numSpots = 0;
		return false;
	}

	int2* d_cornerPos = sd.spotROICornerPositions.ptr();
	int2 imgSize = this->imgSize;
	int roisize = this->roisize, numframes = this->numframes;

	LaunchKernel(sd.numSpots, [=]__device__(int i) {
		int centerPixelIndex = spotList.spotIndices[i];
		int centerX = centerPixelIndex % imgSize.x;
		int centerY = centerPixelIndex / imgSize.x;
		int xpos = centerX - roisize / 2;
		int ypos = centerY - roisize / 2;

		d_cornerPos[i] = make_int2(xpos, ypos);
	}, 0,tm->GetStream(streamIndex));

	for (int f = 0; f < numframes; f++) {
		auto *frame = src.GetFrame(f + startFrame);
		if (frame == 0)
			return false;

		const DeviceImage<float>& srcImage = *frame;

		if (sd.numSpots > maxSpots)
			sd.numSpots = maxSpots;

		float* d_out = sd.spotImages.ptr();
		auto src = srcImage.GetIndexer();

		//DebugPrintf("extracting %d spots from frame %d\n", sd.numSpots, queue->GetFrameIndex(streamIndex));

		LaunchKernel(roisize, roisize, sd.numSpots, [=]__device__(int x, int y, int spot) {
			int2 pos = d_cornerPos[spot] + int2{ x,y };
			
			if (pos.x < 0) pos.x = 0;
			if (pos.y < 0) pos.y = 0;
			if (pos.x >= imgSize.x) pos.x = imgSize.x - 1;
			if (pos.y >= imgSize.y) pos.y = imgSize.y - 1;

			d_out[y*roisize + x + spot * (roisize*roisize * numframes) + roisize * roisize*f] = src.pixel(pos.x, pos.y);

		}, 0, tm->GetStream(streamIndex));
	}

//	cudaDeviceSynchronize();
//	auto data=sd.spotImages.ToVector();
//	WriteTIFF("firstspots.tiff", &data[0], roisize, roisize * 10);

	//auto result = cudaStreamQuery(tm->GetStream(streamIndex));

	return true;
}

bool SpotImageExtractorTask::Completed(IComputeTaskManager*tm, int streamIndex)
{
	return true;
}


FrameSumTask::FrameSumTask(int nStreams, int numframes, bool average, IFrameSource& src)
	:average(average), src(src), numframes(numframes)
{
	streamData.resize(nStreams);
	for (int i = 0; i < nStreams; i++) {
		streamData[i].summedImage.Init(src.GetImageSize());
	}
}

bool FrameSumTask::Execute(IComputeTaskManager* tm, int streamIndex)
{
	auto& sd = streamData[streamIndex];
	cudaStream_t stream = tm->GetStream(streamIndex);
	sd.summedImage.Clear(stream);
	sd.frameCount = 0;

	int curFrame = tm->GetFrameIndex(streamIndex);
	auto dst = sd.summedImage.GetIndexer();

	for (int f = 0; f < numframes; f++) {
		if (src.GetFrame(curFrame - numframes + f + 1))
			sd.frameCount++;
	}
	float factor = average ? 1.0f / sd.frameCount : 1.0f;

	for (int f = 0; f < numframes; f++) {
		int idx = curFrame - numframes + f + 1;
		auto* frame = src.GetFrame(idx);
		if (frame)
		{
			auto src = frame->GetConstIndexer();
			LaunchKernel(sd.summedImage.width, sd.summedImage.height, [=]__device__(int x, int y) {
				dst.pixel(x, y) += src.pixel(x, y) * factor;
			}, 0, stream);
		}
	}

	return true;
}



CDLL_EXPORT int SpotDetector_ProcessFrame(const float* frame, int width, int height,
	int maxSpots, float* spotScores, Int2* cornerPos, float* rois, const SpotDetectorConfig & cfg)
{
	SpotDetector inst;
	inst.Init({ width,height }); 

	DeviceImage<float> img(frame, width, height);
	inst.Detect(img, 0, cfg);

	cudaStreamSynchronize(0);
	size_t numspots = inst.numspots;
	if (numspots > maxSpots) numspots = maxSpots;
	PinnedArray<int> pixelIndices(maxSpots);
	inst.selectedIndices.CopyToHost(pixelIndices.data(), numspots);

	int roisize = cfg.roisize;
	for (int i = 0; i < numspots; i++)
	{
		int centerPixelIndex = pixelIndices[i];
		Int2 center = { centerPixelIndex % width, centerPixelIndex / height };
		Int2 corner = center - roisize / 2;
		cornerPos[i] = corner;
		spotScores[i] = inst.spotScores[i];

		for (int y = 0; y < roisize; y++) {
			for (int x = 0; x < roisize; x++)
				rois[i*roisize*roisize + y * roisize + x] = frame[(corner[1] + y)*width + corner[0] + x];
		}
	}

	return numspots;
}

