/*

Spot detection algorithm, implemented following the method in: 

Huang F, Schwartz SL, Byars JM, Lidke KA (2011). Simultaneous multiple emitter
fitting for single molecule super-resolution imaging. Biomed Opt Express 2, 13771393.

*/
#pragma once

#include "CudaUtils.h"
#include "ImageProcessingQueue.h"

struct SpotLocationList {
	const float* spotScores; // actual unit depends on implementation of SpotDetector
	const int* spotIndices; // pixel indices into frame
	int numSpots;
};

class ISpotLocationListSource
{
public:
	virtual SpotLocationList GetSpotList(int streamIndex) = 0;
};

class ISpotImageSource
{
public:
	virtual const int2 * GetROICornerPositions(int streamIndex) = 0;
	virtual const float* GetSpotImages(int& numspots, int streamIndex) = 0; // should return cuda memory
	virtual ISpotLocationListSource& GetSpotLocationSource() = 0;
	virtual int ROISize() = 0;
	virtual int MaxSpotCount() = 0;
};



class ISpotDetectorTask : public IComputeTask, public ISpotLocationListSource
{
public:
	virtual SpotLocationList GetSpotList(int streamIndex) = 0;
};

struct SpotDetectorConfig
{
	// In a nutshell:
	// detectionImage = uniform1 - uniform2
	// Selected spot locations = max(detectionImage, maxFilterSize) == detectionImage
	int uniformFilter1Size, uniformFilter2Size, maxFilterSize;

	// Roisize is used to remove ROIs near the image border
	int roisize;

	// Only spots where detectionImage > intensityThreshold are selected
	float minIntensity, maxIntensity;

	SpotDetectorConfig(float sigma, int roisize, float minIntensity,float maxIntensity) {
		uniformFilter1Size = int(1 + sigma * 2);
		uniformFilter2Size = uniformFilter1Size * 2;
		this->roisize = roisize;
		this->minIntensity = minIntensity;
		this->maxIntensity = maxIntensity;
		maxFilterSize = int(sigma * 3+1);
	}
};

class ISpotDetectorFactory
{
public:
	virtual ~ISpotDetectorFactory() {}
	virtual ISpotDetectorTask* CreateInstance(const ImageQueueConfig& iqc, IImageSource* imgSource) = 0;
};


class SpotDetector
{
public:
	DeviceImage<float> temp, filtered1, filtered2;
	DeviceImage<float> maxFiltered;
	DeviceVector<int> indices, selectedIndices;
	DeviceVector<float> d_spotValidity;
	PinnedArray<float> spotScores;

	DeviceVector<int> numFoundSpots;
	DeviceVector<uint8_t> partitionTempStorage; // for cub::DevicePartition::If

	int numspots;

	void Init(int2 imgsize);
	void Detect(const DeviceImage<float>& srcImage, cudaStream_t stream, const SpotDetectorConfig&config);
};

class SpotDetectorTask : public ISpotDetectorTask
{
public:
	class Factory : public ISpotDetectorFactory
	{
	public:
		Factory(SpotDetectorConfig cfg) : config(cfg) {}
		ISpotDetectorTask* CreateInstance(const ImageQueueConfig& iqc, IImageSource* imgSource);
		SpotDetectorConfig config;
	};

	SpotDetectorTask(const ImageQueueConfig& iqc, IImageSource& imgSource, const SpotDetectorConfig& sdcfg);
	~SpotDetectorTask();

	SpotLocationList GetSpotList(int streamIndex); 

	bool Execute(IComputeTaskManager* tm, int streamIndex);
	bool Completed(IComputeTaskManager* tm, int streamIndex);
	const char *GetTaskName();

protected:
	SpotDetectorConfig config;
	IImageSource& imgSource;
	
	struct StreamData
	{
		SpotDetector instance;
	};

	std::vector<StreamData> streamData;
};

CDLL_EXPORT void SpotDetector_DestroyFactory(ISpotDetectorFactory* factory);
CDLL_EXPORT ISpotDetectorFactory* SpotDetector_Configure(const SpotDetectorConfig& config);


// Standalone usage of spot detector
CDLL_EXPORT int SpotDetector_ProcessFrame(const float* frame, int width, int height,
	int maxSpots, float* spotScores, Int2* cornerPos, float* rois, const SpotDetectorConfig & cfg);


class SpotImageExtractorTask : public IComputeTask, public ISpotImageSource
{
public:
	SpotImageExtractorTask(int nStreams, IFrameSource& src,  ISpotLocationListSource& listSource, int maxSpots, int roisize, int numframes);

	bool Execute(IComputeTaskManager* tm, int streamIndex);
	bool Completed(IComputeTaskManager* tm, int streamIndex);

	const int2 * GetROICornerPositions(int streamIndex);
	ISpotLocationListSource& GetSpotLocationSource() { return listSource; 	}
	const float* GetSpotImages(int& numspots, int streamIndex);
	int ROISize() { return roisize; }
	int MaxSpotCount() { return maxSpots; }
	int NumFrames() { return numframes; }

	const char *GetTaskName();

protected:
	struct StreamData {
		DeviceVector<float> spotImages;
		DeviceVector<int2> spotROICornerPositions;
		int numSpots;
	};

	IFrameSource& src;
	ISpotLocationListSource & listSource;
	int roisize, maxSpots, numframes;
	int2 imgSize;
	std::vector<StreamData> streamData;
};


class FrameSumTask : public IComputeTask, public IImageSource {
public:
	FrameSumTask(int nstreams, int numframes, bool average, IFrameSource& src);
	bool Execute(IComputeTaskManager* tm, int streamIndex);

	int GetSummedFrameCount(int streamIndex) const { return streamData[streamIndex].frameCount; }
	const DeviceImage<float>& GetImage(int streamIndex) { return streamData[streamIndex].summedImage; }
	int2 GetImageSize() { return src.GetImageSize(); }

	struct StreamData {
		DeviceImage<float> summedImage;
		int frameCount;
	};
	bool average;
protected:
	std::vector<StreamData> streamData;
	int numframes;
	IFrameSource& src;
};


